"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareClaimGovernanceRewardsTransaction = exports.prepareEarlyClaimGovernanceRewardsTransaction = exports.prepareBurnTransactions = exports.prepareClaimPremintTransaction = exports.prepareUnmintTransactions = exports.prepareUnmintPremintTransaction = exports.prepareMintTransactions = exports.getUserLiquidGovernanceInfo = exports.getDistributorInfo = void 0;
const algosdk_1 = require("algosdk");
const utils_1 = require("../../utils");
const abiContracts_1 = require("./constants/abiContracts");
/**
 *
 * Returns information regarding the given liquid governance distributor.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param distributor - distributor to query about
 * @returns DistributorInfo[] distributor info
 */
function getDistributorInfo(indexerClient, distributor) {
    return __awaiter(this, void 0, void 0, function* () {
        const { appId } = distributor;
        const res = yield indexerClient.lookupApplications(appId).do();
        const state = res["application"]["params"]["global-state"];
        const dispenserAppId = Number((0, utils_1.getParsedValueFromState)(state, "dispenser_app_id") || 0);
        const commitEnd = BigInt((0, utils_1.getParsedValueFromState)(state, "commit_end") || 0);
        const periodEnd = BigInt((0, utils_1.getParsedValueFromState)(state, "period_end") || 0);
        const totalCommitment = BigInt((0, utils_1.getParsedValueFromState)(state, "total_commitment") || 0);
        const totalCommitmentClaimed = BigInt((0, utils_1.getParsedValueFromState)(state, "total_commitment_claimed") || 0);
        const canClaimAlgoRewards = Boolean((0, utils_1.getParsedValueFromState)(state, "can_claim_algo_rewards") || 0);
        const rewardsPerAlgo = BigInt((0, utils_1.getParsedValueFromState)(state, "rewards_per_algo") || 0);
        const totalRewardsClaimed = BigInt((0, utils_1.getParsedValueFromState)(state, "total_rewards_claimed") || 0);
        const isBurningPaused = Boolean((0, utils_1.getParsedValueFromState)(state, "is_burning_paused") || 0);
        // optional
        const premintEndState = (0, utils_1.getParsedValueFromState)(state, "premint_end");
        const premintEnd = premintEndState !== undefined ? BigInt(premintEndState) : undefined;
        return {
            currentRound: res["current-round"],
            dispenserAppId,
            premintEnd,
            commitEnd,
            periodEnd,
            totalCommitment,
            totalCommitmentClaimed,
            canClaimAlgoRewards,
            rewardsPerAlgo,
            totalRewardsClaimed,
            isBurningPaused,
        };
    });
}
exports.getDistributorInfo = getDistributorInfo;
/**
 *
 * Returns information regarding a user's liquid governance commitment.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param distributor - distributor to query about
 * @param userAddr - user address to get governance info about
 * @returns UserCommitmentInfo[] user commitment info
 */
function getUserLiquidGovernanceInfo(indexerClient, distributor, userAddr) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const { appId } = distributor;
        // get user account local state
        const req = indexerClient.lookupAccountAppLocalStates(userAddr).applicationID(appId);
        const res = yield req.do();
        // user local state
        const state = (_b = (_a = res["apps-local-states"]) === null || _a === void 0 ? void 0 : _a.find((app) => app.id === appId)) === null || _b === void 0 ? void 0 : _b["key-value"];
        if (state === undefined)
            throw new Error("Unable to find commitment for: " + userAddr + ".");
        const commitment = BigInt((0, utils_1.getParsedValueFromState)(state, "commitment") || 0);
        const commitmentClaimed = BigInt((0, utils_1.getParsedValueFromState)(state, "commitment_claimed") || 0);
        // optional
        const premintState = (0, utils_1.getParsedValueFromState)(state, "premint");
        const premint = premintState !== undefined ? BigInt(premintState) : undefined;
        return {
            currentRound: res["current-round"],
            premint,
            commitment,
            commitmentClaimed,
        };
    });
}
exports.getUserLiquidGovernanceInfo = getUserLiquidGovernanceInfo;
/**
 *
 * Returns a group transaction to mint gALGO for ALGO at a one-to-one rate.
 * If in the commitment period then also commits user into governance.
 *
 * @param dispenser - dispenser to mint gALGO from
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to send and gALGO to mint
 * @param includeOptIn - whether to include an opt in transaction (must be opted in if minting in commitment period)
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the minter (must pass to be eligible for revenue share)
 * @returns Transaction[] mint transactions
 */
function prepareMintTransactions(dispenser, distributor, senderAddr, amount, includeOptIn, params, note) {
    var _a;
    const atc = new algosdk_1.AtomicTransactionComposer();
    const payment = {
        txn: (0, utils_1.transferAlgoOrAsset)(0, senderAddr, (0, algosdk_1.getApplicationAddress)(distributor.appId), amount, Object.assign(Object.assign({}, params), { fee: 0, flatFee: true })),
        signer: utils_1.signer,
    };
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "mint"),
        methodArgs: [payment, dispenser.gAlgoId, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
        note,
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    // for ledger compatibility (max 2 app args), remove index references which are not strictly needed
    txns[1].appArgs = (_a = txns[1].appArgs) === null || _a === void 0 ? void 0 : _a.slice(0, -2);
    // user must be opted in before they can mint in the commitment period
    if (includeOptIn)
        txns.unshift((0, algosdk_1.makeApplicationOptInTxn)(senderAddr, Object.assign(Object.assign({}, params), { fee: 1000, flatFee: true }), distributor.appId));
    return (0, algosdk_1.assignGroupID)(txns);
}
exports.prepareMintTransactions = prepareMintTransactions;
/**
 *
 * Returns a transaction to unmint pre-minted gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to unmint and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction unmint pre-mint transaction
 */
function prepareUnmintPremintTransaction(dispenser, distributor, senderAddr, amount, params, note) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "unmint_premint"),
        methodArgs: [amount, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
        note,
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareUnmintPremintTransaction = prepareUnmintPremintTransaction;
/**
 *
 * Returns a group transaction to unmint gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to send and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction[] unmint transactions
 */
function prepareUnmintTransactions(dispenser, distributor, senderAddr, amount, params, note) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    const assetTransfer = {
        txn: (0, utils_1.transferAlgoOrAsset)(dispenser.gAlgoId, senderAddr, (0, algosdk_1.getApplicationAddress)(dispenser.appId), amount, Object.assign(Object.assign({}, params), { fee: 0, flatFee: true })),
        signer: utils_1.signer,
    };
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "unmint"),
        methodArgs: [assetTransfer, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
        note,
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return (0, algosdk_1.assignGroupID)(txns);
}
exports.prepareUnmintTransactions = prepareUnmintTransactions;
/**
 *
 * Returns a transaction to claim pre-minted gALGO.
 * Can be called on behalf of yourself or another user.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param receiverAddr - account address for the pre-minter that will receiver the gALGO
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction claim pre-mint transaction
 */
function prepareClaimPremintTransaction(dispenser, distributor, senderAddr, receiverAddr, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "claim_premint"),
        methodArgs: [receiverAddr, dispenser.gAlgoId, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareClaimPremintTransaction = prepareClaimPremintTransaction;
/**
 *
 * Returns a group transaction to burn gALGO for ALGO at a one-to-one rate.
 * Must be after period end.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to send and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] burn transactions
 */
function prepareBurnTransactions(dispenser, distributor, senderAddr, amount, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    const assetTransfer = {
        txn: (0, utils_1.transferAlgoOrAsset)(dispenser.gAlgoId, senderAddr, (0, algosdk_1.getApplicationAddress)(dispenser.appId), amount, Object.assign(Object.assign({}, params), { fee: 0, flatFee: true })),
        signer: utils_1.signer,
    };
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "burn"),
        methodArgs: [assetTransfer, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return (0, algosdk_1.assignGroupID)(txns);
}
exports.prepareBurnTransactions = prepareBurnTransactions;
/**
 *
 * Returns a group transaction to early claim governance rewards for a given amount of ALGO.
 * Must be after commitment end.
 * Rewards received is in gALGO. Amount of rewards is determined by rewards_per_algo.
 *
 * @param dispenser - distributor to receive gALGO from
 * @param distributor - distributor which has sender's commitment
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to early claim rewards on
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction early claim governance rewards transaction
 */
function prepareEarlyClaimGovernanceRewardsTransaction(dispenser, distributor, senderAddr, amount, params) {
    var _a;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "early_claim_rewards"),
        methodArgs: [amount, dispenser.gAlgoId, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    // for ledger compatibility (max 2 app args), remove index references which are not strictly needed
    txns[0].appArgs = (_a = txns[0].appArgs) === null || _a === void 0 ? void 0 : _a.slice(0, -2);
    return txns[0];
}
exports.prepareEarlyClaimGovernanceRewardsTransaction = prepareEarlyClaimGovernanceRewardsTransaction;
/**
 *
 * Returns a group transaction to claim governance rewards for unclaimed commitment.
 * Must be after period end and rewards have been distributed from Algorand Foundation.
 * Rewards received is in ALGO. Amount of rewards is determined by rewards_per_algo.
 *
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction claim governance rewards transaction
 */
function prepareClaimGovernanceRewardsTransaction(distributor, senderAddr, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "claim_rewards"),
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareClaimGovernanceRewardsTransaction = prepareClaimGovernanceRewardsTransaction;
//# sourceMappingURL=governance.js.map