"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareBurnTransactions = exports.prepareRemoveLiquidGovernanceEscrowTransactions = exports.prepareCommitOrVoteTransaction = exports.prepareRegisterEscrowOfflineTransaction = exports.prepareRegisterEscrowOnlineTransaction = exports.prepareClaimPremintTransaction = exports.prepareUnmintTransactions = exports.prepareUnmintPremintTransaction = exports.prepareMintTransactions = exports.prepareAddLiquidGovernanceEscrowTransactions = exports.getEscrowGovernanceStatus = exports.getUserLiquidGovernanceInfo = exports.getDistributorInfo = exports.getDistributorLogicSig = void 0;
const algosdk_1 = require("algosdk");
const utils_1 = require("../../utils");
const abiContracts_1 = require("./constants/abiContracts");
function getDistributorLogicSig(userAddr) {
    const prefix = Uint8Array.from([
        7, 32, 1, 1, 128, 36, 70, 79, 76, 75, 83, 95, 70, 73, 78, 65, 78, 67, 69, 95, 65, 76, 71, 79, 95, 76, 73, 81, 85,
        73, 68, 95, 71, 79, 86, 69, 82, 78, 65, 78, 67, 69, 72, 49, 22, 34, 9, 56, 16, 34, 18, 68, 49, 22, 34, 9, 56, 0,
        128, 32,
    ]);
    const suffix = Uint8Array.from([
        18, 68, 49, 22, 34, 9, 56, 8, 20, 68, 49, 22, 34, 9, 56, 32, 50, 3, 18, 68, 49, 22, 34, 9, 56, 9, 50, 3, 18, 68, 49,
        22, 34, 9, 56, 21, 50, 3, 18, 68, 34, 67,
    ]);
    return new algosdk_1.LogicSigAccount(new Uint8Array([...prefix, ...(0, algosdk_1.decodeAddress)(userAddr).publicKey, ...suffix]));
}
exports.getDistributorLogicSig = getDistributorLogicSig;
/**
 *
 * Returns information regarding the given liquid governance distributor.
 *
 * @param client - Algorand client to query
 * @param distributor - distributor to query about
 * @returns DistributorInfo[] distributor info
 */
function getDistributorInfo(client, distributor) {
    return __awaiter(this, void 0, void 0, function* () {
        const { appId } = distributor;
        const { currentRound, globalState: state } = yield (0, utils_1.getApplicationGlobalState)(client, appId);
        if (state === undefined)
            throw Error("Could not find Distributor");
        const dispenserAppId = Number((0, utils_1.getParsedValueFromState)(state, "dispenser_app_id") || 0);
        const premintEnd = BigInt((0, utils_1.getParsedValueFromState)(state, "premint_end") || 0);
        const commitEnd = BigInt((0, utils_1.getParsedValueFromState)(state, "commit_end") || 0);
        const periodEnd = BigInt((0, utils_1.getParsedValueFromState)(state, "period_end") || 0);
        const fee = BigInt((0, utils_1.getParsedValueFromState)(state, "fee") || 0);
        const totalCommitment = BigInt((0, utils_1.getParsedValueFromState)(state, "total_commitment") || 0);
        const isBurningPaused = Boolean((0, utils_1.getParsedValueFromState)(state, "is_burning_paused") || 0);
        return {
            currentRound,
            dispenserAppId,
            premintEnd,
            commitEnd,
            periodEnd,
            fee,
            totalCommitment,
            isBurningPaused,
        };
    });
}
exports.getDistributorInfo = getDistributorInfo;
/**
 *
 * Returns information regarding a user's liquid governance commitment.
 *
 * @param client - Algorand client to query
 * @param distributor - distributor to query about
 * @param userAddr - user address to get governance info about
 * @returns UserCommitmentInfo user commitment info
 */
function getUserLiquidGovernanceInfo(client, distributor, userAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const { appId } = distributor;
        const escrowAddr = getDistributorLogicSig(userAddr).address();
        // get user account local state
        const { currentRound, localState: state } = yield (0, utils_1.getAccountApplicationLocalState)(client, appId, escrowAddr);
        if (state === undefined)
            throw Error(`Could not find user ${userAddr} in liquid gov ${appId}`);
        // user local state
        const canDelegate = Boolean((0, utils_1.getParsedValueFromState)(state, "d"));
        const premint = BigInt((0, utils_1.getParsedValueFromState)(state, "p") || 0);
        const commitment = BigInt((0, utils_1.getParsedValueFromState)(state, "c") || 0);
        const nonCommitment = BigInt((0, utils_1.getParsedValueFromState)(state, "n") || 0);
        return {
            currentRound,
            userAddress: userAddr,
            canDelegate,
            premint,
            commitment,
            nonCommitment,
        };
    });
}
exports.getUserLiquidGovernanceInfo = getUserLiquidGovernanceInfo;
/**
 *
 * Returns information regarding a user's escrow governance status.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param userAddr - user address to get governance info about
 * @param signUpAddr - sign up address for the governance period
 * @returns EscrowGovernanceStatus escrow governance status
 */
function getEscrowGovernanceStatus(indexerClient, userAddr, signUpAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const escrowAddr = getDistributorLogicSig(userAddr).address();
        const notePrefix = "af/gov";
        const req = indexerClient
            .searchForTransactions()
            .address(escrowAddr)
            .addressRole("sender")
            .txType("pay")
            .notePrefix(utils_1.enc.encode(notePrefix));
        const [res, bal] = yield Promise.all([req.do(), (0, utils_1.getAccountDetails)(indexerClient, escrowAddr)]);
        const { currentRound, isOnline, holdings: assetHoldings } = bal;
        const algoBalance = assetHoldings.get(0);
        for (const txn of res["transactions"]) {
            const payTxn = txn["tx-type"] === "appl" ? txn["inner-txns"][0] : txn;
            const receiver = payTxn["payment-transaction"]["receiver"];
            if (receiver === signUpAddr) {
                const note = Buffer.from(payTxn["note"], "base64").toString();
                const NOTE_SPECS_REGEX = new RegExp(String.raw `^${notePrefix}(?<version>\d+):j(?<jsonData>.*)$`);
                const match = note.match(NOTE_SPECS_REGEX);
                if (!(match === null || match === void 0 ? void 0 : match.groups))
                    continue;
                const { version, jsonData } = match.groups;
                try {
                    const data = JSON.parse(jsonData);
                    const commitment = data["com"];
                    if (commitment !== undefined)
                        return {
                            currentRound,
                            balance: algoBalance,
                            isOnline,
                            status: {
                                version: Number(version),
                                commitment: BigInt(commitment),
                                beneficiaryAddress: data["bnf"],
                                xGovControlAddress: data["xGv"],
                            },
                        };
                }
                catch (e) { }
            }
        }
        return {
            currentRound,
            balance: algoBalance,
            isOnline,
        };
    });
}
exports.getEscrowGovernanceStatus = getEscrowGovernanceStatus;
/**
 *
 * Returns a group transaction to add liquid governance escrow.
 *
 * @param distributor - distributor that adding escrow to
 * @param userAddr - account address for the user
 * @param delegatable - whether governance action for escrow can be delegated to gov admin
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: LogicSigAccount } object containing group transaction and generated escrow account
 */
function prepareAddLiquidGovernanceEscrowTransactions(distributor, userAddr, delegatable, params) {
    const { appId } = distributor;
    const escrow = getDistributorLogicSig(userAddr);
    const userCall = (0, utils_1.addEscrowNoteTransaction)(userAddr, escrow.address(), appId, "ga ", Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: escrow.address(),
        signer: utils_1.signer,
        appID: appId,
        onComplete: algosdk_1.OnApplicationComplete.OptInOC,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "add_escrow"),
        methodArgs: [{ txn: userCall, signer: utils_1.signer }, delegatable],
        rekeyTo: (0, algosdk_1.getApplicationAddress)(appId),
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return { txns, escrow };
}
exports.prepareAddLiquidGovernanceEscrowTransactions = prepareAddLiquidGovernanceEscrowTransactions;
/**
 *
 * Returns a group transaction to mint gALGO for ALGO at a one-to-one rate.
 *
 * If you want to commit then MUST be paired with prepareCommitOrVoteTransaction
 * as to send the required note to the signup address.
 *
 * @param dispenser - dispenser to mint gALGO from
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to send and gALGO to mint
 * @param ensureCommit - whether to ensure in commitment period
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the minter (must pass to be eligible for revenue share)
 * @returns Transaction[] mint transactions
 */
function prepareMintTransactions(dispenser, distributor, senderAddr, amount, ensureCommit, params, note) {
    const escrowAddr = getDistributorLogicSig(senderAddr).address();
    const sendAlgo = {
        txn: (0, utils_1.transferAlgoOrAsset)(0, senderAddr, escrowAddr, amount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 })),
        signer: utils_1.signer,
    };
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "mint"),
        methodArgs: [sendAlgo, escrowAddr, dispenser.appId, dispenser.gAlgoId, ensureCommit],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
        note,
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareMintTransactions = prepareMintTransactions;
/**
 *
 * Returns a transaction to unmint pre-minted gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to unmint and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction unmint pre-mint transaction
 */
function prepareUnmintPremintTransaction(distributor, senderAddr, amount, params, note) {
    const escrowAddr = getDistributorLogicSig(senderAddr).address();
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "unmint_premint"),
        methodArgs: [escrowAddr, amount],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
        note,
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareUnmintPremintTransaction = prepareUnmintPremintTransaction;
/**
 *
 * Returns a transaction to unmint pre-minted gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to unmint and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction[] unmint transactions
 */
function prepareUnmintTransactions(dispenser, distributor, senderAddr, amount, params, note) {
    const escrowAddr = getDistributorLogicSig(senderAddr).address();
    const sendgALGO = {
        txn: (0, utils_1.transferAlgoOrAsset)(dispenser.gAlgoId, senderAddr, (0, algosdk_1.getApplicationAddress)(dispenser.appId), amount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 })),
        signer: utils_1.signer,
    };
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "unmint"),
        methodArgs: [sendgALGO, escrowAddr, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
        note,
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareUnmintTransactions = prepareUnmintTransactions;
/**
 *
 * Returns a transaction to claim pre-minted gALGO.
 * Can be called on behalf of yourself or another user.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param receiverAddr - account address for the pre-minter that will receiver the gALGO
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction claim pre-mint transaction
 */
function prepareClaimPremintTransaction(dispenser, distributor, senderAddr, receiverAddr, params) {
    const escrowAddr = getDistributorLogicSig(receiverAddr).address();
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "claim_premint"),
        methodArgs: [escrowAddr, receiverAddr, dispenser.appId, dispenser.gAlgoId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareClaimPremintTransaction = prepareClaimPremintTransaction;
/**
 *
 * Returns a transaction to register escrow online.
 *
 * @param distributor - distributor that has escrow
 * @param senderAddr - account address for the sender
 * @param voteKey - vote key
 * @param selectionKey - selection key
 * @param stateProofKey - state proof key
 * @param voteFirstRound - vote first round
 * @param voteLastRound - vote last round
 * @param voteKeyDilution - vote key dilution
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction register online transaction
 */
function prepareRegisterEscrowOnlineTransaction(distributor, senderAddr, voteKey, selectionKey, stateProofKey, voteFirstRound, voteLastRound, voteKeyDilution, params) {
    const escrowAddr = getDistributorLogicSig(senderAddr).address();
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "register_online"),
        methodArgs: [
            escrowAddr,
            (0, algosdk_1.encodeAddress)(voteKey),
            (0, algosdk_1.encodeAddress)(selectionKey),
            stateProofKey,
            voteFirstRound,
            voteLastRound,
            voteKeyDilution,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareRegisterEscrowOnlineTransaction = prepareRegisterEscrowOnlineTransaction;
/**
 *
 * Returns a transaction to register escrow offline.
 *
 * @param distributor - distributor that has escrow
 * @param senderAddr - account address for the sender
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction register offline transaction
 */
function prepareRegisterEscrowOfflineTransaction(distributor, senderAddr, params) {
    const escrowAddr = getDistributorLogicSig(senderAddr).address();
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "register_offline"),
        methodArgs: [escrowAddr],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareRegisterEscrowOfflineTransaction = prepareRegisterEscrowOfflineTransaction;
/**
 *
 * Returns a transaction to commit or vote in governance.
 *
 * @param distributor - distributor that has escrow
 * @param senderAddr - account address for the sender
 * @param destAddr - destination address
 * @param note - note to send
 * @param params - suggested params the transactions with the fees overwritten
 * @returns Transaction commit/vote transaction
 */
function prepareCommitOrVoteTransaction(distributor, senderAddr, destAddr, note, params) {
    const escrowAddr = getDistributorLogicSig(senderAddr).address();
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "governance"),
        methodArgs: [escrowAddr, destAddr, note],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareCommitOrVoteTransaction = prepareCommitOrVoteTransaction;
/**
 *
 * Returns a group transaction to remove escrow from distributor.
 * Must have zero balance or be after period end.
 *
 * @param distributor - distributor that removing escrow from
 * @param senderAddr - account address for the sender
 * @param ownerAddr - account address for the owner of the escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] burn transactions
 */
function prepareRemoveLiquidGovernanceEscrowTransactions(distributor, senderAddr, ownerAddr, params) {
    const escrowAddr = getDistributorLogicSig(ownerAddr).address();
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "remove_escrow"),
        methodArgs: [escrowAddr],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    const optOutTx = (0, algosdk_1.makeApplicationCloseOutTxn)(escrowAddr, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }), distributor.appId, undefined, undefined, undefined, undefined, undefined, undefined, escrowAddr);
    return [txns[0], optOutTx];
}
exports.prepareRemoveLiquidGovernanceEscrowTransactions = prepareRemoveLiquidGovernanceEscrowTransactions;
/**
 *
 * Returns a group transaction to burn gALGO for ALGO at a one-to-one rate.
 * Must be after period end.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to send and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] burn transactions
 */
function prepareBurnTransactions(dispenser, distributor, senderAddr, amount, params) {
    const sendgALGO = {
        txn: (0, utils_1.transferAlgoOrAsset)(dispenser.gAlgoId, senderAddr, (0, algosdk_1.getApplicationAddress)(dispenser.appId), amount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 })),
        signer: utils_1.signer,
    };
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: distributor.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.abiDistributor.methods, "burn"),
        methodArgs: [sendgALGO, dispenser.appId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareBurnTransactions = prepareBurnTransactions;
//# sourceMappingURL=governance.js.map