"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.calcLPPrice = exports.sqrt = exports.calcConversionRate = exports.calcHealthFactor = exports.calcInterestIndex = exports.calcUtilizationRatio = exports.minimum = exports.maximum = exports.calcBorrowBalance = exports.calcThreshold = exports.divVariable = void 0;
const utils_1 = require("../../utils");
const DECIMALS = BigInt(1e14);
const SECONDS_IN_YEAR = 365 * 24 * 60 * 60;
function maximum(n1, n2) {
    return n1 > n2 ? n1 : n2;
}
exports.maximum = maximum;
function minimum(n1, n2) {
    return n1 < n2 ? n1 : n2;
}
exports.minimum = minimum;
function mul14(n1, n2) {
    return (n1 * n2) / BigInt(1e14);
}
function mulVariable(n1, n2, dec) {
    return (n1 * n2) / BigInt(Math.pow(10, dec));
}
function div14(n1, n2) {
    return (n1 * BigInt(1e14)) / n2;
}
function divVariable(n1, n2, dec) {
    return (n1 * BigInt(Math.pow(10, dec))) / n2;
}
exports.divVariable = divVariable;
/**
 * Calculates the utilization ratio of the pool.
 * @param totalBorrows (Xdp)
 * @param totalDeposits (Xdp)
 * @return utilizationRate (14dp)
 */
function calcUtilizationRatio(totalBorrows, totalDeposits) {
    if (totalDeposits === BigInt(0))
        return BigInt(0);
    return div14(totalBorrows, totalDeposits);
}
exports.calcUtilizationRatio = calcUtilizationRatio;
/**
 * Calculates the current interest index. If epsilon undefined then deposit interest index, else borrow interest index.
 * @param interestIndex (14dp)
 * @param interestRate (14dp)
 * @param latestUpdate (0dp)
 * @param epsilon (14dp)
 * @return interestIndex (14dp)
 */
function calcInterestIndex(interestIndex, interestRate, latestUpdate, epsilon) {
    const dt = BigInt((0, utils_1.unixTime)()) - latestUpdate;
    const interest = (interestRate / BigInt(SECONDS_IN_YEAR)) * dt;
    return mul14(interestIndex, DECIMALS + (epsilon !== undefined ? mul14(epsilon, interest) : interest));
}
exports.calcInterestIndex = calcInterestIndex;
/**
 * Calculates the threshold of under-collaterization of the loan
 * @param collateralAmount (Xdp)
 * @param depositInterestIndex (14dp)
 * @param s2 (14dp)
 * @param conversionRate (<r_dec>dp)
 * @param conversionRateDec (0dp)
 * @return threshold (Xdp)
 */
function calcThreshold(collateralAmount, depositInterestIndex, s2, conversionRate, conversionRateDec) {
    return mulVariable(mul14(mul14(collateralAmount, depositInterestIndex), s2), conversionRate, conversionRateDec);
}
exports.calcThreshold = calcThreshold;
/**
 * Calculates the borrow balance of the loan at time t
 * @param borrowBalanceAtLastOperation (Xdp)
 * @param borrowInterestIndex (14dp)
 * @param borrowInterestIndexAtLastOperation (14dp)
 * @return borrowBalance (Xdp)
 */
function calcBorrowBalance(borrowBalanceAtLastOperation, borrowInterestIndex, borrowInterestIndexAtLastOperation) {
    return (mul14(borrowBalanceAtLastOperation, div14(borrowInterestIndex, borrowInterestIndexAtLastOperation)) + BigInt(1));
}
exports.calcBorrowBalance = calcBorrowBalance;
/**
 * Calculates the health factor of the loan at time t
 * @param threshold (Xdp)
 * @param borrowBalance (Xdp)
 * @return healthFactor (14dp)
 */
function calcHealthFactor(threshold, borrowBalance) {
    return div14(threshold, borrowBalance);
}
exports.calcHealthFactor = calcHealthFactor;
/**
 * Calculate the conversion rate between two assets
 * @param collateralPrice (Xdp)
 * @param borrowPrice (Xdp)
 * @return { rate: bigint, decimals: number } (14dp)
 */
function calcConversionRate(collateralPrice, borrowPrice) {
    let decimals = 18;
    if (collateralPrice >= borrowPrice) {
        let borrowExpPrice = borrowPrice;
        for (; borrowExpPrice < collateralPrice && decimals > 0; decimals--) {
            borrowExpPrice *= BigInt(10);
        }
    }
    const rate = divVariable(collateralPrice, borrowPrice, decimals);
    return { rate, decimals };
}
exports.calcConversionRate = calcConversionRate;
/**
 * Calculate the sqrt of a bigint (rounded down to nearest integer)
 * @param value value to be square-rooted
 * @return bigint sqrt
 */
function sqrt(value) {
    if (value < BigInt(0))
        throw Error("square root of negative numbers is not supported");
    if (value < BigInt(2))
        return value;
    function newtonIteration(n, x0) {
        const x1 = (n / x0 + x0) >> BigInt(1);
        if (x0 === x1 || x0 === x1 - BigInt(1))
            return x0;
        return newtonIteration(n, x1);
    }
    return newtonIteration(value, BigInt(1));
}
exports.sqrt = sqrt;
/**
 * Calculates the LP price
 * @param r0 pool supply of asset 0
 * @param r1 pool supply of asset 1
 * @param p0 price of asset 0
 * @param p1 price of asset 1
 * @param lts circulating supply of liquidity token
 * @return bigint LP price
 */
function calcLPPrice(r0, r1, p0, p1, lts) {
    return BigInt(2) * (sqrt(r0 * p0 * r1 * p1) / lts);
}
exports.calcLPPrice = calcLPPrice;
//# sourceMappingURL=math.js.map