"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getConversionRate = exports.getOraclePrices = void 0;
const algosdk_1 = require("algosdk");
const utils_1 = require("../../utils");
const math_1 = require("./math");
function parseOracleValue(base64Value) {
    const value = Buffer.from(base64Value, "base64").toString("hex");
    // first 8 bytes are the price
    const price = BigInt("0x" + value.slice(0, 16));
    // next 8 bytes are the timestamp
    const timestamp = BigInt("0x" + value.slice(16, 32));
    return { price, timestamp };
}
function parseOracleAdapterValue(base64Value) {
    const value = Buffer.from(base64Value, "base64").toString("hex");
    // first 8 bytes are if Tinyman
    const isTinyman = Number(`0x${value.slice(0, 16)}`);
    // next 16 bytes are asset ids
    const asset0Id = Number(`0x${value.slice(16, 32)}`);
    const asset1Id = Number(`0x${value.slice(32, 48)}`);
    // check if LP token is Tinyman or Pact
    if (isTinyman) {
        // next 32 bytes are tinyman pool address
        const poolAddress = (0, algosdk_1.encodeAddress)(Buffer.from(value.slice(48, 112), "hex"));
        return { provider: "Tinyman", asset0Id, asset1Id, poolAddress };
    }
    else {
        // next 8 bytes are pact pool app id
        const poolAppId = Number(`0x${value.slice(48, 64)}`);
        return { provider: "Pact", asset0Id, asset1Id, poolAppId };
    }
}
function getTinymanLPPrice(indexerClient, validatorAppId, poolAddress, p0, p1) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const res = yield indexerClient.lookupAccountByID(poolAddress).do();
        const { account } = res;
        const state = (_b = (_a = account["apps-local-state"]) === null || _a === void 0 ? void 0 : _a.find((app) => app.id === validatorAppId)) === null || _b === void 0 ? void 0 : _b["key-value"];
        if (state === undefined)
            throw new Error(`Unable to find Tinyman Pool: ${poolAddress} for validator app ${validatorAppId}.`);
        const r0 = BigInt((0, utils_1.getParsedValueFromState)(state, "s1") || 0);
        const r1 = BigInt((0, utils_1.getParsedValueFromState)(state, "s2") || 0);
        const lts = BigInt((0, utils_1.getParsedValueFromState)(state, "ilt") || 0);
        return (0, math_1.calcLPPrice)(r0, r1, p0, p1, lts);
    });
}
function getPactLPPrice(indexerClient, poolAppId, p0, p1) {
    return __awaiter(this, void 0, void 0, function* () {
        const res = yield indexerClient.lookupApplications(poolAppId).do();
        const state = res.application.params["global-state"];
        const r0 = BigInt((0, utils_1.getParsedValueFromState)(state, "A") || 0);
        const r1 = BigInt((0, utils_1.getParsedValueFromState)(state, "B") || 0);
        const lts = BigInt((0, utils_1.getParsedValueFromState)(state, "L") || 0);
        return (0, math_1.calcLPPrice)(r0, r1, p0, p1, lts);
    });
}
/**
 *
 * Returns oracle prices for given oracle and provided assets.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param oracle - oracle to query
 * @param assets - assets to get prices for
 * @returns OraclePrices oracle prices
 */
function getOraclePrices(indexerClient, oracle, assets) {
    return __awaiter(this, void 0, void 0, function* () {
        const { oracle1AppId, oracleAdapterAppId, tinymanValidatorAppId } = oracle;
        const [oracleRes, oracleAdapterRes] = yield Promise.all([
            indexerClient.lookupApplications(oracle1AppId).do(),
            indexerClient.lookupApplications(oracleAdapterAppId).do(),
        ]);
        const oracleState = oracleRes.application.params["global-state"];
        const oracleAdapterState = oracleAdapterRes.application.params["global-state"];
        const prices = {};
        let price;
        const promises = assets.map((assetId) => __awaiter(this, void 0, void 0, function* () {
            const base64Value = (0, utils_1.getParsedValueFromState)(oracleAdapterState, (0, utils_1.fromIntToBytes8Hex)(assetId), "hex");
            // check if liquidity token
            if (base64Value !== undefined) {
                const oracleAdapterValue = parseOracleAdapterValue(base64Value);
                const { price: p0, timestamp: t0 } = parseOracleValue(String((0, utils_1.getParsedValueFromState)(oracleState, (0, utils_1.fromIntToBytes8Hex)(oracleAdapterValue.asset0Id), "hex")));
                const { price: p1, timestamp: t1 } = parseOracleValue(String((0, utils_1.getParsedValueFromState)(oracleState, (0, utils_1.fromIntToBytes8Hex)(oracleAdapterValue.asset1Id), "hex")));
                price = {
                    price: oracleAdapterValue.provider === "Tinyman"
                        ? yield getTinymanLPPrice(indexerClient, tinymanValidatorAppId, oracleAdapterValue.poolAddress, p0, p1)
                        : yield getPactLPPrice(indexerClient, oracleAdapterValue.poolAppId, p0, p1),
                    timestamp: (0, math_1.minimum)(t0, t1),
                };
            }
            else {
                price = parseOracleValue(String((0, utils_1.getParsedValueFromState)(oracleState, (0, utils_1.fromIntToBytes8Hex)(assetId), "hex")));
            }
            prices[assetId] = price;
        }));
        yield Promise.all(promises);
        return { currentRound: oracleRes["current-round"], prices };
    });
}
exports.getOraclePrices = getOraclePrices;
/**
 *
 * Returns conversion rate between two prices.
 *
 * @param collateralPrice - collateral asset price
 * @param borrowPrice - borrow asset price
 * @returns ConversionRate conversion rate
 */
function getConversionRate(collateralPrice, borrowPrice) {
    return (0, math_1.calcConversionRate)(collateralPrice, borrowPrice);
}
exports.getConversionRate = getConversionRate;
//# sourceMappingURL=oracle.js.map