"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareRemoveDepositEscrowFromDeposits = exports.prepareOptOutDepositEscrowFromAssetInDeposits = exports.prepareUpdatePoolInterestIndexes = exports.prepareWithdrawFromPool = exports.prepareWithdrawFromDepositEscrowInDeposits = exports.prepareDepositIntoPool = exports.prepareOptDepositEscrowIntoAssetInDeposits = exports.prepareAddDepositEscrowToDeposits = exports.retrieveUserDepositInfo = exports.retrieveUserDepositsFullInfo = exports.retrieveUserDepositsInfo = exports.retrievePoolInfo = exports.retrievePoolManagerInfo = void 0;
const algosdk_1 = require("algosdk");
const mathLib_1 = require("../../mathLib");
const utils_1 = require("../../utils");
const abiContracts_1 = require("./abiContracts");
const formulae_1 = require("./formulae");
const oracle_1 = require("./oracle");
const utils_2 = require("./utils");
/**
 *
 * Returns information regarding the given pool manager.
 *
 * @param client - Algorand client to query
 * @param poolManagerAppId - pool manager application to query about
 * @returns Promise<PoolManagerInfo[]> pool manager info
 */
function retrievePoolManagerInfo(client, poolManagerAppId) {
    return __awaiter(this, void 0, void 0, function* () {
        const { currentRound, globalState: state } = yield (0, utils_1.getApplicationGlobalState)(client, poolManagerAppId);
        if (state === undefined)
            throw Error("Could not find Pool Manager");
        const pools = {};
        for (let i = 0; i < 63; i++) {
            const poolBase64Value = String((0, utils_1.getParsedValueFromState)(state, (0, utils_1.fromIntToByteHex)(i), "hex"));
            const poolValue = Buffer.from(poolBase64Value, "base64").toString("hex");
            for (let j = 0; j < 3; j++) {
                const basePos = j * 84;
                const poolAppId = Number("0x" + poolValue.slice(basePos, basePos + 12));
                // add pool
                if (poolAppId > 0) {
                    const vbir = BigInt("0x" + poolValue.slice(basePos + 12, basePos + 28));
                    const vbiit1 = BigInt("0x" + poolValue.slice(basePos + 28, basePos + 44));
                    const dir = BigInt("0x" + poolValue.slice(basePos + 44, basePos + 60));
                    const diit1 = BigInt("0x" + poolValue.slice(basePos + 60, basePos + 76));
                    const lu = BigInt("0x" + poolValue.slice(basePos + 76, basePos + 84));
                    const vbii = (0, formulae_1.calcBorrowInterestIndex)(vbir, vbiit1, lu);
                    const dii = (0, formulae_1.calcDepositInterestIndex)(dir, diit1, lu);
                    pools[poolAppId] = {
                        variableBorrowInterestRate: vbir,
                        variableBorrowInterestYield: (0, mathLib_1.compoundEverySecond)(vbir, mathLib_1.ONE_16_DP),
                        variableBorrowInterestIndex: vbii,
                        depositInterestRate: dir,
                        depositInterestYield: (0, mathLib_1.compoundEveryHour)(dir, mathLib_1.ONE_16_DP),
                        depositInterestIndex: dii,
                        metadata: {
                            oldVariableBorrowInterestIndex: vbiit1,
                            oldDepositInterestIndex: diit1,
                            oldTimestamp: lu,
                        },
                    };
                }
            }
        }
        return { currentRound, pools };
    });
}
exports.retrievePoolManagerInfo = retrievePoolManagerInfo;
/**
 *
 * Returns information regarding the given pool.
 *
 * @param client - Algorand client to query
 * @param pool - pool application to query about
 * @returns Promise<PoolInfo> pool info
 */
function retrievePoolInfo(client, pool) {
    return __awaiter(this, void 0, void 0, function* () {
        const { currentRound, globalState: state } = yield (0, utils_1.getApplicationGlobalState)(client, pool.appId);
        if (state === undefined)
            throw Error("Could not find Pool");
        const varBor = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "v")));
        const stblBor = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "s")));
        const interest = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "i")));
        const caps = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "ca")));
        const config = (0, utils_1.parseBitsAsBooleans)(String((0, utils_1.getParsedValueFromState)(state, "co")));
        // combine
        return {
            currentRound,
            variableBorrow: {
                vr0: varBor[0],
                vr1: varBor[1],
                vr2: varBor[2],
                totalVariableBorrowAmount: varBor[3],
                variableBorrowInterestRate: varBor[4],
                variableBorrowInterestYield: (0, mathLib_1.compoundEverySecond)(varBor[4], mathLib_1.ONE_16_DP),
                variableBorrowInterestIndex: (0, formulae_1.calcBorrowInterestIndex)(varBor[4], varBor[5], interest[6]),
            },
            stableBorrow: {
                sr0: stblBor[0],
                sr1: stblBor[1],
                sr2: stblBor[2],
                sr3: stblBor[3],
                optimalStableToTotalDebtRatio: stblBor[4],
                rebalanceUpUtilisationRatio: stblBor[5],
                rebalanceUpDepositInterestRate: stblBor[6],
                rebalanceDownDelta: stblBor[7],
                totalStableBorrowAmount: stblBor[8],
                stableBorrowInterestRate: stblBor[9],
                stableBorrowInterestYield: (0, mathLib_1.compoundEverySecond)(stblBor[9], mathLib_1.ONE_16_DP),
                overallStableBorrowInterestAmount: stblBor[10] * mathLib_1.UINT64 + stblBor[11],
            },
            interest: {
                retentionRate: interest[0],
                flashLoanFee: interest[1],
                optimalUtilisationRatio: interest[2],
                totalDeposits: interest[3],
                depositInterestRate: interest[4],
                depositInterestYield: (0, mathLib_1.compoundEveryHour)(interest[4], mathLib_1.ONE_16_DP),
                depositInterestIndex: (0, formulae_1.calcDepositInterestIndex)(interest[4], interest[5], interest[6]),
                latestUpdate: interest[6],
            },
            caps: {
                borrowCap: caps[0],
                stableBorrowPercentageCap: caps[1],
            },
            config: {
                depreciated: config[0],
                rewardsPaused: config[1],
                stableBorrowSupported: config[2],
                flashLoanSupported: config[3],
            },
        };
    });
}
exports.retrievePoolInfo = retrievePoolInfo;
/**
 *
 * Returns basic information regarding the given user's deposit escrows.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param depositsAppId - deposits application to query about
 * @param userAddr - account address for the user
 * @returns Promise<UserDepositInfo[]> user deposits info
 */
function retrieveUserDepositsInfo(indexerClient, depositsAppId, userAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const userDepositsInfo = [];
        // get users' escrows
        const escrows = yield (0, utils_2.getEscrows)(indexerClient, userAddr, depositsAppId, "da ", "dr ");
        // get all remaining escrows' holdings
        for (const escrowAddr of escrows) {
            const { currentRound, holdings: assetHoldings } = yield (0, utils_1.getAccountDetails)(indexerClient, escrowAddr);
            const holdings = [];
            assetHoldings.forEach((balance, assetId) => {
                if (assetId !== 0)
                    holdings.push({ fAssetId: assetId, fAssetBalance: balance });
            });
            userDepositsInfo.push({ currentRound, escrowAddress: escrowAddr, holdings });
        }
        return userDepositsInfo;
    });
}
exports.retrieveUserDepositsInfo = retrieveUserDepositsInfo;
/**
 *
 * Returns full information regarding the given user's deposit escrows.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param poolManagerAppId - pool manager application to query about
 * @param depositsAppId - deposits application to query about
 * @param pools - pools in pool manager (either MainnetPools or TestnetPools)
 * @param oracle - oracle to query
 * @param userAddr - account address for the user
 * @returns Promise<UserDepositFullInfo[]> user deposits full info
 */
function retrieveUserDepositsFullInfo(indexerClient, poolManagerAppId, depositsAppId, pools, oracle, userAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        // get all prerequisites
        const userDepositsInfoReq = retrieveUserDepositsInfo(indexerClient, depositsAppId, userAddr);
        const poolManagerInfoReq = retrievePoolManagerInfo(indexerClient, poolManagerAppId);
        const oraclePricesReq = (0, oracle_1.getOraclePrices)(indexerClient, oracle);
        const [userDepositsInfo, poolManagerInfo, { prices }] = yield Promise.all([
            userDepositsInfoReq,
            poolManagerInfoReq,
            oraclePricesReq,
        ]);
        // map from UserDepositInfo to ExtendedUserDepositInfo
        return userDepositsInfo.map((deposit) => {
            const holdings = deposit.holdings.map(({ fAssetId, fAssetBalance }) => {
                const pool = Object.entries(pools)
                    .map(([, pool]) => pool)
                    .find((pool) => pool.fAssetId === fAssetId);
                if (pool === undefined)
                    throw Error("Could not find pool with fAsset " + fAssetId);
                const poolAppId = pool.appId;
                const assetId = pool.assetId;
                const poolInfo = poolManagerInfo.pools[poolAppId];
                if (poolInfo === undefined)
                    throw Error("Could not find pool " + poolAppId);
                const { depositInterestIndex, depositInterestRate, depositInterestYield } = poolInfo;
                const oraclePrice = prices[assetId];
                if (oraclePrice === undefined)
                    throw Error("Could not find asset price " + assetId);
                const { price: assetPrice } = oraclePrice;
                const assetBalance = (0, formulae_1.calcWithdrawReturn)(fAssetBalance, depositInterestIndex);
                const balanceValue = (0, mathLib_1.mulScale)(assetBalance, assetPrice, mathLib_1.ONE_10_DP); // 4 d.p.
                return {
                    fAssetId,
                    fAssetBalance,
                    poolAppId,
                    assetId,
                    assetPrice,
                    assetBalance,
                    balanceValue,
                    interestRate: depositInterestRate,
                    interestYield: depositInterestYield,
                };
            });
            return Object.assign(Object.assign({}, deposit), { holdings });
        });
    });
}
exports.retrieveUserDepositsFullInfo = retrieveUserDepositsFullInfo;
/**
 *
 * Returns information regarding the given user's deposit escrows.
 *
 * @param client - Algorand client to query
 * @param depositsAppId - deposits application to query about
 * @param escrowAddr - account address for the deposit escrow
 * @returns Promise<UserDepositInfo> user deposit info
 */
function retrieveUserDepositInfo(client, depositsAppId, escrowAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const { currentRound, holdings: assetHoldings } = yield (0, utils_1.getAccountDetails)(client, escrowAddr);
        const holdings = [];
        assetHoldings.forEach((balance, assetId) => {
            if (assetId !== 0)
                holdings.push({ fAssetId: assetId, fAssetBalance: balance });
        });
        return { currentRound, escrowAddress: escrowAddr, holdings };
    });
}
exports.retrieveUserDepositInfo = retrieveUserDepositInfo;
/**
 *
 * Returns a group transaction to add escrow before depositing.
 *
 * @param depositsAppId - deposits application to add an escrow for
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: Account } object containing group transaction and generated escrow account
 */
function prepareAddDepositEscrowToDeposits(depositsAppId, userAddr, params) {
    const escrow = (0, algosdk_1.generateAccount)();
    const userCall = (0, utils_1.addEscrowNoteTransaction)(userAddr, escrow.addr, depositsAppId, "da ", Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: escrow.addr,
        signer: utils_1.signer,
        appID: depositsAppId,
        onComplete: algosdk_1.OnApplicationComplete.OptInOC,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositsABIContract.methods, "add_deposit_escrow"),
        methodArgs: [{ txn: userCall, signer: utils_1.signer }],
        rekeyTo: (0, algosdk_1.getApplicationAddress)(depositsAppId),
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return { txns, escrow };
}
exports.prepareAddDepositEscrowToDeposits = prepareAddDepositEscrowToDeposits;
/**
 *
 * Returns a transaction to opt deposits escrow into asset so that it can hold a given pool's f asset.
 *
 * @param depositsAppId - deposits application of escrow
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param pool - pool to add f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt deposit escrow into asset transaction
 */
function prepareOptDepositEscrowIntoAssetInDeposits(depositsAppId, poolManagerAppId, userAddr, escrowAddr, pool, params) {
    const { appId: poolAppId, fAssetId, poolManagerIndex } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositsAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositsABIContract.methods, "opt_escrow_into_asset"),
        methodArgs: [escrowAddr, poolManagerAppId, poolAppId, fAssetId, poolManagerIndex],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareOptDepositEscrowIntoAssetInDeposits = prepareOptDepositEscrowIntoAssetInDeposits;
/**
 *
 * Returns a transaction to deposit asset into given pool.
 *
 * @param pool - pool application to deposit into
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the deposit (typically the user's deposit escrow or loan escrow)
 * @param assetAmount - the asset amount to deposit
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note field to add
 * @returns Transaction[] deposit asset group transaction
 */
function prepareDepositIntoPool(pool, poolManagerAppId, userAddr, receiverAddr, assetAmount, params, note) {
    const { appId, assetId, fAssetId } = pool;
    const sendAsset = (0, utils_1.transferAlgoOrAsset)(assetId, userAddr, (0, algosdk_1.getApplicationAddress)(appId), assetAmount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.poolABIContract.methods, "deposit"),
        methodArgs: [{ txn: sendAsset, signer: utils_1.signer }, receiverAddr, assetId, fAssetId, poolManagerAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
        note,
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareDepositIntoPool = prepareDepositIntoPool;
/**
 *
 * Returns a transaction to withdraw from a deposits escrow
 *
 * @param depositsAppId - deposits application of escrow
 * @param pool - pool to withdraw from
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param amount - the amount of asset / f asset to send to withdraw from escrow.
 * @param isfAssetAmount - whether the amount to withdraw is expressed in terms of f asset or asset
 * @param remainDeposited - whether receiver should get f asset or asset (cannot remain deposited and use asset amount)
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction withdraw from deposit escrow transaction
 */
function prepareWithdrawFromDepositEscrowInDeposits(depositsAppId, pool, poolManagerAppId, userAddr, escrowAddr, receiverAddr, amount, isfAssetAmount, remainDeposited, params) {
    const { appId: poolAppId, assetId, fAssetId, poolManagerIndex } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositsAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositsABIContract.methods, "withdraw"),
        methodArgs: [
            escrowAddr,
            receiverAddr,
            poolManagerAppId,
            poolAppId,
            assetId,
            fAssetId,
            amount,
            isfAssetAmount,
            remainDeposited,
            poolManagerIndex,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: remainDeposited ? 2000 : 6000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareWithdrawFromDepositEscrowInDeposits = prepareWithdrawFromDepositEscrowInDeposits;
/**
 *
 * Returns a group transaction to withdraw from a user's wallet
 *
 * @param pool - pool to withdraw from
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param fAssetAmount - the amount of f asset to send to the pool
 * @param receivedAssetAmount - the amount of asset to receive. Any excess f asset sent will be returned to the deposit escrow. If zero then interpreted as variable withdrawal.
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] withdraw from user's wallet group transaction
 */
function prepareWithdrawFromPool(pool, poolManagerAppId, userAddr, receiverAddr, fAssetAmount, receivedAssetAmount, params) {
    const { appId: poolAppId, assetId, fAssetId } = pool;
    const sendfAsset = (0, utils_1.transferAlgoOrAsset)(fAssetId, userAddr, (0, algosdk_1.getApplicationAddress)(poolAppId), fAssetAmount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: poolAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.poolABIContract.methods, "withdraw"),
        methodArgs: [{ txn: sendfAsset, signer: utils_1.signer }, receivedAssetAmount, receiverAddr, assetId, fAssetId, poolManagerAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 5000 }),
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareWithdrawFromPool = prepareWithdrawFromPool;
/**
 *
 * Returns a transaction to update a pool's interest indexes
 *
 * @param pool - pool to update
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] update pool's interest indexes transaction
 */
function prepareUpdatePoolInterestIndexes(pool, poolManagerAppId, userAddr, params) {
    const { appId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.poolABIContract.methods, "update_pool_interest_indexes"),
        methodArgs: [poolManagerAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareUpdatePoolInterestIndexes = prepareUpdatePoolInterestIndexes;
/**
 *
 * Returns a transaction to remove asset from deposit escrow
 *
 * @param depositsAppId - deposits application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param pool - pool to remove f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt out deposit escrow from asset transaction
 */
function prepareOptOutDepositEscrowFromAssetInDeposits(depositsAppId, userAddr, escrowAddr, pool, params) {
    const { appId: poolAppId, fAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositsAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositsABIContract.methods, "close_out_escrow_from_asset"),
        methodArgs: [escrowAddr, (0, algosdk_1.getApplicationAddress)(poolAppId), fAssetId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareOptOutDepositEscrowFromAssetInDeposits = prepareOptOutDepositEscrowFromAssetInDeposits;
/**
 *
 * Returns a group transaction to remove a user's deposit escrow and return its minimum balance to the user.
 *
 * @param depositsAppId - deposits application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] remove and close out deposit escrow group transaction
 */
function prepareRemoveDepositEscrowFromDeposits(depositsAppId, userAddr, escrowAddr, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositsAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositsABIContract.methods, "remove_deposit_escrow"),
        methodArgs: [escrowAddr],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    const optOutTx = (0, algosdk_1.makeApplicationCloseOutTxn)(escrowAddr, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }), depositsAppId);
    const closeToTx = (0, utils_1.removeEscrowNoteTransaction)(escrowAddr, userAddr, "dr ", Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }));
    return [txns[0], optOutTx, closeToTx];
}
exports.prepareRemoveDepositEscrowFromDeposits = prepareRemoveDepositEscrowFromDeposits;
//# sourceMappingURL=deposit.js.map