"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareRemoveDepositStakingEscrow = exports.prepareOptOutDepositStakingEscrowFromAsset = exports.prepareWithdrawFromDepositStakingEscrow = exports.prepareClaimRewardsOfDepositStakingEscrow = exports.prepareSyncStakeInDepositStakingEscrow = exports.prepareOptDepositStakingEscrowIntoAsset = exports.prepareAddDepositStakingEscrow = exports.userDepositStakingInfo = exports.retrieveUserDepositStakingLocalState = exports.retrieveUserDepositStakingsLocalState = exports.depositStakingProgramsInfo = exports.retrieveDepositStakingInfo = void 0;
const algosdk_1 = require("algosdk");
const mathLib_1 = require("../../mathLib");
const utils_1 = require("../../utils");
const abiContracts_1 = require("./abiContracts");
const utils_2 = require("./utils");
Object.defineProperty(exports, "depositStakingProgramsInfo", { enumerable: true, get: function () { return utils_2.depositStakingProgramsInfo; } });
Object.defineProperty(exports, "userDepositStakingInfo", { enumerable: true, get: function () { return utils_2.userDepositStakingInfo; } });
/**
 *
 * Returns information regarding the given deposit staking application.
 *
 * @param client - Algorand client to query
 * @param depositStakingAppId - deposit staking application to query about
 * @returns Promise<DepositStakingInfo> pool info
 */
function retrieveDepositStakingInfo(client, depositStakingAppId) {
    return __awaiter(this, void 0, void 0, function* () {
        const { currentRound, globalState: state } = yield (0, utils_1.getApplicationGlobalState)(client, depositStakingAppId);
        if (state === undefined)
            throw Error("Could not find Deposit Staking");
        // initialise staking program
        const stakingPrograms = [];
        for (let i = 0; i <= 5; i++) {
            const prefix = "S".charCodeAt(0).toString(16);
            const stakeBase64Value = String((0, utils_1.getParsedValueFromState)(state, prefix + (0, utils_1.fromIntToByteHex)(i), "hex"));
            const stakeValue = Buffer.from(stakeBase64Value, "base64").toString("hex");
            for (let j = 0; j <= 4; j++) {
                const basePos = j * 46;
                const rewards = [];
                stakingPrograms.push({
                    poolAppId: Number("0x" + stakeValue.slice(basePos, basePos + 12)),
                    totalStaked: BigInt("0x" + stakeValue.slice(basePos + 12, basePos + 28)),
                    minTotalStaked: BigInt("0x" + stakeValue.slice(basePos + 28, basePos + 44)),
                    stakeIndex: i * 5 + j,
                    numRewards: Number("0x" + stakeValue.slice(basePos + 44, basePos + 46)),
                    rewards,
                });
            }
        }
        // add rewards
        for (let i = 0; i <= 22; i++) {
            const prefix = "R".charCodeAt(0).toString(16);
            const rewardBase64Value = String((0, utils_1.getParsedValueFromState)(state, prefix + (0, utils_1.fromIntToByteHex)(i), "hex"));
            const rewardValue = Buffer.from(rewardBase64Value, "base64").toString("hex");
            for (let j = 0; j <= (i !== 22 ? 3 : 1); j++) {
                const basePos = j * 60;
                const stakeIndex = Number(BigInt(i * 4 + j) / BigInt(3));
                const localRewardIndex = Number(BigInt(i * 4 + j) % BigInt(3));
                const { totalStaked, minTotalStaked, rewards, numRewards } = stakingPrograms[stakeIndex];
                if (localRewardIndex >= numRewards)
                    continue;
                const ts = (0, mathLib_1.maximum)(totalStaked, minTotalStaked);
                const endTimestamp = BigInt("0x" + rewardValue.slice(basePos + 12, basePos + 20));
                const lu = BigInt("0x" + rewardValue.slice(basePos + 20, basePos + 28));
                const rewardRate = BigInt("0x" + rewardValue.slice(basePos + 28, basePos + 44));
                const rpt = BigInt("0x" + rewardValue.slice(basePos + 44, basePos + 60));
                const currTime = BigInt((0, utils_1.unixTime)());
                const dt = currTime <= endTimestamp ? currTime - lu : lu <= endTimestamp ? endTimestamp - lu : BigInt(0);
                const rewardPerToken = rpt + (rewardRate * dt) / ts;
                rewards.push({
                    rewardAssetId: Number("0x" + rewardValue.slice(basePos, basePos + 12)),
                    endTimestamp,
                    rewardRate,
                    rewardPerToken,
                });
            }
        }
        // combine
        return { currentRound, stakingPrograms };
    });
}
exports.retrieveDepositStakingInfo = retrieveDepositStakingInfo;
/**
 *
 * Returns local state regarding the deposit staking escrows of a given user.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param depositStakingAppId - deposit staking application to query about
 * @param userAddr - account address for the user
 * @returns Promise<UserDepositStakingLocalState[]> deposit staking escrows' local state
 */
function retrieveUserDepositStakingsLocalState(indexerClient, depositStakingAppId, userAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const depositStakingsLocalState = [];
        const escrows = yield (0, utils_2.getEscrows)(indexerClient, userAddr, depositStakingAppId, "fa ", "fr ");
        // get all remaining deposit stakings' local state
        for (const escrowAddr of escrows) {
            const [{ holdings }, { currentRound, localState: state }] = yield Promise.all([
                (0, utils_1.getAccountDetails)(indexerClient, escrowAddr),
                (0, utils_1.getAccountApplicationLocalState)(indexerClient, depositStakingAppId, escrowAddr),
            ]);
            const optedIntoAssets = new Set(holdings.keys());
            if (state === undefined)
                throw Error(`Could not find deposit staking ${depositStakingAppId} in escrow ${escrowAddr}`);
            depositStakingsLocalState.push(Object.assign(Object.assign({}, (0, utils_2.depositStakingLocalState)(state, depositStakingAppId, escrowAddr)), { currentRound,
                optedIntoAssets }));
        }
        return depositStakingsLocalState;
    });
}
exports.retrieveUserDepositStakingsLocalState = retrieveUserDepositStakingsLocalState;
/**
 *
 * Returns local state regarding the deposit staking escrows of a given user.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param depositStakingAppId - deposit staking application to query about
 * @param escrowAddr - account address for the deposit staking escrow
 * @returns Promise<UserDepositStakingLocalState> deposit staking escrows' local state
 */
function retrieveUserDepositStakingLocalState(indexerClient, depositStakingAppId, escrowAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const { currentRound, localState: state } = yield (0, utils_1.getAccountApplicationLocalState)(indexerClient, depositStakingAppId, escrowAddr);
        if (state === undefined)
            throw Error(`Could not find deposit staking ${depositStakingAppId} in escrow ${escrowAddr}`);
        return Object.assign({ currentRound }, (0, utils_2.depositStakingLocalState)(state, depositStakingAppId, escrowAddr));
    });
}
exports.retrieveUserDepositStakingLocalState = retrieveUserDepositStakingLocalState;
/**
 *
 * Returns a group transaction to deposit staking escrow.
 *
 * @param depositStakingAppId - deposit staking application to query about
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: Account } object containing group transaction and generated escrow account
 */
function prepareAddDepositStakingEscrow(depositStakingAppId, userAddr, params) {
    const escrow = (0, algosdk_1.generateAccount)();
    const userCall = (0, utils_1.addEscrowNoteTransaction)(userAddr, escrow.addr, depositStakingAppId, "fa ", Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: escrow.addr,
        signer: utils_1.signer,
        appID: depositStakingAppId,
        onComplete: algosdk_1.OnApplicationComplete.OptInOC,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositStakingABIContract.methods, "add_f_staking_escrow"),
        methodArgs: [{ txn: userCall, signer: utils_1.signer }],
        rekeyTo: (0, algosdk_1.getApplicationAddress)(depositStakingAppId),
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return { txns, escrow };
}
exports.prepareAddDepositStakingEscrow = prepareAddDepositStakingEscrow;
/**
 *
 * Returns a transaction to opt deposit staking escrow into asset so that it can hold a given pool's f asset.
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param pool - pool to add f asset of
 * @param stakeIndex - staking program index
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt deposit staking escrow into asset transaction
 */
function prepareOptDepositStakingEscrowIntoAsset(depositStakingAppId, poolManagerAppId, userAddr, escrowAddr, pool, stakeIndex, params) {
    const { appId: poolAppId, fAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositStakingAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositStakingABIContract.methods, "opt_escrow_into_asset"),
        methodArgs: [escrowAddr, poolAppId, fAssetId, stakeIndex],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareOptDepositStakingEscrowIntoAsset = prepareOptDepositStakingEscrowIntoAsset;
/**
 *
 * Returns a transaction to sync stake of deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param pool - pool to sync stake of
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param stakeIndex - staking program index
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction sync stake of deposit staking escrow transaction
 */
function prepareSyncStakeInDepositStakingEscrow(depositStakingAppId, pool, userAddr, escrowAddr, stakeIndex, params) {
    const { appId: poolAppId, fAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositStakingAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositStakingABIContract.methods, "sync_stake"),
        methodArgs: [escrowAddr, poolAppId, fAssetId, stakeIndex],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareSyncStakeInDepositStakingEscrow = prepareSyncStakeInDepositStakingEscrow;
/**
 *
 * Returns a transaction to claim the rewards of a deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param stakeIndex - staking program index
 * @param rewardAssetIds - the asset ids of all the rewards assets claiming
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction withdraw from deposit staking escrow transaction
 */
function prepareClaimRewardsOfDepositStakingEscrow(depositStakingAppId, userAddr, escrowAddr, receiverAddr, stakeIndex, rewardAssetIds, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositStakingAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositStakingABIContract.methods, "claim_rewards"),
        methodArgs: [escrowAddr, receiverAddr, stakeIndex],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    txns[0].appForeignAssets = rewardAssetIds;
    return txns[0];
}
exports.prepareClaimRewardsOfDepositStakingEscrow = prepareClaimRewardsOfDepositStakingEscrow;
/**
 *
 * Returns a transaction to withdraw from a deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param pool - pool to withdraw from
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param amount - the amount of asset / f asset to send to withdraw from escrow.
 * @param isfAssetAmount - whether the amount to withdraw is expressed in terms of f asset or asset
 * @param remainDeposited - whether receiver should get f asset or asset (cannot remain deposited and use asset amount)
 * @param stakeIndex - staking program index
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction withdraw from deposit staking escrow transaction
 */
function prepareWithdrawFromDepositStakingEscrow(depositStakingAppId, pool, poolManagerAppId, userAddr, escrowAddr, receiverAddr, amount, isfAssetAmount, remainDeposited, stakeIndex, params) {
    const { appId: poolAppId, assetId, fAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositStakingAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositStakingABIContract.methods, "withdraw_stake"),
        methodArgs: [
            escrowAddr,
            receiverAddr,
            poolManagerAppId,
            poolAppId,
            assetId,
            fAssetId,
            amount,
            isfAssetAmount,
            remainDeposited,
            stakeIndex,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 6000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareWithdrawFromDepositStakingEscrow = prepareWithdrawFromDepositStakingEscrow;
/**
 *
 * Returns a transaction to remove asset from deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param pool - pool to remove f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt out deposit staking escrow from asset transaction
 */
function prepareOptOutDepositStakingEscrowFromAsset(depositStakingAppId, userAddr, escrowAddr, pool, params) {
    const { appId: poolAppId, fAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositStakingAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositStakingABIContract.methods, "close_out_escrow_from_asset"),
        methodArgs: [escrowAddr, (0, algosdk_1.getApplicationAddress)(poolAppId), fAssetId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareOptOutDepositStakingEscrowFromAsset = prepareOptOutDepositStakingEscrowFromAsset;
/**
 *
 * Returns a group transaction to remove a user's deposit staking escrow and return its minimum balance to the user.
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] remove and close out deposit staking escrow group transaction
 */
function prepareRemoveDepositStakingEscrow(depositStakingAppId, userAddr, escrowAddr, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: depositStakingAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.depositStakingABIContract.methods, "remove_f_staking_escrow"),
        methodArgs: [escrowAddr],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    const optOutTx = (0, algosdk_1.makeApplicationCloseOutTxn)(escrowAddr, params, depositStakingAppId);
    const closeToTx = (0, utils_1.removeEscrowNoteTransaction)(escrowAddr, userAddr, "fr ", params);
    return [txns[0], optOutTx, closeToTx];
}
exports.prepareRemoveDepositStakingEscrow = prepareRemoveDepositStakingEscrow;
//# sourceMappingURL=depositStaking.js.map