"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.calcLPPrice = exports.calcFlashLoanRepayment = exports.calcRebalanceDownThreshold = exports.calcRebalanceUpThreshold = exports.calcLoanStableInterestRate = exports.calcBorrowBalance = exports.calcLiquidationMargin = exports.calcBorrowUtilisationRatio = exports.calcLTVRatio = exports.calcBorrowAssetLoanValue = exports.calcCollateralAssetLoanValue = exports.calcWithdrawReturn = exports.calcDepositReturn = exports.calcDepositInterestIndex = exports.calcBorrowInterestIndex = exports.calcDepositInterestRate = exports.calcOverallBorrowInterestRate = exports.calcStableBorrowInterestRate = exports.calcVariableBorrowInterestRate = exports.calcStableDebtToTotalDebtRatio = exports.calcUtilisationRatio = exports.calcMaxSingleStableBorrow = exports.calcStableBorrowRatio = exports.calcAvailableLiquidity = exports.calcTotalDebt = exports.calcAssetDollarValue = void 0;
const mathLib_1 = require("../../mathLib");
const utils_1 = require("../../utils");
/**
 * Calculates the dollar value of a given asset amount
 * @param amount (0dp)
 * @param price (14dp)
 * @return value (0dp)
 */
function calcAssetDollarValue(amount, price) {
    return (0, mathLib_1.mulScaleRoundUp)(amount, price, mathLib_1.ONE_14_DP);
}
exports.calcAssetDollarValue = calcAssetDollarValue;
/**
 * Calculates the total debt of a pool
 * @param totalVarDebt (0dp)
 * @param totalStblDebt (0dp)
 * @return totalDebt (0dp)
 */
function calcTotalDebt(totalVarDebt, totalStblDebt) {
    return totalVarDebt + totalStblDebt;
}
exports.calcTotalDebt = calcTotalDebt;
/**
 * Calculates the total debt of a pool
 * @param totalDebt (0dp)
 * @param totalDeposits (0dp)
 * @return availableLiquidity (0dp)
 */
function calcAvailableLiquidity(totalDebt, totalDeposits) {
    return totalDeposits - totalDebt;
}
exports.calcAvailableLiquidity = calcAvailableLiquidity;
/**
 * Calculates the ratio of the available liquidity that is being stable borrowed
 * @param stblBorAmount (0dp)
 * @param availableLiquidity (0dp)
 * @return stableBorrowRatio (16dp)
 */
function calcStableBorrowRatio(stblBorAmount, availableLiquidity) {
    return (0, mathLib_1.divScale)(stblBorAmount, availableLiquidity, mathLib_1.ONE_16_DP);
}
exports.calcStableBorrowRatio = calcStableBorrowRatio;
/**
 * Calculates the maximum stable borrow amount a user can make in one go
 * @param availableLiquidity (0dp)
 * @param sbpc (0dp)
 * @return stableBorrowRatio (16dp)
 */
function calcMaxSingleStableBorrow(availableLiquidity, sbpc) {
    return (0, mathLib_1.mulScale)(availableLiquidity, sbpc, mathLib_1.ONE_16_DP);
}
exports.calcMaxSingleStableBorrow = calcMaxSingleStableBorrow;
/**
 * Calculates the utilisation ratio of a pool
 * @param totalDebt (0dp)
 * @param totalDeposits (0dp)
 * @return utilisationRatio (16dp)
 */
function calcUtilisationRatio(totalDebt, totalDeposits) {
    if (totalDeposits === BigInt(0))
        return BigInt(0);
    return (0, mathLib_1.divScale)(totalDebt, totalDeposits, mathLib_1.ONE_16_DP);
}
exports.calcUtilisationRatio = calcUtilisationRatio;
/**
 * Calculates the stable debt to total debt ratio of a pool
 * @param totalStblDebt (0dp)
 * @param totalDebt (0dp)
 * @return stableDebtToTotalDebtRatio (16dp)
 */
function calcStableDebtToTotalDebtRatio(totalStblDebt, totalDebt) {
    if (totalDebt === BigInt(0))
        return BigInt(0);
    return (0, mathLib_1.divScale)(totalStblDebt, totalDebt, mathLib_1.ONE_16_DP);
}
exports.calcStableDebtToTotalDebtRatio = calcStableDebtToTotalDebtRatio;
/**
 * Calculate the variable borrow interest rate of a pool
 * @param vr0 (16dp)
 * @param vr1 (16dp)
 * @param vr2 (16dp)
 * @param ut (16dp)
 * @param uopt (16dp)
 * @return variableBorrowInterestRate (16dp)
 */
function calcVariableBorrowInterestRate(vr0, vr1, vr2, ut, uopt) {
    return ut < uopt
        ? vr0 + (0, mathLib_1.divScale)((0, mathLib_1.mulScale)(ut, vr1, mathLib_1.ONE_16_DP), uopt, mathLib_1.ONE_16_DP)
        : vr0 + vr1 + (0, mathLib_1.divScale)((0, mathLib_1.mulScale)(ut - uopt, vr2, mathLib_1.ONE_16_DP), mathLib_1.ONE_16_DP - uopt, mathLib_1.ONE_16_DP);
}
exports.calcVariableBorrowInterestRate = calcVariableBorrowInterestRate;
/**
 * Calculate the stable borrow interest rate of a pool
 * @param vr1 (16dp)
 * @param sr0 (16dp)
 * @param sr1 (16dp)
 * @param sr2 (16dp)
 * @param sr3 (16dp)
 * @param ut (16dp)
 * @param uopt (16dp)
 * @param ratiot (16dp)
 * @param ratioopt (16dp)
 * @return stableBorrowInterestRate (16dp)
 */
function calcStableBorrowInterestRate(vr1, sr0, sr1, sr2, sr3, ut, uopt, ratiot, ratioopt) {
    const base = ut <= uopt
        ? vr1 + sr0 + (0, mathLib_1.divScale)((0, mathLib_1.mulScale)(ut, sr1, mathLib_1.ONE_16_DP), uopt, mathLib_1.ONE_16_DP)
        : vr1 + sr0 + sr1 + (0, mathLib_1.divScale)((0, mathLib_1.mulScale)(ut - uopt, sr2, mathLib_1.ONE_16_DP), mathLib_1.ONE_16_DP - uopt, mathLib_1.ONE_16_DP);
    const extra = ratiot <= ratioopt
        ? BigInt(0)
        : (0, mathLib_1.divScale)((0, mathLib_1.mulScale)(sr3, ratiot - ratioopt, mathLib_1.ONE_16_DP), mathLib_1.ONE_16_DP - ratioopt, mathLib_1.ONE_16_DP);
    return base + extra;
}
exports.calcStableBorrowInterestRate = calcStableBorrowInterestRate;
/**
 * Calculate the overall borrow interest rate of a pool
 * @param totalVarDebt (0dp)
 * @param totalDebt (0dp)
 * @param vbirt (16dp)
 * @param osbiat (16dp)
 * @return overallBorrowInterestRate (16dp)
 */
function calcOverallBorrowInterestRate(totalVarDebt, totalDebt, vbirt, osbiat) {
    if (totalDebt === BigInt(0))
        return BigInt(0);
    return (totalVarDebt * vbirt + osbiat) / totalDebt;
}
exports.calcOverallBorrowInterestRate = calcOverallBorrowInterestRate;
/**
 * Calculate the deposit interest rate of a pool
 * @param obirt (16dp)
 * @param ut (16dp)
 * @param rr (16dp)
 * @return overallBorrowInterestRate (16dp)
 */
function calcDepositInterestRate(obirt, rr, ut) {
    return (0, mathLib_1.mulScale)((0, mathLib_1.mulScale)(ut, obirt, mathLib_1.ONE_16_DP), mathLib_1.ONE_16_DP - rr, mathLib_1.ONE_16_DP);
}
exports.calcDepositInterestRate = calcDepositInterestRate;
/**
 * Calculate the borrow interest index of pool
 * @param birt1 (16dp)
 * @param biit1 (16dp)
 * @param latestUpdate (0dp)
 * @return borrowInterestIndex (14dp)
 */
function calcBorrowInterestIndex(birt1, biit1, latestUpdate) {
    const dt = BigInt((0, utils_1.unixTime)()) - latestUpdate;
    return (0, mathLib_1.mulScale)(biit1, (0, mathLib_1.expBySquaring)(mathLib_1.ONE_16_DP + birt1 / mathLib_1.SECONDS_IN_YEAR, dt, mathLib_1.ONE_16_DP), mathLib_1.ONE_16_DP);
}
exports.calcBorrowInterestIndex = calcBorrowInterestIndex;
/**
 * Calculate the deposit interest index of pool
 * @param dirt1 (16dp)
 * @param diit1 (16dp)
 * @param latestUpdate (0dp)
 * @return depositInterestIndex (14dp)
 */
function calcDepositInterestIndex(dirt1, diit1, latestUpdate) {
    const dt = BigInt((0, utils_1.unixTime)()) - latestUpdate;
    return (0, mathLib_1.mulScale)(diit1, mathLib_1.ONE_16_DP + (dirt1 * dt) / mathLib_1.SECONDS_IN_YEAR, mathLib_1.ONE_16_DP);
}
exports.calcDepositInterestIndex = calcDepositInterestIndex;
/**
 * Calculates the fAsset received from a deposit
 * @param depositAmount (0dp)
 * @param diit (14dp)
 * @return depositReturn (0dp)
 */
function calcDepositReturn(depositAmount, diit) {
    return (0, mathLib_1.divScale)(depositAmount, diit, mathLib_1.ONE_14_DP);
}
exports.calcDepositReturn = calcDepositReturn;
/**
 * Calculates the asset received from a withdraw
 * @param withdrawAmount (0dp)
 * @param diit (14dp)
 * @return withdrawReturn (0dp)
 */
function calcWithdrawReturn(withdrawAmount, diit) {
    return (0, mathLib_1.mulScale)(withdrawAmount, diit, mathLib_1.ONE_14_DP);
}
exports.calcWithdrawReturn = calcWithdrawReturn;
/**
 * Calculates the collateral asset loan value
 * @param amount (0dp)
 * @param price (14dp)
 * @param factor (4dp)
 * @return loanValue (4dp)
 */
function calcCollateralAssetLoanValue(amount, price, factor) {
    return (0, mathLib_1.mulScale)((0, mathLib_1.mulScale)(amount, price, mathLib_1.ONE_10_DP), factor, mathLib_1.ONE_4_DP);
}
exports.calcCollateralAssetLoanValue = calcCollateralAssetLoanValue;
/**
 * Calculates the borrow asset loan value
 * @param amount (0dp)
 * @param price (14dp)
 * @param factor (4dp)
 * @return loanValue (4dp)
 */
function calcBorrowAssetLoanValue(amount, price, factor) {
    return (0, mathLib_1.mulScaleRoundUp)((0, mathLib_1.mulScaleRoundUp)(amount, price, mathLib_1.ONE_10_DP), factor, mathLib_1.ONE_4_DP);
}
exports.calcBorrowAssetLoanValue = calcBorrowAssetLoanValue;
/**
 * Calculates the loan's LTV ratio
 * @param totalBorrowBalanceValue (4dp)
 * @param totalCollateralBalanceValue (4dp)
 * @return LTVRatio (4dp)
 */
function calcLTVRatio(totalBorrowBalanceValue, totalCollateralBalanceValue) {
    if (totalCollateralBalanceValue === BigInt(0))
        return BigInt(0);
    return (0, mathLib_1.divScale)(totalBorrowBalanceValue, totalCollateralBalanceValue, mathLib_1.ONE_4_DP);
}
exports.calcLTVRatio = calcLTVRatio;
/**
 * Calculates the loan's borrow utilisation ratio
 * @param totalEffectiveBorrowBalanceValue (4dp)
 * @param totalEffectiveCollateralBalanceValue (4dp)
 * @return borrowUtilisationRatio (4dp)
 */
function calcBorrowUtilisationRatio(totalEffectiveBorrowBalanceValue, totalEffectiveCollateralBalanceValue) {
    if (totalEffectiveCollateralBalanceValue === BigInt(0))
        return BigInt(0);
    return (0, mathLib_1.divScale)(totalEffectiveBorrowBalanceValue, totalEffectiveCollateralBalanceValue, mathLib_1.ONE_4_DP);
}
exports.calcBorrowUtilisationRatio = calcBorrowUtilisationRatio;
/**
 * Calculates the loan's liquidation margin
 * @param totalEffectiveBorrowBalanceValue (4dp)
 * @param totalEffectiveCollateralBalanceValue (4dp)
 * @return liquidationMargin (4dp)
 */
function calcLiquidationMargin(totalEffectiveBorrowBalanceValue, totalEffectiveCollateralBalanceValue) {
    if (totalEffectiveCollateralBalanceValue === BigInt(0))
        return BigInt(0);
    return (0, mathLib_1.divScale)(totalEffectiveCollateralBalanceValue - totalEffectiveBorrowBalanceValue, totalEffectiveCollateralBalanceValue, mathLib_1.ONE_4_DP);
}
exports.calcLiquidationMargin = calcLiquidationMargin;
/**
 * Calculates the borrow balance of the loan at time t
 * @param bbtn1 (0dp)
 * @param biit (14dp)
 * @param biitn1 (14dp)
 * @return borrowBalance (0dp)
 */
function calcBorrowBalance(bbtn1, biit, biitn1) {
    return (0, mathLib_1.mulScaleRoundUp)(bbtn1, (0, mathLib_1.divScaleRoundUp)(biit, biitn1, mathLib_1.ONE_14_DP), mathLib_1.ONE_14_DP);
}
exports.calcBorrowBalance = calcBorrowBalance;
/**
 * Calculates the stable borrow interest rate of the loan after a borrow increase
 * @param bbt (0dp)
 * @param amount (0dp)
 * @param sbirtn1 (16dp)
 * @param sbirt1 (16dp)
 * @return stableInterestRate (16dp)
 */
function calcLoanStableInterestRate(bbt, amount, sbirtn1, sbirt1) {
    return (bbt * sbirtn1 + amount * sbirt1) / (bbt + amount);
}
exports.calcLoanStableInterestRate = calcLoanStableInterestRate;
/**
 * Calculates the deposit interest rate condition required to rebalance up stable borrow.
 * Note that there is also a second condition on the pool utilisation ratio.
 * @param rudir (16dp)
 * @param vr0 (16dp)
 * @param vr1 (16dp)
 * @param vr2 (16dp)
 * @return rebalanceUpThreshold (16dp)
 */
function calcRebalanceUpThreshold(rudir, vr0, vr1, vr2) {
    return (0, mathLib_1.mulScale)(rudir, vr0 + vr1 + vr2, mathLib_1.ONE_16_DP);
}
exports.calcRebalanceUpThreshold = calcRebalanceUpThreshold;
/**
 * Calculates the stable interest rate condition required to rebalance down stable borrow
 * @param rdd (16dp)
 * @param sbirt1 (16dp)
 * @return rebalanceDownThreshold (16dp)
 */
function calcRebalanceDownThreshold(rdd, sbirt1) {
    return (0, mathLib_1.mulScale)(mathLib_1.ONE_16_DP + rdd, sbirt1, mathLib_1.ONE_16_DP);
}
exports.calcRebalanceDownThreshold = calcRebalanceDownThreshold;
/**
 * Calculates the flash loan repayment amount for a given borrow amount and fee
 * @param borrowAmount (0dp)
 * @param fee (16dp)
 * @return repaymentAmount (0dp)
 */
function calcFlashLoanRepayment(borrowAmount, fee) {
    return borrowAmount + (0, mathLib_1.mulScaleRoundUp)(borrowAmount, fee, mathLib_1.ONE_16_DP);
}
exports.calcFlashLoanRepayment = calcFlashLoanRepayment;
/**
 * Calculates the LP price
 * @param r0 pool supply of asset 0
 * @param r1 pool supply of asset 1
 * @param p0 price of asset 0
 * @param p1 price of asset 1
 * @param lts circulating supply of liquidity token
 * @return bigint LP price
 */
function calcLPPrice(r0, r1, p0, p1, lts) {
    return BigInt(2) * ((0, mathLib_1.sqrt)(r0 * p0 * r1 * p1) / lts);
}
exports.calcLPPrice = calcLPPrice;
//# sourceMappingURL=formulae.js.map