"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.wrapWithFlashLoan = exports.prepareFlashLoanEnd = exports.prepareFlashLoanBegin = exports.prepareRemoveUserLoan = exports.prepareRebalanceDownLoan = exports.prepareRebalanceUpLoan = exports.prepareLiquidateLoan = exports.prepareRepayLoanWithCollateral = exports.prepareRepayLoanWithTxn = exports.prepareSwitchBorrowTypeInLoan = exports.prepareBorrowFromLoan = exports.prepareRemoveCollateralFromLoan = exports.prepareSwapCollateralInLoanEnd = exports.prepareSwapCollateralInLoanBegin = exports.prepareReduceCollateralFromLoan = exports.prepareSyncCollateralInLoan = exports.prepareAddCollateralToLoan = exports.prepareCreateUserLoan = exports.getMaxBorrowForBorrowUtilisationRatio = exports.getMaxReduceCollateralForBorrowUtilisationRatio = exports.retrieveLiquidatableLoans = exports.retrieveUserLoanInfo = exports.retrieveUserLoansInfo = exports.userLoanInfo = exports.retrieveLoanLocalState = exports.retrieveLoansLocalState = exports.retrieveLoanInfo = void 0;
const algosdk_1 = require("algosdk");
const mathLib_1 = require("../../mathLib");
const utils_1 = require("../../utils");
const abiContracts_1 = require("./abiContracts");
const deposit_1 = require("./deposit");
const formulae_1 = require("./formulae");
const oracle_1 = require("./oracle");
const utils_2 = require("./utils");
Object.defineProperty(exports, "userLoanInfo", { enumerable: true, get: function () { return utils_2.userLoanInfo; } });
/**
 *
 * Returns information regarding the given pool.
 *
 * @param client - Algorand client to query
 * @param loanAppId - loan application to query about
 * @returns Promise<LoanInfo[]> loan info
 */
function retrieveLoanInfo(client, loanAppId) {
    return __awaiter(this, void 0, void 0, function* () {
        const { currentRound, globalState: state } = yield (0, utils_1.getApplicationGlobalState)(client, loanAppId);
        if (state === undefined)
            throw Error("Could not find Loan");
        const paramsBase64Value = String((0, utils_1.getParsedValueFromState)(state, "pa"));
        const paramsValue = Buffer.from(paramsBase64Value, "base64").toString("hex");
        const canSwapCollateral = Boolean(BigInt("0x" + paramsValue.slice(96, 98)));
        const pools = {};
        for (let i = 0; i < 63; i++) {
            const poolBase64Value = String((0, utils_1.getParsedValueFromState)(state, (0, utils_1.fromIntToByteHex)(i), "hex"));
            const poolValue = Buffer.from(poolBase64Value, "base64").toString("hex");
            for (let j = 0; j < 3; j++) {
                const basePos = j * 84;
                const poolAppId = Number("0x" + poolValue.slice(basePos, basePos + 16));
                // add pool
                if (poolAppId > 0) {
                    pools[poolAppId] = {
                        poolAppId,
                        assetId: Number("0x" + poolValue.slice(basePos + 16, basePos + 32)),
                        collateralCap: BigInt("0x" + poolValue.slice(basePos + 32, basePos + 48)),
                        collateralUsed: BigInt("0x" + poolValue.slice(basePos + 48, basePos + 64)),
                        collateralFactor: BigInt("0x" + poolValue.slice(basePos + 64, basePos + 68)),
                        borrowFactor: BigInt("0x" + poolValue.slice(basePos + 68, basePos + 72)),
                        liquidationMax: BigInt("0x" + poolValue.slice(basePos + 72, basePos + 76)),
                        liquidationBonus: BigInt("0x" + poolValue.slice(basePos + 76, basePos + 80)),
                        liquidationFee: BigInt("0x" + poolValue.slice(basePos + 80, basePos + 84)),
                    };
                }
            }
        }
        // combine
        return { currentRound, canSwapCollateral, pools };
    });
}
exports.retrieveLoanInfo = retrieveLoanInfo;
/**
 *
 * Returns local state regarding the loan escrows of a given user.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param userAddr - account address for the user
 * @returns Promise<LoanLocalState[]> loan escrows' local state
 */
function retrieveLoansLocalState(indexerClient, loanAppId, userAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const loansLocalState = [];
        const escrows = yield (0, utils_2.getEscrows)(indexerClient, userAddr, loanAppId, "la ", "lr ");
        // get all remaining loans' local state
        for (const escrowAddr of escrows) {
            const { currentRound, localState: state } = yield (0, utils_1.getAccountApplicationLocalState)(indexerClient, loanAppId, escrowAddr);
            if (state === undefined)
                throw Error(`Could not find loan ${loanAppId} in escrow ${escrowAddr}`);
            loansLocalState.push(Object.assign({ currentRound }, (0, utils_2.loanLocalState)(state, loanAppId, escrowAddr)));
        }
        return loansLocalState;
    });
}
exports.retrieveLoansLocalState = retrieveLoansLocalState;
/**
 *
 * Returns local state of given escrow.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param client - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param escrowAddr - account address for the loan escrow
 * @returns Promise<LoanLocalState> loan escrow local state
 */
function retrieveLoanLocalState(client, loanAppId, escrowAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const { currentRound, localState: state } = yield (0, utils_1.getAccountApplicationLocalState)(client, loanAppId, escrowAddr);
        if (state === undefined)
            throw Error(`Could not find loan ${loanAppId} in escrow ${escrowAddr}`);
        return Object.assign({ currentRound }, (0, utils_2.loanLocalState)(state, loanAppId, escrowAddr));
    });
}
exports.retrieveLoanLocalState = retrieveLoanLocalState;
/**
 *
 * Returns information regarding the loan escrows of a given user.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param poolManagerAppId - pool manager application to query about
 * @param oracle - oracle to query
 * @param userAddr - account address for the user
 * @returns Promise<UserLoanInfo[]> user loans infos
 */
function retrieveUserLoansInfo(indexerClient, loanAppId, poolManagerAppId, oracle, userAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        const userLoanInfos = [];
        // get all prerequisites
        const escrowsReq = (0, utils_2.getEscrows)(indexerClient, userAddr, loanAppId, "la ", "lr ");
        const loanInfoReq = retrieveLoanInfo(indexerClient, loanAppId);
        const poolManagerInfoReq = (0, deposit_1.retrievePoolManagerInfo)(indexerClient, poolManagerAppId);
        const oraclePricesReq = (0, oracle_1.getOraclePrices)(indexerClient, oracle);
        const [escrows, poolManagerInfo, loanInfo, oraclePrices] = yield Promise.all([
            escrowsReq,
            poolManagerInfoReq,
            loanInfoReq,
            oraclePricesReq,
        ]);
        // get all remaining loans' info
        for (const escrowAddr of escrows) {
            const { currentRound, localState: state } = yield (0, utils_1.getAccountApplicationLocalState)(indexerClient, loanAppId, escrowAddr);
            if (state === undefined)
                throw Error(`Could not find loan ${loanAppId} in escrow ${escrowAddr}`);
            const localState = (0, utils_2.loanLocalState)(state, loanAppId, escrowAddr);
            userLoanInfos.push(Object.assign(Object.assign({}, (0, utils_2.userLoanInfo)(localState, poolManagerInfo, loanInfo, oraclePrices)), { currentRound }));
        }
        return userLoanInfos;
    });
}
exports.retrieveUserLoansInfo = retrieveUserLoansInfo;
/**
 *
 * Returns information regarding the given user loan escrow.
 *
 * @param client - Algorand client to query
 * @param loanAppId - loan application to query about
 * @param poolManagerAppId - pool manager application to query about
 * @param oracle - oracle to query
 * @param escrowAddr - account address for the loan escrow
 * @returns Promise<UserLoanInfo> user loan info
 */
function retrieveUserLoanInfo(client, loanAppId, poolManagerAppId, oracle, escrowAddr) {
    return __awaiter(this, void 0, void 0, function* () {
        // get all prerequisites
        const loanInfoReq = retrieveLoanInfo(client, loanAppId);
        const poolManagerInfoReq = (0, deposit_1.retrievePoolManagerInfo)(client, poolManagerAppId);
        const oraclePricesReq = (0, oracle_1.getOraclePrices)(client, oracle);
        const [poolInfo, loanInfo, oraclePrices] = yield Promise.all([poolManagerInfoReq, loanInfoReq, oraclePricesReq]);
        // get loan info
        const { currentRound, localState: state } = yield (0, utils_1.getAccountApplicationLocalState)(client, loanAppId, escrowAddr);
        if (state === undefined)
            throw Error(`Could not find loan ${loanAppId} in escrow ${escrowAddr}`);
        const localState = (0, utils_2.loanLocalState)(state, loanAppId, escrowAddr);
        return Object.assign(Object.assign({}, (0, utils_2.userLoanInfo)(localState, poolInfo, loanInfo, oraclePrices)), { currentRound });
    });
}
exports.retrieveUserLoanInfo = retrieveUserLoanInfo;
/**
 *
 * Returns all loans that are liquidatable.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param poolManagerInfo - pool manager info which is returned by retrievePoolManagerInfo function
 * @param loanInfo - loan info which is returned by retrieveLoanInfo function
 * @param oraclePrices - oracle prices which is returned by getOraclePrices function
 * @param nextToken - token for retrieving next escrows
 * @returns Promise<{ loans: UserLoanInfo[], nextToken?: string}> object containing liquidatable loans and next token
 */
function retrieveLiquidatableLoans(indexerClient, loanAppId, poolManagerInfo, loanInfo, oraclePrices, nextToken) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const loans = [];
        const req = yield indexerClient
            .searchAccounts()
            .applicationID(loanAppId)
            .exclude("assets,created-assets,created-apps");
        if (nextToken !== undefined)
            req.nextToken(nextToken);
        const res = yield req.do();
        // metadata
        const currentRound = res["current-round"];
        nextToken = res["next-token"];
        // convert to user loan info and add if liquidatable
        for (const acc of res["accounts"]) {
            const escrowAddr = acc["address"];
            const state = (_b = (_a = acc["apps-local-state"]) === null || _a === void 0 ? void 0 : _a.find(({ id }) => id === loanAppId)) === null || _b === void 0 ? void 0 : _b["key-value"];
            const localState = (0, utils_2.loanLocalState)(state, loanAppId, escrowAddr);
            const loan = (0, utils_2.userLoanInfo)(localState, poolManagerInfo, loanInfo, oraclePrices);
            if (loan.totalEffectiveCollateralBalanceValue < loan.totalEffectiveBorrowBalanceValue)
                loans.push(Object.assign(Object.assign({}, loan), { currentRound }));
        }
        return { loans, nextToken };
    });
}
exports.retrieveLiquidatableLoans = retrieveLiquidatableLoans;
/**
 *
 * Returns the maximum reduce collateral of a given f asset considering a target borrow utilisation ratio.
 * Returns 0 if no collateral in loan.
 * Returns at most the collateral balance.
 *
 * @param loan - user loan
 * @param colPoolAppId - price of asset borrowing 14 d.p.
 * @param targetBorrowUtilisationRatio - the utilisation ratio that you are targeting 4 d.p.
 * @returns bigint max f asset amount
 */
function getMaxReduceCollateralForBorrowUtilisationRatio(loan, colPoolAppId, targetBorrowUtilisationRatio) {
    const collateral = loan.collaterals.find(({ poolAppId }) => poolAppId === colPoolAppId);
    // if could not find collateral or target is below actual, return 0
    if (!collateral || targetBorrowUtilisationRatio <= loan.borrowUtilisationRatio)
        return BigInt(0);
    // check if can reduce all collateral (special case as lack required precision otherwise)
    const newEffectiveBalanceValue = loan.totalEffectiveCollateralBalanceValue - collateral.effectiveBalanceValue;
    const newBorrowUtilisationRatio = (0, formulae_1.calcBorrowUtilisationRatio)(loan.totalEffectiveBorrowBalanceValue, newEffectiveBalanceValue);
    if (!(newEffectiveBalanceValue === BigInt(0) && loan.totalEffectiveBorrowBalanceValue > BigInt(0)) &&
        newBorrowUtilisationRatio <= targetBorrowUtilisationRatio)
        return collateral.fAssetBalance;
    // calculate max
    const targetEffectiveCollateralBalanceValue = (0, mathLib_1.divScale)(loan.totalEffectiveBorrowBalanceValue, targetBorrowUtilisationRatio, mathLib_1.ONE_4_DP); // 4 d.p.
    const deltaEffectiveBalanceValue = loan.totalEffectiveCollateralBalanceValue - targetEffectiveCollateralBalanceValue; // 4 d.p.
    const deltaBalanceValue = (0, mathLib_1.divScale)(deltaEffectiveBalanceValue, collateral.collateralFactor, mathLib_1.ONE_4_DP); // 4 d.p.
    const deltaAssetBalance = (0, mathLib_1.divScale)(deltaBalanceValue, collateral.assetPrice, mathLib_1.ONE_10_DP); // 0 d.p.
    const deltafAssetBalance = (0, formulae_1.calcDepositReturn)(deltaAssetBalance, collateral.depositInterestIndex);
    return (0, mathLib_1.minimum)(deltafAssetBalance, collateral.fAssetBalance);
}
exports.getMaxReduceCollateralForBorrowUtilisationRatio = getMaxReduceCollateralForBorrowUtilisationRatio;
/**
 *
 * Returns the maximum borrow of a given asset considering a target borrow utilisation ratio.
 * Returns 0 if cannot borrow anything more.
 *
 * @param loan - user loan
 * @param assetPrice - price of asset borrowing 14 d.p.
 * @param borrowFactor - borrow factor of asset borrowing 4 d.p.
 * @param targetBorrowUtilisationRatio - the utilisation ratio that you are targeting 4 d.p.
 * @returns bigint max asset amount
 */
function getMaxBorrowForBorrowUtilisationRatio(loan, assetPrice, borrowFactor, targetBorrowUtilisationRatio) {
    // if target is below actual, return 0
    if (targetBorrowUtilisationRatio <= loan.borrowUtilisationRatio)
        return BigInt(0);
    // calculate max
    const targetEffectiveBorrowBalanceValue = (0, mathLib_1.mulScale)(loan.totalEffectiveCollateralBalanceValue, targetBorrowUtilisationRatio, mathLib_1.ONE_4_DP); // 4 d.p.
    const deltaEffectiveBalanceValue = targetEffectiveBorrowBalanceValue - loan.totalEffectiveBorrowBalanceValue; // 4 d.p.
    const deltaBalanceValue = (0, mathLib_1.divScale)(deltaEffectiveBalanceValue, borrowFactor, mathLib_1.ONE_4_DP); // 4 d.p.
    const deltaAssetBalance = (0, mathLib_1.divScale)(deltaBalanceValue, assetPrice, mathLib_1.ONE_10_DP); // 0 d.p.
    return deltaAssetBalance;
}
exports.getMaxBorrowForBorrowUtilisationRatio = getMaxBorrowForBorrowUtilisationRatio;
/**
 *
 * Returns a group transaction to create loan escrow.
 *
 * @param loanAppId - loan application to add escrow for
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: Account } object containing group transaction and generated escrow account
 */
function prepareCreateUserLoan(loanAppId, userAddr, params) {
    const escrow = (0, algosdk_1.generateAccount)();
    const userCall = (0, utils_1.addEscrowNoteTransaction)(userAddr, escrow.addr, loanAppId, "la ", Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: escrow.addr,
        signer: utils_1.signer,
        appID: loanAppId,
        onComplete: algosdk_1.OnApplicationComplete.OptInOC,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "create_loan"),
        methodArgs: [{ txn: userCall, signer: utils_1.signer }],
        rekeyTo: (0, algosdk_1.getApplicationAddress)(loanAppId),
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return { txns, escrow };
}
exports.prepareCreateUserLoan = prepareCreateUserLoan;
/**
 *
 * Returns a transaction to add a collateral to a loan escrow.
 *
 * @param loanAppId - loan application to add collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to add f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction add collateral to loan escrow transaction
 */
function prepareAddCollateralToLoan(loanAppId, poolManagerAppId, userAddr, escrowAddr, pool, params) {
    const { appId: poolAppId, fAssetId, poolManagerIndex, loans: poolLoans } = pool;
    const poolLoanIndex = poolLoans[loanAppId];
    if (poolLoanIndex === undefined)
        throw Error("Pool is not in loan");
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "add_collateral"),
        methodArgs: [escrowAddr, fAssetId, poolAppId, poolManagerIndex, poolLoanIndex, poolManagerAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareAddCollateralToLoan = prepareAddCollateralToLoan;
/**
 *
 * Returns a transaction to sync collateral of a loan escrow.
 *
 * @param loanAppId - loan application to sync collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to sync collateral of
 * @param oracle - oracle application to retrieve asset price from
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] sync collateral group transaction
 */
function prepareSyncCollateralInLoan(loanAppId, poolManagerAppId, userAddr, escrowAddr, pool, oracle, params) {
    const { appId: poolAppId, fAssetId } = pool;
    const { oracleAdapterAppId } = oracle;
    // refresh prices
    const lpAssets = "lpToken" in pool ? [pool.lpToken] : [];
    const baseAssetIds = "lpToken" in pool ? [] : [pool.assetId];
    const refreshPrices = (0, oracle_1.prepareRefreshPricesInOracleAdapter)(oracle, userAddr, lpAssets, baseAssetIds, params);
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "sync_collateral"),
        methodArgs: [
            { txn: refreshPrices[refreshPrices.length - 1], signer: utils_1.signer },
            escrowAddr,
            fAssetId,
            poolAppId,
            poolManagerAppId,
            oracleAdapterAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 1000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return [...refreshPrices.slice(0, -1), ...txns];
}
exports.prepareSyncCollateralInLoan = prepareSyncCollateralInLoan;
/**
 *
 * Returns a transaction to sync collateral of a loan escrow.
 *
 * @param loanAppId - loan application to reduce collateral in
 * @param poolManagerAppId - pool manager application*
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the collateral (typically the user's deposit escrow)
 * @param pool - pool to reduce collateral of
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param amount - the amount of asset / f asset to reduce the collateral by
 * @param isfAssetAmount - whether the amount of collateral to reduce by is expressed in terms of f asset or asset
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] reduce collateral group transaction
 */
function prepareReduceCollateralFromLoan(loanAppId, poolManagerAppId, userAddr, escrowAddr, receiverAddr, pool, oracle, lpAssets, baseAssetIds, amount, isfAssetAmount, params) {
    const { appId: poolAppId, assetId, fAssetId } = pool;
    const { oracleAdapterAppId } = oracle;
    const refreshPrices = (0, oracle_1.prepareRefreshPricesInOracleAdapter)(oracle, userAddr, lpAssets, baseAssetIds, params);
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "reduce_collateral"),
        methodArgs: [
            { txn: refreshPrices[refreshPrices.length - 1], signer: utils_1.signer },
            escrowAddr,
            receiverAddr,
            assetId,
            fAssetId,
            amount,
            isfAssetAmount,
            poolAppId,
            poolManagerAppId,
            oracleAdapterAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 6000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return [...refreshPrices.slice(0, -1), ...txns];
}
exports.prepareReduceCollateralFromLoan = prepareReduceCollateralFromLoan;
/**
 *
 * Returns a transaction to begin swap collateral in a loan escrow.
 * Must be groped together with swap_collateral_end call.
 *
 * @param loanAppId - loan application to swap collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the collateral (typically the user's deposit escrow)
 * @param pool - pool to swap collateral of
 * @param amount - the amount of asset / f asset to reduce the collateral by
 * @param isfAssetAmount - whether the amount of collateral to reduce by is expressed in terms of f asset or asset
 * @param txnIndexForSwapCollateralEnd - transaction index in the group transaction for the swap_collateral_end call.
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction swap collateral begin transaction
 */
function prepareSwapCollateralInLoanBegin(loanAppId, poolManagerAppId, userAddr, escrowAddr, receiverAddr, pool, amount, isfAssetAmount, txnIndexForSwapCollateralEnd, params) {
    const { appId: poolAppId, assetId, fAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "swap_collateral_begin"),
        methodArgs: [
            escrowAddr,
            receiverAddr,
            assetId,
            fAssetId,
            amount,
            isfAssetAmount,
            txnIndexForSwapCollateralEnd,
            poolAppId,
            poolManagerAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 6000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareSwapCollateralInLoanBegin = prepareSwapCollateralInLoanBegin;
/**
 *
 * Returns a group transaction to end swap collateral in a loan escrow.
 * Must be groped together with swap_collateral_begin call.
 *
 * @param loanAppId - loan application to swap collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] swap collateral end group transaction
 */
function prepareSwapCollateralInLoanEnd(loanAppId, poolManagerAppId, userAddr, escrowAddr, oracle, lpAssets, baseAssetIds, params) {
    const { oracleAdapterAppId } = oracle;
    const refreshPrices = (0, oracle_1.prepareRefreshPricesInOracleAdapter)(oracle, userAddr, lpAssets, baseAssetIds, params);
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "swap_collateral_end"),
        methodArgs: [
            { txn: refreshPrices[refreshPrices.length - 1], signer: utils_1.signer },
            escrowAddr,
            poolManagerAppId,
            oracleAdapterAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 1000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return [...refreshPrices.slice(0, -1), ...txns];
}
exports.prepareSwapCollateralInLoanEnd = prepareSwapCollateralInLoanEnd;
/**
 *
 * Returns a transaction to remove collateral from loan escrow.
 *
 * @param loanAppId - loan application to remove collateral in
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to remove collateral of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction remove collateral transaction
 */
function prepareRemoveCollateralFromLoan(loanAppId, userAddr, escrowAddr, pool, params) {
    const { appId: poolAppId, fAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "remove_collateral"),
        methodArgs: [escrowAddr, fAssetId, poolAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareRemoveCollateralFromLoan = prepareRemoveCollateralFromLoan;
/**
 *
 * Returns a group transaction to borrow using loan escrow.
 *
 * @param loanAppId - loan application to borrow in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the borrow (typically the user)
 * @param pool - pool to borrow from
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param borrowAmount - amount to borrow of asset
 * @param maxStableRate - maximum stable rate of the borrow, if zero then borrow is interpreted as a variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] borrow group transaction
 */
function prepareBorrowFromLoan(loanAppId, poolManagerAppId, userAddr, escrowAddr, receiverAddr, pool, oracle, lpAssets, baseAssetIds, borrowAmount, maxStableRate, // if zero then variable borrow
params) {
    const { appId: poolAppId, assetId, poolManagerIndex, loans: poolLoans } = pool;
    const { oracleAdapterAppId } = oracle;
    const poolLoanIndex = poolLoans[loanAppId];
    if (poolLoanIndex === undefined)
        throw Error("Pool is not in loan");
    const refreshPrices = (0, oracle_1.prepareRefreshPricesInOracleAdapter)(oracle, userAddr, lpAssets, baseAssetIds, params);
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "borrow"),
        methodArgs: [
            { txn: refreshPrices[refreshPrices.length - 1], signer: utils_1.signer },
            escrowAddr,
            receiverAddr,
            assetId,
            borrowAmount,
            maxStableRate,
            poolManagerIndex,
            poolLoanIndex,
            poolAppId,
            poolManagerAppId,
            oracleAdapterAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 8000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return [...refreshPrices.slice(0, -1), ...txns];
}
exports.prepareBorrowFromLoan = prepareBorrowFromLoan;
/**
 *
 * Returns a transaction to switch borrow type of a loan borrow.
 *
 * @param loanAppId - loan application to switch borrow type in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to switch borrow type of
 * @param maxStableRate - maximum stable rate of the borrow, if zero then interpreted as switching a stable rate borrow to a variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] switch borrow type transaction
 */
function prepareSwitchBorrowTypeInLoan(loanAppId, poolManagerAppId, userAddr, escrowAddr, pool, maxStableRate, // ignored if not switching to stable borrow
params) {
    const { appId: poolAppId, assetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "switch_borrow_type"),
        methodArgs: [escrowAddr, assetId, maxStableRate, poolAppId, poolManagerAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 6000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareSwitchBorrowTypeInLoan = prepareSwitchBorrowTypeInLoan;
/**
 *
 * Returns a group transaction to repay borrow in a loan escrow using assets sent from user.
 *
 * @param loanAppId - loan application to repay borrow in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the rewards if any (typically the user's deposit escrow)
 * @param reserveAddr - account address to receive the protocol revenue from the percentage of the accrued interest
 * @param pool - pool to repay borrow of
 * @param repayAmount - amount of borrow to repay
 * @param isStable - whether the borrow that is being repaid is a stable or variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] reduce borrow with transaction group transaction
 */
function prepareRepayLoanWithTxn(loanAppId, poolManagerAppId, userAddr, escrowAddr, receiverAddr, reserveAddr, pool, repayAmount, isStable, params) {
    const { appId: poolAppId, assetId, frAssetId } = pool;
    const sendAsset = (0, utils_1.transferAlgoOrAsset)(assetId, userAddr, (0, algosdk_1.getApplicationAddress)(poolAppId), repayAmount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "repay_with_txn"),
        methodArgs: [
            { txn: sendAsset, signer: utils_1.signer },
            escrowAddr,
            receiverAddr,
            reserveAddr,
            assetId,
            frAssetId,
            isStable,
            poolAppId,
            poolManagerAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 10000 }),
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareRepayLoanWithTxn = prepareRepayLoanWithTxn;
/**
 *
 * Returns a group transaction to repay borrow in a loan escrow using collateral.
 *
 * @param loanAppId - loan application to repay borrow in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the rewards if any (typically the user's deposit escrow)
 * @param reserveAddr - account address to receive the protocol revenue from the percentage of the accrued interest
 * @param pool - pool to repay borrow and use collateral of
 * @param repayAmount - amount of borrow to repay expressed in terms of the asset
 * @param isStable - whether the borrow that is being repaid is a stable or variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] repay borrow with collateral group transaction
 */
function prepareRepayLoanWithCollateral(loanAppId, poolManagerAppId, userAddr, escrowAddr, receiverAddr, reserveAddr, pool, repayAmount, isStable, params) {
    const { appId: poolAppId, assetId, fAssetId, frAssetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "repay_with_collateral"),
        methodArgs: [
            escrowAddr,
            receiverAddr,
            reserveAddr,
            assetId,
            fAssetId,
            frAssetId,
            repayAmount,
            isStable,
            poolAppId,
            poolManagerAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 14000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareRepayLoanWithCollateral = prepareRepayLoanWithCollateral;
/**
 *
 * Returns a group transaction to repay borrow in a loan escrow using assets sent from user.
 *
 * @param loanAppId - loan application to repay borrow in
 * @param poolManagerAppId - pool manager application
 * @param liquidatorAddr - account address for the liquidator
 * @param escrowAddr - account address for the loan escrow
 * @param reserveAddr - account address to receive the protocol revenue from the percentage of the accrued interest
 * @param collateralPool - pool to seize collateral of
 * @param borrowPool - pool to repay borrow of
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param repayAmount - amount of borrow to repay expressed in terms of borrow pool asset
 * @param minCollateralAmount - minimum collateral amount for the liquidator to receive expressed in terms of collateral pool f asset
 * @param isStable - whether the borrow that is being repaid is a stable or variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] liquidate group transaction
 */
function prepareLiquidateLoan(loanAppId, poolManagerAppId, liquidatorAddr, escrowAddr, reserveAddr, collateralPool, borrowPool, oracle, lpAssets, baseAssetIds, repayAmount, minCollateralAmount, isStable, params) {
    const { appId: colPoolAppId, fAssetId } = collateralPool;
    const { appId: borPoolAppId, assetId } = borrowPool;
    const { oracleAdapterAppId } = oracle;
    const refreshPrices = (0, oracle_1.prepareRefreshPricesInOracleAdapter)(oracle, liquidatorAddr, lpAssets, baseAssetIds, params);
    const sendAsset = (0, utils_1.transferAlgoOrAsset)(assetId, liquidatorAddr, (0, algosdk_1.getApplicationAddress)(borPoolAppId), repayAmount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: liquidatorAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "liquidate"),
        methodArgs: [
            { txn: refreshPrices[refreshPrices.length - 1], signer: utils_1.signer },
            { txn: sendAsset, signer: utils_1.signer },
            escrowAddr,
            reserveAddr,
            assetId,
            fAssetId,
            minCollateralAmount,
            isStable,
            colPoolAppId,
            borPoolAppId,
            poolManagerAppId,
            oracleAdapterAppId,
        ],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 10000 }),
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareLiquidateLoan = prepareLiquidateLoan;
/**
 *
 * Returns a transaction to rebalance up borrow in a loan escrow.
 *
 * @param loanAppId - loan application to rebalance up borrow in
 * @param poolManagerAppId - pool manager application*
 * @param rebalancerAddr - account address for the rebalancer
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to rebalance
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction rebalance up transaction
 */
function prepareRebalanceUpLoan(loanAppId, poolManagerAppId, rebalancerAddr, escrowAddr, pool, params) {
    const { appId: poolAppId, assetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: rebalancerAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "rebalance_up"),
        methodArgs: [escrowAddr, assetId, poolAppId, poolManagerAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 5000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareRebalanceUpLoan = prepareRebalanceUpLoan;
/**
 *
 * Returns a transaction to rebalance up borrow in a loan escrow.
 *
 * @param loanAppId - loan application to rebalance down borrow in
 * @param poolManagerAppId - pool manager application
 * @param rebalancerAddr - account address for the rebalancer
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to rebalance
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction rebalance down transaction
 */
function prepareRebalanceDownLoan(loanAppId, poolManagerAppId, rebalancerAddr, escrowAddr, pool, params) {
    const { appId: poolAppId, assetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: rebalancerAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "rebalance_down"),
        methodArgs: [escrowAddr, assetId, poolAppId, poolManagerAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 5000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareRebalanceDownLoan = prepareRebalanceDownLoan;
/**
 *
 * Returns a group transaction to remove a user's loan escrow and return its minimum balance to the user.
 *
 * @param loanAppId - loan application to remove loan in
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] remove and close out loan escrow group transaction
 */
function prepareRemoveUserLoan(loanAppId, userAddr, escrowAddr, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: loanAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.loanABIContract.methods, "remove_loan"),
        methodArgs: [escrowAddr],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 4000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    const optOutTx = (0, algosdk_1.makeApplicationCloseOutTxn)(escrowAddr, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }), loanAppId);
    const closeToTx = (0, utils_1.removeEscrowNoteTransaction)(escrowAddr, userAddr, "lr ", Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }));
    return [txns[0], optOutTx, closeToTx];
}
exports.prepareRemoveUserLoan = prepareRemoveUserLoan;
/**
 *
 * Returns a transaction to begin flash loan
 * Must be groped together with flash_loan_end call.
 *
 * @param pool - pool to borrow from
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the loan
 * @param borrowAmount - the amount of the asset to borrow
 * @param txnIndexForFlashLoanEnd - transaction index in the group transaction for the flash_loan_end call.
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction flash loan begin transaction
 */
function prepareFlashLoanBegin(pool, userAddr, receiverAddr, borrowAmount, txnIndexForFlashLoanEnd, params) {
    const { appId, assetId } = pool;
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.poolABIContract.methods, "flash_loan_begin"),
        methodArgs: [borrowAmount, txnIndexForFlashLoanEnd, receiverAddr, assetId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareFlashLoanBegin = prepareFlashLoanBegin;
/**
 *
 * Returns a group transaction to end flash loan.
 * Must be groped together with flash_loan_begin call.
 *
 * @param pool - pool borrowed from
 * @param userAddr - account address for the user
 * @param reserveAddr - account address to receive the protocol revenue from the flash loan fee
 * @param repaymentAmount - the amount of the asset to repay (borrow amount plus flash loan fee)
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] flash loan end group transaction
 */
function prepareFlashLoanEnd(pool, userAddr, reserveAddr, repaymentAmount, params) {
    const { appId, assetId } = pool;
    const sendAsset = (0, utils_1.transferAlgoOrAsset)(assetId, userAddr, (0, algosdk_1.getApplicationAddress)(appId), repaymentAmount, Object.assign(Object.assign({}, params), { flatFee: true, fee: 0 }));
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.poolABIContract.methods, "flash_loan_end"),
        methodArgs: [{ txn: sendAsset, signer: utils_1.signer }, reserveAddr, assetId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
    });
    return atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
}
exports.prepareFlashLoanEnd = prepareFlashLoanEnd;
/**
 *
 * Wraps given transactions with flash loan.
 *
 * @param txns - txns to wrap flash loan around
 * @param pool - pool to borrow from
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the loan
 * @param reserveAddr - account address to receive the protocol revenue from the flash loan fee
 * @param borrowAmount - the amount of the asset to borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @param flashLoanFee - fee for flash loan as 16 d.p integer (default 0.1%)
 * @returns Transaction[] group transaction wrapped with flash loan
 */
function wrapWithFlashLoan(txns, pool, userAddr, receiverAddr, reserveAddr, borrowAmount, params, flashLoanFee = BigInt(0.001e16)) {
    // clear group id in passed txns
    const wrappedTxns = txns.map((txn) => {
        txn.group = undefined;
        return txn;
    });
    // add flash loan begin
    const txnIndexForFlashLoanEnd = txns.length + 2;
    const flashLoanBegin = prepareFlashLoanBegin(pool, userAddr, receiverAddr, borrowAmount, txnIndexForFlashLoanEnd, params);
    wrappedTxns.unshift(flashLoanBegin);
    // add flash loan end
    const repaymentAmount = (0, formulae_1.calcFlashLoanRepayment)(BigInt(borrowAmount), flashLoanFee);
    const flashLoanEnd = prepareFlashLoanEnd(pool, userAddr, reserveAddr, repaymentAmount, params);
    wrappedTxns.push(...flashLoanEnd);
    // return txns wrapped with flash loan
    return wrappedTxns;
}
exports.wrapWithFlashLoan = wrapWithFlashLoan;
//# sourceMappingURL=loan.js.map