"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareRefreshPricesInOracleAdapter = exports.getOraclePrices = void 0;
const algosdk_1 = require("algosdk");
const mathLib_1 = require("../../mathLib");
const utils_1 = require("../../utils");
const abiContracts_1 = require("./abiContracts");
const formulae_1 = require("./formulae");
const types_1 = require("./types");
function parseOracleValue(base64Value) {
    const value = Buffer.from(base64Value, "base64").toString("hex");
    // [price (uint64), latest_update (uint64), ...]
    const price = BigInt("0x" + value.slice(0, 16));
    const timestamp = BigInt("0x" + value.slice(16, 32));
    return { price, timestamp };
}
function parseLPTokenOracleValue(lpAssetId, base64Value) {
    const value = Buffer.from(base64Value, "base64").toString("hex");
    // [type (uint8), a0_id (uint64), a1_id, (uint64), _ (uint64), _ (uint64), _ (uint64), pool (address/uint64)]
    const provider = Number("0x" + value.slice(0, 2));
    const asset0Id = Number(`0x${value.slice(2, 34)}`);
    const asset1Id = Number(`0x${value.slice(34, 50)}`);
    // check if LP token is Tinyman or Pact
    if (provider == types_1.LPTokenProvider.TINYMAN) {
        const poolAddress = (0, algosdk_1.encodeAddress)(Buffer.from(value.slice(82, 146), "hex"));
        return { provider, lpAssetId, asset0Id, asset1Id, lpPoolAddress: poolAddress };
    }
    else if (provider == types_1.LPTokenProvider.PACT) {
        const poolAppId = Number("0x" + value.slice(82, 98));
        return { provider, lpAssetId, asset0Id, asset1Id, lpPoolAppId: poolAppId };
    }
    else {
        throw Error("Unknown LP Token type");
    }
}
function getTinymanLPPrice(client, validatorAppId, poolAddress, p0, p1) {
    return __awaiter(this, void 0, void 0, function* () {
        const { localState: state } = yield (0, utils_1.getAccountApplicationLocalState)(client, validatorAppId, poolAddress);
        if (state === undefined)
            throw new Error(`Unable to find Tinyman Pool: ${poolAddress} for validator app ${validatorAppId}.`);
        const r0 = BigInt((0, utils_1.getParsedValueFromState)(state, "s1") || 0);
        const r1 = BigInt((0, utils_1.getParsedValueFromState)(state, "s2") || 0);
        const lts = BigInt((0, utils_1.getParsedValueFromState)(state, "ilt") || 0);
        return (0, formulae_1.calcLPPrice)(r0, r1, p0, p1, lts);
    });
}
function getPactLPPrice(client, poolAppId, p0, p1) {
    return __awaiter(this, void 0, void 0, function* () {
        const { globalState: state } = yield (0, utils_1.getApplicationGlobalState)(client, poolAppId);
        if (state === undefined)
            throw new Error(`Unable to find Pact Pool: ${poolAppId}.`);
        const r0 = BigInt((0, utils_1.getParsedValueFromState)(state, "A") || 0);
        const r1 = BigInt((0, utils_1.getParsedValueFromState)(state, "B") || 0);
        const lts = BigInt((0, utils_1.getParsedValueFromState)(state, "L") || 0);
        return (0, formulae_1.calcLPPrice)(r0, r1, p0, p1, lts);
    });
}
/**
 *
 * Returns oracle prices for given oracle and provided assets.
 *
 * @param client - Algorand client to query
 * @param oracle - oracle to query
 * @param assetIds - assets ids to get prices for, if undefined then returns all prices
 * @returns OraclePrices oracle prices
 */
function getOraclePrices(client, oracle, assetIds) {
    return __awaiter(this, void 0, void 0, function* () {
        const { oracle0AppId, lpTokenOracle } = oracle;
        const lookupApps = [(0, utils_1.getApplicationGlobalState)(client, oracle0AppId)];
        if (lpTokenOracle)
            lookupApps.push((0, utils_1.getApplicationGlobalState)(client, lpTokenOracle.appId));
        const [oracleRes, lpTokenOracleRes] = yield Promise.all(lookupApps);
        const { currentRound, globalState: oracleState } = oracleRes;
        const lpTokenOracleState = lpTokenOracleRes === null || lpTokenOracleRes === void 0 ? void 0 : lpTokenOracleRes.globalState;
        if (oracleState === undefined)
            throw Error("Could not find Oracle");
        if (lpTokenOracle && lpTokenOracleState === undefined)
            throw Error("Could not find LP Token Oracle");
        const prices = {};
        // get the assets for which we need to retrieve their prices
        const allAssetIds = oracleState
            .concat(lpTokenOracleState || [])
            .filter(({ key }) => {
            // remove non asset ids global state
            key = Buffer.from(key, "base64").toString("utf8");
            return key !== "updater_addr" && key !== "admin" && key !== "tinyman_validator_app_id" && key !== "td";
        })
            .map(({ key }) => {
            // convert key to asset id
            return (0, algosdk_1.decodeUint64)(Buffer.from(key, "base64"), "safe");
        });
        const assets = assetIds ? assetIds : allAssetIds;
        // retrieve asset prices
        const retrievePrices = assets.map((assetId) => __awaiter(this, void 0, void 0, function* () {
            let assetPrice;
            const lpTokenBase64Value = lpTokenOracle
                ? (0, utils_1.getParsedValueFromState)(lpTokenOracleState, (0, utils_1.fromIntToBytes8Hex)(assetId), "hex")
                : undefined;
            // lpTokenBase64Value defined iff asset is lp token in given lpTokenOracle
            if (lpTokenBase64Value !== undefined) {
                const lpTokenOracleValue = parseLPTokenOracleValue(assetId, lpTokenBase64Value);
                const { price: p0, timestamp: t0 } = parseOracleValue(String((0, utils_1.getParsedValueFromState)(oracleState, (0, utils_1.fromIntToBytes8Hex)(lpTokenOracleValue.asset0Id), "hex")));
                const { price: p1, timestamp: t1 } = parseOracleValue(String((0, utils_1.getParsedValueFromState)(oracleState, (0, utils_1.fromIntToBytes8Hex)(lpTokenOracleValue.asset1Id), "hex")));
                let price;
                switch (lpTokenOracleValue.provider) {
                    case types_1.LPTokenProvider.TINYMAN:
                        price = yield getTinymanLPPrice(client, lpTokenOracle.tinymanValidatorAppId, lpTokenOracleValue.lpPoolAddress, p0, p1);
                        break;
                    case types_1.LPTokenProvider.PACT:
                        price = yield getPactLPPrice(client, lpTokenOracleValue.lpPoolAppId, p0, p1);
                        break;
                    default:
                        throw Error("Unknown LP Token provider");
                }
                assetPrice = { price, timestamp: (0, mathLib_1.minimum)(t0, t1) };
            }
            else {
                assetPrice = parseOracleValue(String((0, utils_1.getParsedValueFromState)(oracleState, (0, utils_1.fromIntToBytes8Hex)(assetId), "hex")));
            }
            prices[assetId] = assetPrice;
        }));
        yield Promise.all(retrievePrices);
        return { currentRound, prices };
    });
}
exports.getOraclePrices = getOraclePrices;
/**
 *
 * Returns a group transaction to refresh the given assets.
 *
 * @param oracle - oracle applications to use
 * @param userAddr - account address for the user
 * @param lpAssets - list of lp assets
 * @param baseAssetIds - list of base asset ids (non-lp assets)
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] refresh prices group transaction
 */
function prepareRefreshPricesInOracleAdapter(oracle, userAddr, lpAssets, baseAssetIds, params) {
    const { oracleAdapterAppId, lpTokenOracle, oracle0AppId } = oracle;
    if (lpTokenOracle === undefined && lpAssets.length > 0)
        throw Error("Cannot refresh LP assets without LP Token Oracle");
    const atc = new algosdk_1.AtomicTransactionComposer();
    // divide lp tokens into Tinyman and Pact
    const tinymanLPAssets = lpAssets.filter(({ provider }) => provider === types_1.LPTokenProvider.TINYMAN);
    const pactLPAssets = lpAssets.filter(({ provider }) => provider === types_1.LPTokenProvider.PACT);
    // update lp tokens
    const foreignAccounts = [];
    const foreignApps = [];
    const MAX_TINYMAN_UPDATE = 4;
    const MAX_PACT_UPDATE = 8;
    const MAX_COMBINATION_UPDATE = 7;
    let tinymanIndex = 0;
    let pactIndex = 0;
    while (tinymanIndex < tinymanLPAssets.length && pactIndex < pactLPAssets.length) {
        // retrieve which lp assets to update
        const tinymanLPUpdates = tinymanLPAssets.slice(tinymanIndex, tinymanIndex + MAX_TINYMAN_UPDATE);
        const maxPactUpdates = tinymanLPUpdates.length === 0 ? MAX_PACT_UPDATE : MAX_COMBINATION_UPDATE - tinymanLPUpdates.length;
        const pactLPUpdates = pactLPAssets.slice(pactIndex, pactIndex + maxPactUpdates);
        // prepare update lp tokens arguments
        const lpAssetIds = [
            ...tinymanLPUpdates.map(({ lpAssetId }) => lpAssetId),
            ...pactLPUpdates.map(({ lpAssetId }) => lpAssetId),
        ];
        // foreign arrays
        foreignAccounts.push(tinymanLPUpdates.map(({ lpPoolAddress }) => lpPoolAddress));
        const apps = [];
        if (tinymanLPUpdates.length > 0)
            apps.push(lpTokenOracle.tinymanValidatorAppId);
        pactLPUpdates.forEach(({ lpPoolAppId }) => apps.push(lpPoolAppId));
        foreignApps.push(apps);
        // update lp
        atc.addMethodCall({
            sender: userAddr,
            signer: utils_1.signer,
            appID: lpTokenOracle.appId,
            method: (0, algosdk_1.getMethodByName)(abiContracts_1.lpTokenOracleABIContract.methods, "update_lp_tokens"),
            methodArgs: [lpAssetIds],
            suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 1000 }),
        });
        // increase indexes
        tinymanIndex += tinymanLPUpdates.length;
        pactIndex += pactLPUpdates.length;
    }
    // prepare refresh prices arguments
    const oracle1AppId = oracle.oracle1AppId || 0;
    const lpTokenOracleAppId = (lpTokenOracle === null || lpTokenOracle === void 0 ? void 0 : lpTokenOracle.appId) || 0;
    const lpAssetIds = lpAssets.map(({ lpAssetId }) => lpAssetId);
    // refresh prices
    atc.addMethodCall({
        sender: userAddr,
        signer: utils_1.signer,
        appID: oracleAdapterAppId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.oracleAdapterABIContract.methods, "refresh_prices"),
        methodArgs: [lpAssetIds, baseAssetIds, oracle0AppId, oracle1AppId, lpTokenOracleAppId],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 1000 }),
    });
    // build
    return atc.buildGroup().map(({ txn }, index) => {
        if (index < foreignAccounts.length && index < foreignApps.length) {
            txn.appAccounts = foreignAccounts[index].map((address) => (0, algosdk_1.decodeAddress)(address));
            txn.appForeignApps = foreignApps[index];
        }
        txn.group = undefined;
        return txn;
    });
}
exports.prepareRefreshPricesInOracleAdapter = prepareRefreshPricesInOracleAdapter;
//# sourceMappingURL=oracle.js.map