"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.userLoanInfo = exports.loanLocalState = exports.userDepositStakingInfo = exports.depositStakingProgramsInfo = exports.depositStakingLocalState = exports.getEscrows = void 0;
const algosdk_1 = require("algosdk");
const mathLib_1 = require("../../mathLib");
const utils_1 = require("../../utils");
const formulae_1 = require("./formulae");
function getEscrows(indexerClient, userAddr, appId, addNotePrefix, removeNotePrefix) {
    return __awaiter(this, void 0, void 0, function* () {
        const escrows = new Set();
        const appAddress = (0, algosdk_1.getApplicationAddress)(appId);
        const addedReq = indexerClient
            .searchForTransactions()
            .address(userAddr)
            .addressRole("sender")
            .txType("pay")
            .notePrefix(utils_1.enc.encode(addNotePrefix))
            .do();
        const removedReq = indexerClient
            .searchForTransactions()
            .address(userAddr)
            .addressRole("receiver")
            .txType("pay")
            .notePrefix(utils_1.enc.encode(removeNotePrefix))
            .do();
        const [added, removed] = yield Promise.all([addedReq, removedReq]);
        for (const txn of added["transactions"]) {
            const receiver = txn["payment-transaction"]["receiver"];
            if (receiver === appAddress) {
                const note = Buffer.from(txn["note"], "base64");
                const address = (0, algosdk_1.encodeAddress)(note.slice(addNotePrefix.length));
                escrows.add(address);
            }
        }
        for (const txn of removed["transactions"]) {
            const sender = txn["sender"];
            escrows.delete(sender);
        }
        return escrows;
    });
}
exports.getEscrows = getEscrows;
/**
 *
 * Derives deposit staking local state from escrow account.
 *
 * @param state - escrow account local state
 * @param depositStakingAppId - deposit staking application to query about
 * @param escrowAddr - escrow address
 * @returns UserDepositStakingLocalState user deposit staking local state
 */
function depositStakingLocalState(state, depositStakingAppId, escrowAddr) {
    // standard
    const userAddress = (0, algosdk_1.encodeAddress)(Buffer.from(String((0, utils_1.getParsedValueFromState)(state, "ua")), "base64"));
    const stakedAmounts = [];
    for (let i = 0; i < 2; i++) {
        const prefix = "S".charCodeAt(0).toString(16);
        const stakeBase64Value = String((0, utils_1.getParsedValueFromState)(state, prefix + (0, utils_1.fromIntToByteHex)(i), "hex"));
        const stakeValue = (0, utils_1.parseUint64s)(stakeBase64Value);
        stakedAmounts.push(...stakeValue);
    }
    const rewardPerTokens = [];
    for (let i = 0; i < 6; i++) {
        const prefix = "R".charCodeAt(0).toString(16);
        const rewardBase64Value = String((0, utils_1.getParsedValueFromState)(state, prefix + (0, utils_1.fromIntToByteHex)(i), "hex"));
        const rewardValue = (0, utils_1.parseUint64s)(rewardBase64Value);
        rewardPerTokens.push(...rewardValue);
    }
    const unclaimedRewards = [];
    for (let i = 0; i < 6; i++) {
        const prefix = "U".charCodeAt(0).toString(16);
        const unclaimedBase64Value = String((0, utils_1.getParsedValueFromState)(state, prefix + (0, utils_1.fromIntToByteHex)(i), "hex"));
        const unclaimedValue = (0, utils_1.parseUint64s)(unclaimedBase64Value);
        unclaimedRewards.push(...unclaimedValue);
    }
    return {
        userAddress,
        escrowAddress: escrowAddr,
        optedIntoAssets: new Set(),
        stakedAmounts,
        rewardPerTokens,
        unclaimedRewards,
    };
}
exports.depositStakingLocalState = depositStakingLocalState;
/**
 *
 * Derives deposit staking programs info from deposit staking info.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param depositStakingInfo - deposit staking info which is returned by retrieveDepositStakingInfo function
 * @param poolManagerInfo - pool manager info which is returned by retrievePoolManagerInfo function
 * @param pools - pools in pool manager (either MainnetPools or TestnetPools)
 * @param oraclePrices - oracle prices which is returned by getOraclePrices function
 * @returns Promise<DepositStakingProgramInfo[]> deposit staking programs info
 */
function depositStakingProgramsInfo(depositStakingInfo, poolManagerInfo, pools, oraclePrices) {
    const stakingPrograms = [];
    const { pools: poolManagerPools } = poolManagerInfo;
    const { prices } = oraclePrices;
    depositStakingInfo.stakingPrograms
        .filter(({ poolAppId }) => poolAppId !== 0)
        .forEach(({ poolAppId, totalStaked, minTotalStaked, rewards, stakeIndex }) => {
        const pool = Object.entries(pools)
            .map(([, pool]) => pool)
            .find((pool) => pool.appId === poolAppId);
        const poolInfo = poolManagerPools[poolAppId];
        if (pool === undefined || poolInfo === undefined)
            throw Error("Could not find pool " + poolAppId);
        const { assetId, fAssetId } = pool;
        const { depositInterestIndex, depositInterestRate, depositInterestYield } = poolInfo;
        const oraclePrice = prices[assetId];
        if (oraclePrice === undefined)
            throw Error("Could not find asset price " + assetId);
        const { price: assetPrice } = oraclePrice;
        const fAssetTotalStakedAmount = (0, mathLib_1.maximum)(totalStaked, minTotalStaked);
        const assetTotalStakedAmount = (0, formulae_1.calcWithdrawReturn)(fAssetTotalStakedAmount, depositInterestIndex);
        const totalStakedAmountValue = (0, mathLib_1.mulScale)(assetTotalStakedAmount, assetPrice, mathLib_1.ONE_10_DP); // 4 d.p.
        const userRewards = [];
        rewards.forEach(({ rewardAssetId, endTimestamp, rewardRate, rewardPerToken }) => {
            const oraclePrice = prices[rewardAssetId];
            if (oraclePrice === undefined)
                throw Error("Could not find asset price " + rewardAssetId);
            const { price: rewardAssetPrice } = oraclePrice;
            const stakedAmountValue = assetTotalStakedAmount * assetPrice;
            const rewardInterestRate = (0, utils_1.unixTime)() < endTimestamp && stakedAmountValue !== BigInt(0)
                ? (rewardRate * BigInt(1e6) * rewardAssetPrice * mathLib_1.SECONDS_IN_YEAR) / stakedAmountValue
                : BigInt(0);
            userRewards.push({
                rewardAssetId,
                endTimestamp,
                rewardAssetPrice,
                rewardInterestRate,
                rewardRate,
                rewardPerToken,
            });
        });
        stakingPrograms.push({
            poolAppId,
            stakeIndex,
            fAssetId,
            fAssetTotalStakedAmount,
            assetId,
            assetPrice,
            assetTotalStakedAmount,
            totalStakedAmountValue,
            depositInterestRate,
            depositInterestYield,
            rewards: userRewards,
        });
    });
    return stakingPrograms;
}
exports.depositStakingProgramsInfo = depositStakingProgramsInfo;
/**
 *
 * Derives user loan info from escrow account.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param localState - local state of escrow account
 * @param poolManagerInfo - pool manager info which is returned by retrievePoolManagerInfo function*
 * @param depositStakingProgramsInfo - deposit staking programs info which is returned by depositStakingProgramsInfo function
 * @returns Promise<UserDepositStakingInfo> user loans info
 */
function userDepositStakingInfo(localState, poolManagerInfo, depositStakingProgramsInfo) {
    const stakingPrograms = [];
    const { pools: poolManagerPools } = poolManagerInfo;
    depositStakingProgramsInfo.forEach((stakingProgram, stakeIndex) => {
        const { poolAppId, fAssetId, assetId, assetPrice, depositInterestRate, depositInterestYield, rewards } = stakingProgram;
        const poolInfo = poolManagerPools[poolAppId];
        if (poolInfo === undefined)
            throw Error("Could not find pool " + poolAppId);
        const { depositInterestIndex } = poolInfo;
        const fAssetStakedAmount = localState.stakedAmounts[stakeIndex];
        const assetStakedAmount = (0, formulae_1.calcWithdrawReturn)(fAssetStakedAmount, depositInterestIndex);
        const stakedAmountValue = (0, mathLib_1.mulScale)(assetStakedAmount, assetPrice, mathLib_1.ONE_10_DP); // 4 d.p.
        const userRewards = [];
        rewards.forEach(({ rewardAssetId, endTimestamp, rewardAssetPrice, rewardInterestRate, rewardPerToken }, localRewardIndex) => {
            const rewardIndex = stakeIndex * 3 + localRewardIndex;
            const oldRewardPerToken = localState.rewardPerTokens[rewardIndex];
            const oldUnclaimedReward = localState.unclaimedRewards[rewardIndex];
            const unclaimedReward = oldUnclaimedReward + (0, mathLib_1.mulScale)(fAssetStakedAmount, rewardPerToken - oldRewardPerToken, mathLib_1.ONE_10_DP);
            const unclaimedRewardValue = (0, mathLib_1.mulScale)(unclaimedReward, rewardAssetPrice, mathLib_1.ONE_10_DP); // 4 d.p.
            userRewards.push({
                rewardAssetId,
                endTimestamp,
                rewardAssetPrice,
                rewardInterestRate,
                unclaimedReward,
                unclaimedRewardValue,
            });
        });
        stakingPrograms.push({
            poolAppId,
            fAssetId,
            fAssetStakedAmount,
            assetId,
            assetPrice,
            assetStakedAmount,
            stakedAmountValue,
            depositInterestRate,
            depositInterestYield,
            rewards: userRewards,
        });
    });
    return {
        currentRound: localState.currentRound,
        userAddress: localState.userAddress,
        escrowAddress: localState.escrowAddress,
        optedIntoAssets: localState.optedIntoAssets,
        stakingPrograms,
    };
}
exports.userDepositStakingInfo = userDepositStakingInfo;
/**
 *
 * Derives loan local state from escrow account.
 *
 * @param state - escrow account local state
 * @param loanAppId - loan application to query about
 * @param escrowAddr - escrow address
 * @returns LoanLocalState loan local state
 */
function loanLocalState(state, loanAppId, escrowAddr) {
    // standard
    const userAddress = (0, algosdk_1.encodeAddress)(Buffer.from(String((0, utils_1.getParsedValueFromState)(state, "u")), "base64"));
    const colPls = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "c")));
    const borPls = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "b")));
    const colBals = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "cb")));
    const borAms = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "ba")));
    const borBals = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "bb")));
    const lbii = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "l")));
    const sbir = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "r")));
    const lsc = (0, utils_1.parseUint64s)(String((0, utils_1.getParsedValueFromState)(state, "t")));
    // custom
    const collaterals = [];
    const borrows = [];
    for (let i = 0; i < 15; i++) {
        // add collateral
        collaterals.push({
            poolAppId: Number(colPls[i]),
            fAssetBalance: colBals[i],
        });
        // add borrow
        borrows.push({
            poolAppId: Number(borPls[i]),
            borrowedAmount: borAms[i],
            borrowBalance: borBals[i],
            latestBorrowInterestIndex: lbii[i],
            stableBorrowInterestRate: sbir[i],
            latestStableChange: lsc[i],
        });
    }
    return {
        userAddress,
        escrowAddress: escrowAddr,
        collaterals,
        borrows,
    };
}
exports.loanLocalState = loanLocalState;
/**
 *
 * Derives user loan info from escrow account.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param localState - local state of escrow account
 * @param poolManagerInfo - pool manager info which is returned by retrievePoolManagerInfo function
 * @param loanInfo - loan info which is returned by retrieveLoanInfo function
 * @param oraclePrices - oracle prices which is returned by getOraclePrices function
 * @returns Promise<UserLoansInfo> user loans info
 */
function userLoanInfo(localState, poolManagerInfo, loanInfo, oraclePrices) {
    const { pools: poolManagerPools } = poolManagerInfo;
    const { pools: loanPools } = loanInfo;
    const { prices } = oraclePrices;
    let netRate = BigInt(0);
    let netYield = BigInt(0);
    // collaterals
    const collaterals = [];
    let totalCollateralBalanceValue = BigInt(0);
    let totalEffectiveCollateralBalanceValue = BigInt(0);
    localState.collaterals.forEach(({ poolAppId, fAssetBalance }) => {
        const isColPresent = poolAppId > 0;
        if (!isColPresent)
            return;
        const poolInfo = poolManagerPools[poolAppId];
        const poolLoanInfo = loanPools[poolAppId];
        if (poolInfo === undefined || poolLoanInfo === undefined)
            throw Error("Could not find collateral pool " + poolAppId);
        const { depositInterestIndex, depositInterestRate, depositInterestYield } = poolInfo;
        const { assetId, collateralFactor } = poolLoanInfo;
        const oraclePrice = prices[assetId];
        if (oraclePrice === undefined)
            throw Error("Could not find asset price " + assetId);
        const { price: assetPrice } = oraclePrice;
        const assetBalance = (0, formulae_1.calcWithdrawReturn)(fAssetBalance, depositInterestIndex);
        const balanceValue = (0, formulae_1.calcCollateralAssetLoanValue)(assetBalance, assetPrice, mathLib_1.ONE_4_DP);
        const effectiveBalanceValue = (0, formulae_1.calcCollateralAssetLoanValue)(assetBalance, assetPrice, collateralFactor);
        totalCollateralBalanceValue += balanceValue;
        totalEffectiveCollateralBalanceValue += effectiveBalanceValue;
        netRate += balanceValue * depositInterestRate;
        netYield += balanceValue * depositInterestYield;
        collaterals.push({
            poolAppId,
            assetId,
            assetPrice,
            depositInterestIndex,
            collateralFactor,
            fAssetBalance,
            assetBalance,
            balanceValue,
            effectiveBalanceValue,
            interestRate: depositInterestRate,
            interestYield: depositInterestYield,
        });
    });
    // borrows
    const borrows = [];
    let totalBorrowedAmountValue = BigInt(0);
    let totalBorrowBalanceValue = BigInt(0);
    let totalEffectiveBorrowBalanceValue = BigInt(0);
    localState.borrows.forEach(({ poolAppId, borrowedAmount, borrowBalance: oldBorrowBalance, latestBorrowInterestIndex, stableBorrowInterestRate, latestStableChange, }) => {
        const isBorPresent = oldBorrowBalance > BigInt(0);
        if (!isBorPresent)
            return;
        const poolInfo = poolManagerPools[poolAppId];
        const poolLoanInfo = loanPools[poolAppId];
        if (poolInfo === undefined || poolLoanInfo === undefined)
            throw Error("Could not find borrow pool " + poolAppId);
        const { assetId, borrowFactor } = poolLoanInfo;
        const oraclePrice = prices[assetId];
        if (oraclePrice === undefined)
            throw Error("Could not find asset price " + assetId);
        const { price: assetPrice } = oraclePrice;
        const isStable = latestStableChange > BigInt(0);
        const bii = isStable
            ? (0, formulae_1.calcBorrowInterestIndex)(stableBorrowInterestRate, latestBorrowInterestIndex, latestStableChange)
            : poolInfo.variableBorrowInterestIndex;
        const borrowedAmountValue = (0, formulae_1.calcCollateralAssetLoanValue)(borrowedAmount, oraclePrice.price, mathLib_1.ONE_4_DP); // no rounding
        const borrowBalance = (0, formulae_1.calcBorrowBalance)(oldBorrowBalance, bii, latestBorrowInterestIndex);
        const borrowBalanceValue = (0, formulae_1.calcBorrowAssetLoanValue)(borrowBalance, assetPrice, mathLib_1.ONE_4_DP);
        const effectiveBorrowBalanceValue = (0, formulae_1.calcBorrowAssetLoanValue)(borrowBalance, assetPrice, borrowFactor);
        const interestRate = isStable ? stableBorrowInterestRate : poolInfo.variableBorrowInterestRate;
        const interestYield = isStable
            ? (0, mathLib_1.compoundEverySecond)(stableBorrowInterestRate, mathLib_1.ONE_16_DP)
            : poolInfo.variableBorrowInterestYield;
        totalBorrowedAmountValue += borrowedAmountValue;
        totalBorrowBalanceValue += borrowBalanceValue;
        totalEffectiveBorrowBalanceValue += effectiveBorrowBalanceValue;
        netRate -= borrowBalanceValue * interestRate;
        netYield -= borrowBalanceValue * interestYield;
        borrows.push({
            poolAppId,
            assetId,
            assetPrice,
            isStable,
            borrowFactor,
            borrowedAmount,
            borrowedAmountValue,
            borrowBalance,
            borrowBalanceValue,
            effectiveBorrowBalanceValue,
            accruedInterest: borrowBalance - borrowedAmount,
            accruedInterestValue: borrowBalanceValue - borrowedAmountValue,
            interestRate,
            interestYield,
        });
    });
    if (totalCollateralBalanceValue > BigInt(0)) {
        netRate /= totalCollateralBalanceValue;
        netYield /= totalCollateralBalanceValue;
    }
    // combine
    return {
        currentRound: localState.currentRound,
        userAddress: localState.userAddress,
        escrowAddress: localState.escrowAddress,
        collaterals,
        borrows,
        netRate,
        netYield,
        totalCollateralBalanceValue,
        totalBorrowedAmountValue,
        totalBorrowBalanceValue,
        totalEffectiveCollateralBalanceValue,
        totalEffectiveBorrowBalanceValue,
        loanToValueRatio: (0, formulae_1.calcLTVRatio)(totalBorrowBalanceValue, totalCollateralBalanceValue),
        borrowUtilisationRatio: (0, formulae_1.calcBorrowUtilisationRatio)(totalEffectiveBorrowBalanceValue, totalEffectiveCollateralBalanceValue),
        liquidationMargin: (0, formulae_1.calcLiquidationMargin)(totalEffectiveBorrowBalanceValue, totalEffectiveCollateralBalanceValue),
    };
}
exports.userLoanInfo = userLoanInfo;
//# sourceMappingURL=utils.js.map