import { Indexer, SuggestedParams, Transaction } from "algosdk";
import { Dispenser, Distributor } from "../common";
import { DistributorInfo, UserCommitmentInfo } from "./types";
/**
 *
 * Returns information regarding the given liquid governance distributor.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param distributor - distributor to query about
 * @returns DistributorInfo[] distributor info
 */
declare function getDistributorInfo(indexerClient: Indexer, distributor: Distributor): Promise<DistributorInfo>;
/**
 *
 * Returns information regarding a user's liquid governance commitment.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param distributor - distributor to query about
 * @param userAddr - user address to get governance info about
 * @returns UserCommitmentInfo[] user commitment info
 */
declare function getUserLiquidGovernanceInfo(indexerClient: Indexer, distributor: Distributor, userAddr: string): Promise<UserCommitmentInfo>;
/**
 *
 * Returns a group transaction to mint gALGO for ALGO at a one-to-one rate.
 * If in the commitment period then also commits user into governance.
 *
 * @param dispenser - dispenser to mint gALGO from
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to send and gALGO to mint
 * @param includeOptIn - whether to include an opt in transaction (must be opted in if minting in commitment period)
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the minter (must pass to be eligible for revenue share)
 * @returns Transaction[] mint transactions
 */
declare function prepareMintTransactions(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, includeOptIn: boolean, params: SuggestedParams, note?: Uint8Array): Transaction[];
/**
 *
 * Returns a transaction to unmint pre-minted gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to unmint and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction unmint pre-mint transaction
 */
declare function prepareUnmintPremintTransaction(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, params: SuggestedParams, note?: Uint8Array): Transaction;
/**
 *
 * Returns a group transaction to unmint gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to send and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction[] unmint transactions
 */
declare function prepareUnmintTransactions(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, params: SuggestedParams, note?: Uint8Array): Transaction[];
/**
 *
 * Returns a transaction to claim pre-minted gALGO.
 * Can be called on behalf of yourself or another user.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param receiverAddr - account address for the pre-minter that will receiver the gALGO
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction claim pre-mint transaction
 */
declare function prepareClaimPremintTransaction(dispenser: Dispenser, distributor: Distributor, senderAddr: string, receiverAddr: string, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to burn gALGO for ALGO at a one-to-one rate.
 * Must be after period end.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to send and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] burn transactions
 */
declare function prepareBurnTransactions(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a group transaction to early claim governance rewards for a given amount of ALGO.
 * Must be after commitment end.
 * Rewards received is in gALGO. Amount of rewards is determined by rewards_per_algo.
 *
 * @param dispenser - distributor to receive gALGO from
 * @param distributor - distributor which has sender's commitment
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to early claim rewards on
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction early claim governance rewards transaction
 */
declare function prepareEarlyClaimGovernanceRewardsTransaction(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to claim governance rewards for unclaimed commitment.
 * Must be after period end and rewards have been distributed from Algorand Foundation.
 * Rewards received is in ALGO. Amount of rewards is determined by rewards_per_algo.
 *
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction claim governance rewards transaction
 */
declare function prepareClaimGovernanceRewardsTransaction(distributor: Distributor, senderAddr: string, params: SuggestedParams): Transaction;
export { getDistributorInfo, getUserLiquidGovernanceInfo, prepareMintTransactions, prepareUnmintPremintTransaction, prepareUnmintTransactions, prepareClaimPremintTransaction, prepareBurnTransactions, prepareEarlyClaimGovernanceRewardsTransaction, prepareClaimGovernanceRewardsTransaction, };
//# sourceMappingURL=governance.d.ts.map