/// <reference types="node" />
import { Algodv2, Indexer, LogicSigAccount, SuggestedParams, Transaction } from "algosdk";
import { Dispenser, Distributor } from "../common";
import { DistributorInfo, EscrowGovernanceStatus, UserCommitmentInfo } from "./types";
declare function getDistributorLogicSig(userAddr: string): LogicSigAccount;
/**
 *
 * Returns information regarding the given liquid governance distributor.
 *
 * @param client - Algorand client to query
 * @param distributor - distributor to query about
 * @returns DistributorInfo[] distributor info
 */
declare function getDistributorInfo(client: Algodv2 | Indexer, distributor: Distributor): Promise<DistributorInfo>;
/**
 *
 * Returns information regarding a user's liquid governance commitment.
 *
 * @param client - Algorand client to query
 * @param distributor - distributor to query about
 * @param userAddr - user address to get governance info about
 * @returns UserCommitmentInfo user commitment info
 */
declare function getUserLiquidGovernanceInfo(client: Algodv2 | Indexer, distributor: Distributor, userAddr: string): Promise<UserCommitmentInfo>;
/**
 *
 * Returns information regarding a user's escrow governance status.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param userAddr - user address to get governance info about
 * @param signUpAddr - sign up address for the governance period
 * @returns EscrowGovernanceStatus escrow governance status
 */
declare function getEscrowGovernanceStatus(indexerClient: Indexer, userAddr: string, signUpAddr: string): Promise<EscrowGovernanceStatus>;
/**
 *
 * Returns a group transaction to add liquid governance escrow.
 *
 * @param distributor - distributor that adding escrow to
 * @param userAddr - account address for the user
 * @param delegatable - whether governance action for escrow can be delegated to gov admin
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: LogicSigAccount } object containing group transaction and generated escrow account
 */
declare function prepareAddLiquidGovernanceEscrowTransactions(distributor: Distributor, userAddr: string, delegatable: boolean, params: SuggestedParams): {
    txns: Transaction[];
    escrow: LogicSigAccount;
};
/**
 *
 * Returns a group transaction to mint gALGO for ALGO at a one-to-one rate.
 *
 * If you want to commit then MUST be paired with prepareCommitOrVoteTransaction
 * as to send the required note to the signup address.
 *
 * @param dispenser - dispenser to mint gALGO from
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to send and gALGO to mint
 * @param ensureCommit - whether to ensure in commitment period
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the minter (must pass to be eligible for revenue share)
 * @returns Transaction[] mint transactions
 */
declare function prepareMintTransactions(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, ensureCommit: boolean, params: SuggestedParams, note?: Uint8Array): Transaction[];
/**
 *
 * Returns a transaction to unmint pre-minted gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to unmint and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction unmint pre-mint transaction
 */
declare function prepareUnmintPremintTransaction(distributor: Distributor, senderAddr: string, amount: number | bigint, params: SuggestedParams, note?: Uint8Array): Transaction[];
/**
 *
 * Returns a transaction to unmint pre-minted gALGO for ALGO at a one-to-one rate.
 * Must be in commitment period. By unminting, you will lose your governance rewards.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to unmint and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note to distinguish who is the unminter (must pass to be eligible for revenue share)
 * @returns Transaction[] unmint transactions
 */
declare function prepareUnmintTransactions(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, params: SuggestedParams, note?: Uint8Array): Transaction[];
/**
 *
 * Returns a transaction to claim pre-minted gALGO.
 * Can be called on behalf of yourself or another user.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor to receive ALGO from
 * @param senderAddr - account address for the sender
 * @param receiverAddr - account address for the pre-minter that will receiver the gALGO
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction claim pre-mint transaction
 */
declare function prepareClaimPremintTransaction(dispenser: Dispenser, distributor: Distributor, senderAddr: string, receiverAddr: string, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to register escrow online.
 *
 * @param distributor - distributor that has escrow
 * @param senderAddr - account address for the sender
 * @param voteKey - vote key
 * @param selectionKey - selection key
 * @param stateProofKey - state proof key
 * @param voteFirstRound - vote first round
 * @param voteLastRound - vote last round
 * @param voteKeyDilution - vote key dilution
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction register online transaction
 */
declare function prepareRegisterEscrowOnlineTransaction(distributor: Distributor, senderAddr: string, voteKey: Buffer, selectionKey: Buffer, stateProofKey: Buffer, voteFirstRound: number | bigint, voteLastRound: number | bigint, voteKeyDilution: number | bigint, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to register escrow offline.
 *
 * @param distributor - distributor that has escrow
 * @param senderAddr - account address for the sender
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction register offline transaction
 */
declare function prepareRegisterEscrowOfflineTransaction(distributor: Distributor, senderAddr: string, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to commit or vote in governance.
 *
 * @param distributor - distributor that has escrow
 * @param senderAddr - account address for the sender
 * @param destAddr - destination address
 * @param note - note to send
 * @param params - suggested params the transactions with the fees overwritten
 * @returns Transaction commit/vote transaction
 */
declare function prepareCommitOrVoteTransaction(distributor: Distributor, senderAddr: string, destAddr: string, note: string, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to remove escrow from distributor.
 * Must have zero balance or be after period end.
 *
 * @param distributor - distributor that removing escrow from
 * @param senderAddr - account address for the sender
 * @param ownerAddr - account address for the owner of the escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] burn transactions
 */
declare function prepareRemoveLiquidGovernanceEscrowTransactions(distributor: Distributor, senderAddr: string, ownerAddr: string, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a group transaction to burn gALGO for ALGO at a one-to-one rate.
 * Must be after period end.
 *
 * @param dispenser - dispenser to send gALGO to
 * @param distributor - distributor that calls dispenser and to send ALGO to
 * @param senderAddr - account address for the sender
 * @param amount - amount of gALGO to send and ALGO to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] burn transactions
 */
declare function prepareBurnTransactions(dispenser: Dispenser, distributor: Distributor, senderAddr: string, amount: number | bigint, params: SuggestedParams): Transaction[];
export { getDistributorLogicSig, getDistributorInfo, getUserLiquidGovernanceInfo, getEscrowGovernanceStatus, prepareAddLiquidGovernanceEscrowTransactions, prepareMintTransactions, prepareUnmintPremintTransaction, prepareUnmintTransactions, prepareClaimPremintTransaction, prepareRegisterEscrowOnlineTransaction, prepareRegisterEscrowOfflineTransaction, prepareCommitOrVoteTransaction, prepareRemoveLiquidGovernanceEscrowTransactions, prepareBurnTransactions, };
//# sourceMappingURL=governance.d.ts.map