import { Account, Algodv2, Indexer, SuggestedParams, Transaction } from "algosdk";
import { UserDepositFullInfo, Oracle, Pool, PoolInfo, PoolManagerInfo, UserDepositInfo } from "./types";
/**
 *
 * Returns information regarding the given pool manager.
 *
 * @param client - Algorand client to query
 * @param poolManagerAppId - pool manager application to query about
 * @returns Promise<PoolManagerInfo[]> pool manager info
 */
declare function retrievePoolManagerInfo(client: Algodv2 | Indexer, poolManagerAppId: number): Promise<PoolManagerInfo>;
/**
 *
 * Returns information regarding the given pool.
 *
 * @param client - Algorand client to query
 * @param pool - pool application to query about
 * @returns Promise<PoolInfo> pool info
 */
declare function retrievePoolInfo(client: Algodv2 | Indexer, pool: Pool): Promise<PoolInfo>;
/**
 *
 * Returns basic information regarding the given user's deposit escrows.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param depositsAppId - deposits application to query about
 * @param userAddr - account address for the user
 * @returns Promise<UserDepositInfo[]> user deposits info
 */
declare function retrieveUserDepositsInfo(indexerClient: Indexer, depositsAppId: number, userAddr: string): Promise<UserDepositInfo[]>;
/**
 *
 * Returns full information regarding the given user's deposit escrows.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param poolManagerAppId - pool manager application to query about
 * @param depositsAppId - deposits application to query about
 * @param pools - pools in pool manager (either MainnetPools or TestnetPools)
 * @param oracle - oracle to query
 * @param userAddr - account address for the user
 * @returns Promise<UserDepositFullInfo[]> user deposits full info
 */
declare function retrieveUserDepositsFullInfo(indexerClient: Indexer, poolManagerAppId: number, depositsAppId: number, pools: Record<string, Pool>, oracle: Oracle, userAddr: string): Promise<UserDepositFullInfo[]>;
/**
 *
 * Returns information regarding the given user's deposit escrows.
 *
 * @param client - Algorand client to query
 * @param depositsAppId - deposits application to query about
 * @param escrowAddr - account address for the deposit escrow
 * @returns Promise<UserDepositInfo> user deposit info
 */
declare function retrieveUserDepositInfo(client: Algodv2 | Indexer, depositsAppId: number, escrowAddr: string): Promise<UserDepositInfo>;
/**
 *
 * Returns a group transaction to add escrow before depositing.
 *
 * @param depositsAppId - deposits application to add an escrow for
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: Account } object containing group transaction and generated escrow account
 */
declare function prepareAddDepositEscrowToDeposits(depositsAppId: number, userAddr: string, params: SuggestedParams): {
    txns: Transaction[];
    escrow: Account;
};
/**
 *
 * Returns a transaction to opt deposits escrow into asset so that it can hold a given pool's f asset.
 *
 * @param depositsAppId - deposits application of escrow
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param pool - pool to add f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt deposit escrow into asset transaction
 */
declare function prepareOptDepositEscrowIntoAssetInDeposits(depositsAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, pool: Pool, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to deposit asset into given pool.
 *
 * @param pool - pool application to deposit into
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the deposit (typically the user's deposit escrow or loan escrow)
 * @param assetAmount - the asset amount to deposit
 * @param params - suggested params for the transactions with the fees overwritten
 * @param note - optional note field to add
 * @returns Transaction[] deposit asset group transaction
 */
declare function prepareDepositIntoPool(pool: Pool, poolManagerAppId: number, userAddr: string, receiverAddr: string, assetAmount: number | bigint, params: SuggestedParams, note?: Uint8Array): Transaction[];
/**
 *
 * Returns a transaction to withdraw from a deposits escrow
 *
 * @param depositsAppId - deposits application of escrow
 * @param pool - pool to withdraw from
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param amount - the amount of asset / f asset to send to withdraw from escrow.
 * @param isfAssetAmount - whether the amount to withdraw is expressed in terms of f asset or asset
 * @param remainDeposited - whether receiver should get f asset or asset (cannot remain deposited and use asset amount)
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction withdraw from deposit escrow transaction
 */
declare function prepareWithdrawFromDepositEscrowInDeposits(depositsAppId: number, pool: Pool, poolManagerAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, amount: number | bigint, isfAssetAmount: boolean, remainDeposited: boolean, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to withdraw from a user's wallet
 *
 * @param pool - pool to withdraw from
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param fAssetAmount - the amount of f asset to send to the pool
 * @param receivedAssetAmount - the amount of asset to receive. Any excess f asset sent will be returned to the deposit escrow. If zero then interpreted as variable withdrawal.
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] withdraw from user's wallet group transaction
 */
declare function prepareWithdrawFromPool(pool: Pool, poolManagerAppId: number, userAddr: string, receiverAddr: string, fAssetAmount: number | bigint, receivedAssetAmount: number | bigint, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a transaction to update a pool's interest indexes
 *
 * @param pool - pool to update
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] update pool's interest indexes transaction
 */
declare function prepareUpdatePoolInterestIndexes(pool: Pool, poolManagerAppId: number, userAddr: string, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to remove asset from deposit escrow
 *
 * @param depositsAppId - deposits application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param pool - pool to remove f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt out deposit escrow from asset transaction
 */
declare function prepareOptOutDepositEscrowFromAssetInDeposits(depositsAppId: number, userAddr: string, escrowAddr: string, pool: Pool, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to remove a user's deposit escrow and return its minimum balance to the user.
 *
 * @param depositsAppId - deposits application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] remove and close out deposit escrow group transaction
 */
declare function prepareRemoveDepositEscrowFromDeposits(depositsAppId: number, userAddr: string, escrowAddr: string, params: SuggestedParams): Transaction[];
export { retrievePoolManagerInfo, retrievePoolInfo, retrieveUserDepositsInfo, retrieveUserDepositsFullInfo, retrieveUserDepositInfo, prepareAddDepositEscrowToDeposits, prepareOptDepositEscrowIntoAssetInDeposits, prepareDepositIntoPool, prepareWithdrawFromDepositEscrowInDeposits, prepareWithdrawFromPool, prepareUpdatePoolInterestIndexes, prepareOptOutDepositEscrowFromAssetInDeposits, prepareRemoveDepositEscrowFromDeposits, };
//# sourceMappingURL=deposit.d.ts.map