import { Account, Algodv2, Indexer, SuggestedParams, Transaction } from "algosdk";
import { DepositStakingInfo, Pool, UserDepositStakingLocalState } from "./types";
import { depositStakingProgramsInfo, userDepositStakingInfo } from "./utils";
/**
 *
 * Returns information regarding the given deposit staking application.
 *
 * @param client - Algorand client to query
 * @param depositStakingAppId - deposit staking application to query about
 * @returns Promise<DepositStakingInfo> pool info
 */
declare function retrieveDepositStakingInfo(client: Algodv2 | Indexer, depositStakingAppId: number): Promise<DepositStakingInfo>;
/**
 *
 * Returns local state regarding the deposit staking escrows of a given user.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param depositStakingAppId - deposit staking application to query about
 * @param userAddr - account address for the user
 * @returns Promise<UserDepositStakingLocalState[]> deposit staking escrows' local state
 */
declare function retrieveUserDepositStakingsLocalState(indexerClient: Indexer, depositStakingAppId: number, userAddr: string): Promise<UserDepositStakingLocalState[]>;
/**
 *
 * Returns local state regarding the deposit staking escrows of a given user.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param depositStakingAppId - deposit staking application to query about
 * @param escrowAddr - account address for the deposit staking escrow
 * @returns Promise<UserDepositStakingLocalState> deposit staking escrows' local state
 */
declare function retrieveUserDepositStakingLocalState(indexerClient: Indexer, depositStakingAppId: number, escrowAddr: string): Promise<UserDepositStakingLocalState>;
/**
 *
 * Returns a group transaction to deposit staking escrow.
 *
 * @param depositStakingAppId - deposit staking application to query about
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: Account } object containing group transaction and generated escrow account
 */
declare function prepareAddDepositStakingEscrow(depositStakingAppId: number, userAddr: string, params: SuggestedParams): {
    txns: Transaction[];
    escrow: Account;
};
/**
 *
 * Returns a transaction to opt deposit staking escrow into asset so that it can hold a given pool's f asset.
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param pool - pool to add f asset of
 * @param stakeIndex - staking program index
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt deposit staking escrow into asset transaction
 */
declare function prepareOptDepositStakingEscrowIntoAsset(depositStakingAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, pool: Pool, stakeIndex: number, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to sync stake of deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param pool - pool to sync stake of
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param stakeIndex - staking program index
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction sync stake of deposit staking escrow transaction
 */
declare function prepareSyncStakeInDepositStakingEscrow(depositStakingAppId: number, pool: Pool, userAddr: string, escrowAddr: string, stakeIndex: number, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to claim the rewards of a deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param stakeIndex - staking program index
 * @param rewardAssetIds - the asset ids of all the rewards assets claiming
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction withdraw from deposit staking escrow transaction
 */
declare function prepareClaimRewardsOfDepositStakingEscrow(depositStakingAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, stakeIndex: number, rewardAssetIds: number[], params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to withdraw from a deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param pool - pool to withdraw from
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param receiverAddr - account address to receive the withdrawal (typically the same as the user address)
 * @param amount - the amount of asset / f asset to send to withdraw from escrow.
 * @param isfAssetAmount - whether the amount to withdraw is expressed in terms of f asset or asset
 * @param remainDeposited - whether receiver should get f asset or asset (cannot remain deposited and use asset amount)
 * @param stakeIndex - staking program index
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction withdraw from deposit staking escrow transaction
 */
declare function prepareWithdrawFromDepositStakingEscrow(depositStakingAppId: number, pool: Pool, poolManagerAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, amount: number | bigint, isfAssetAmount: boolean, remainDeposited: boolean, stakeIndex: number, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to remove asset from deposit staking escrow
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param pool - pool to remove f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction opt out deposit staking escrow from asset transaction
 */
declare function prepareOptOutDepositStakingEscrowFromAsset(depositStakingAppId: number, userAddr: string, escrowAddr: string, pool: Pool, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to remove a user's deposit staking escrow and return its minimum balance to the user.
 *
 * @param depositStakingAppId - deposit staking application of escrow
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the deposit staking escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] remove and close out deposit staking escrow group transaction
 */
declare function prepareRemoveDepositStakingEscrow(depositStakingAppId: number, userAddr: string, escrowAddr: string, params: SuggestedParams): Transaction[];
export { retrieveDepositStakingInfo, depositStakingProgramsInfo, retrieveUserDepositStakingsLocalState, retrieveUserDepositStakingLocalState, userDepositStakingInfo, prepareAddDepositStakingEscrow, prepareOptDepositStakingEscrowIntoAsset, prepareSyncStakeInDepositStakingEscrow, prepareClaimRewardsOfDepositStakingEscrow, prepareWithdrawFromDepositStakingEscrow, prepareOptOutDepositStakingEscrowFromAsset, prepareRemoveDepositStakingEscrow, };
//# sourceMappingURL=depositStaking.d.ts.map