/**
 * Calculates the dollar value of a given asset amount
 * @param amount (0dp)
 * @param price (14dp)
 * @return value (0dp)
 */
declare function calcAssetDollarValue(amount: bigint, price: bigint): bigint;
/**
 * Calculates the total debt of a pool
 * @param totalVarDebt (0dp)
 * @param totalStblDebt (0dp)
 * @return totalDebt (0dp)
 */
declare function calcTotalDebt(totalVarDebt: bigint, totalStblDebt: bigint): bigint;
/**
 * Calculates the total debt of a pool
 * @param totalDebt (0dp)
 * @param totalDeposits (0dp)
 * @return availableLiquidity (0dp)
 */
declare function calcAvailableLiquidity(totalDebt: bigint, totalDeposits: bigint): bigint;
/**
 * Calculates the ratio of the available liquidity that is being stable borrowed
 * @param stblBorAmount (0dp)
 * @param availableLiquidity (0dp)
 * @return stableBorrowRatio (16dp)
 */
declare function calcStableBorrowRatio(stblBorAmount: bigint, availableLiquidity: bigint): bigint;
/**
 * Calculates the maximum stable borrow amount a user can make in one go
 * @param availableLiquidity (0dp)
 * @param sbpc (0dp)
 * @return stableBorrowRatio (16dp)
 */
declare function calcMaxSingleStableBorrow(availableLiquidity: bigint, sbpc: bigint): bigint;
/**
 * Calculates the utilisation ratio of a pool
 * @param totalDebt (0dp)
 * @param totalDeposits (0dp)
 * @return utilisationRatio (16dp)
 */
declare function calcUtilisationRatio(totalDebt: bigint, totalDeposits: bigint): bigint;
/**
 * Calculates the stable debt to total debt ratio of a pool
 * @param totalStblDebt (0dp)
 * @param totalDebt (0dp)
 * @return stableDebtToTotalDebtRatio (16dp)
 */
declare function calcStableDebtToTotalDebtRatio(totalStblDebt: bigint, totalDebt: bigint): bigint;
/**
 * Calculate the variable borrow interest rate of a pool
 * @param vr0 (16dp)
 * @param vr1 (16dp)
 * @param vr2 (16dp)
 * @param ut (16dp)
 * @param uopt (16dp)
 * @return variableBorrowInterestRate (16dp)
 */
declare function calcVariableBorrowInterestRate(vr0: bigint, vr1: bigint, vr2: bigint, ut: bigint, uopt: bigint): bigint;
/**
 * Calculate the stable borrow interest rate of a pool
 * @param vr1 (16dp)
 * @param sr0 (16dp)
 * @param sr1 (16dp)
 * @param sr2 (16dp)
 * @param sr3 (16dp)
 * @param ut (16dp)
 * @param uopt (16dp)
 * @param ratiot (16dp)
 * @param ratioopt (16dp)
 * @return stableBorrowInterestRate (16dp)
 */
declare function calcStableBorrowInterestRate(vr1: bigint, sr0: bigint, sr1: bigint, sr2: bigint, sr3: bigint, ut: bigint, uopt: bigint, ratiot: bigint, ratioopt: bigint): bigint;
/**
 * Calculate the overall borrow interest rate of a pool
 * @param totalVarDebt (0dp)
 * @param totalDebt (0dp)
 * @param vbirt (16dp)
 * @param osbiat (16dp)
 * @return overallBorrowInterestRate (16dp)
 */
declare function calcOverallBorrowInterestRate(totalVarDebt: bigint, totalDebt: bigint, vbirt: bigint, osbiat: bigint): bigint;
/**
 * Calculate the deposit interest rate of a pool
 * @param obirt (16dp)
 * @param ut (16dp)
 * @param rr (16dp)
 * @return overallBorrowInterestRate (16dp)
 */
declare function calcDepositInterestRate(obirt: bigint, rr: bigint, ut: bigint): bigint;
/**
 * Calculate the borrow interest index of pool
 * @param birt1 (16dp)
 * @param biit1 (16dp)
 * @param latestUpdate (0dp)
 * @return borrowInterestIndex (14dp)
 */
declare function calcBorrowInterestIndex(birt1: bigint, biit1: bigint, latestUpdate: bigint): bigint;
/**
 * Calculate the deposit interest index of pool
 * @param dirt1 (16dp)
 * @param diit1 (16dp)
 * @param latestUpdate (0dp)
 * @return depositInterestIndex (14dp)
 */
declare function calcDepositInterestIndex(dirt1: bigint, diit1: bigint, latestUpdate: bigint): bigint;
/**
 * Calculates the fAsset received from a deposit
 * @param depositAmount (0dp)
 * @param diit (14dp)
 * @return depositReturn (0dp)
 */
declare function calcDepositReturn(depositAmount: bigint, diit: bigint): bigint;
/**
 * Calculates the asset received from a withdraw
 * @param withdrawAmount (0dp)
 * @param diit (14dp)
 * @return withdrawReturn (0dp)
 */
declare function calcWithdrawReturn(withdrawAmount: bigint, diit: bigint): bigint;
/**
 * Calculates the collateral asset loan value
 * @param amount (0dp)
 * @param price (14dp)
 * @param factor (4dp)
 * @return loanValue (4dp)
 */
declare function calcCollateralAssetLoanValue(amount: bigint, price: bigint, factor: bigint): bigint;
/**
 * Calculates the borrow asset loan value
 * @param amount (0dp)
 * @param price (14dp)
 * @param factor (4dp)
 * @return loanValue (4dp)
 */
declare function calcBorrowAssetLoanValue(amount: bigint, price: bigint, factor: bigint): bigint;
/**
 * Calculates the loan's LTV ratio
 * @param totalBorrowBalanceValue (4dp)
 * @param totalCollateralBalanceValue (4dp)
 * @return LTVRatio (4dp)
 */
declare function calcLTVRatio(totalBorrowBalanceValue: bigint, totalCollateralBalanceValue: bigint): bigint;
/**
 * Calculates the loan's borrow utilisation ratio
 * @param totalEffectiveBorrowBalanceValue (4dp)
 * @param totalEffectiveCollateralBalanceValue (4dp)
 * @return borrowUtilisationRatio (4dp)
 */
declare function calcBorrowUtilisationRatio(totalEffectiveBorrowBalanceValue: bigint, totalEffectiveCollateralBalanceValue: bigint): bigint;
/**
 * Calculates the loan's liquidation margin
 * @param totalEffectiveBorrowBalanceValue (4dp)
 * @param totalEffectiveCollateralBalanceValue (4dp)
 * @return liquidationMargin (4dp)
 */
declare function calcLiquidationMargin(totalEffectiveBorrowBalanceValue: bigint, totalEffectiveCollateralBalanceValue: bigint): bigint;
/**
 * Calculates the borrow balance of the loan at time t
 * @param bbtn1 (0dp)
 * @param biit (14dp)
 * @param biitn1 (14dp)
 * @return borrowBalance (0dp)
 */
declare function calcBorrowBalance(bbtn1: bigint, biit: bigint, biitn1: bigint): bigint;
/**
 * Calculates the stable borrow interest rate of the loan after a borrow increase
 * @param bbt (0dp)
 * @param amount (0dp)
 * @param sbirtn1 (16dp)
 * @param sbirt1 (16dp)
 * @return stableInterestRate (16dp)
 */
declare function calcLoanStableInterestRate(bbt: bigint, amount: bigint, sbirtn1: bigint, sbirt1: bigint): bigint;
/**
 * Calculates the deposit interest rate condition required to rebalance up stable borrow.
 * Note that there is also a second condition on the pool utilisation ratio.
 * @param rudir (16dp)
 * @param vr0 (16dp)
 * @param vr1 (16dp)
 * @param vr2 (16dp)
 * @return rebalanceUpThreshold (16dp)
 */
declare function calcRebalanceUpThreshold(rudir: bigint, vr0: bigint, vr1: bigint, vr2: bigint): bigint;
/**
 * Calculates the stable interest rate condition required to rebalance down stable borrow
 * @param rdd (16dp)
 * @param sbirt1 (16dp)
 * @return rebalanceDownThreshold (16dp)
 */
declare function calcRebalanceDownThreshold(rdd: bigint, sbirt1: bigint): bigint;
/**
 * Calculates the flash loan repayment amount for a given borrow amount and fee
 * @param borrowAmount (0dp)
 * @param fee (16dp)
 * @return repaymentAmount (0dp)
 */
declare function calcFlashLoanRepayment(borrowAmount: bigint, fee: bigint): bigint;
/**
 * Calculates the LP price
 * @param r0 pool supply of asset 0
 * @param r1 pool supply of asset 1
 * @param p0 price of asset 0
 * @param p1 price of asset 1
 * @param lts circulating supply of liquidity token
 * @return bigint LP price
 */
declare function calcLPPrice(r0: bigint, r1: bigint, p0: bigint, p1: bigint, lts: bigint): bigint;
export { calcAssetDollarValue, calcTotalDebt, calcAvailableLiquidity, calcStableBorrowRatio, calcMaxSingleStableBorrow, calcUtilisationRatio, calcStableDebtToTotalDebtRatio, calcVariableBorrowInterestRate, calcStableBorrowInterestRate, calcOverallBorrowInterestRate, calcDepositInterestRate, calcBorrowInterestIndex, calcDepositInterestIndex, calcDepositReturn, calcWithdrawReturn, calcCollateralAssetLoanValue, calcBorrowAssetLoanValue, calcLTVRatio, calcBorrowUtilisationRatio, calcLiquidationMargin, calcBorrowBalance, calcLoanStableInterestRate, calcRebalanceUpThreshold, calcRebalanceDownThreshold, calcFlashLoanRepayment, calcLPPrice, };
//# sourceMappingURL=formulae.d.ts.map