import { Account, Algodv2, Indexer, SuggestedParams, Transaction } from "algosdk";
import { LoanInfo, LoanLocalState, LPToken, Oracle, OraclePrices, Pool, PoolManagerInfo, ReserveAddress, UserLoanInfo } from "./types";
import { userLoanInfo } from "./utils";
/**
 *
 * Returns information regarding the given pool.
 *
 * @param client - Algorand client to query
 * @param loanAppId - loan application to query about
 * @returns Promise<LoanInfo[]> loan info
 */
declare function retrieveLoanInfo(client: Algodv2 | Indexer, loanAppId: number): Promise<LoanInfo>;
/**
 *
 * Returns local state regarding the loan escrows of a given user.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param userAddr - account address for the user
 * @returns Promise<LoanLocalState[]> loan escrows' local state
 */
declare function retrieveLoansLocalState(indexerClient: Indexer, loanAppId: number, userAddr: string): Promise<LoanLocalState[]>;
/**
 *
 * Returns local state of given escrow.
 * Use for advanced use cases where optimising number of network request.
 *
 * @param client - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param escrowAddr - account address for the loan escrow
 * @returns Promise<LoanLocalState> loan escrow local state
 */
declare function retrieveLoanLocalState(client: Algodv2 | Indexer, loanAppId: number, escrowAddr: string): Promise<LoanLocalState>;
/**
 *
 * Returns information regarding the loan escrows of a given user.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param poolManagerAppId - pool manager application to query about
 * @param oracle - oracle to query
 * @param userAddr - account address for the user
 * @returns Promise<UserLoanInfo[]> user loans infos
 */
declare function retrieveUserLoansInfo(indexerClient: Indexer, loanAppId: number, poolManagerAppId: number, oracle: Oracle, userAddr: string): Promise<UserLoanInfo[]>;
/**
 *
 * Returns information regarding the given user loan escrow.
 *
 * @param client - Algorand client to query
 * @param loanAppId - loan application to query about
 * @param poolManagerAppId - pool manager application to query about
 * @param oracle - oracle to query
 * @param escrowAddr - account address for the loan escrow
 * @returns Promise<UserLoanInfo> user loan info
 */
declare function retrieveUserLoanInfo(client: Algodv2 | Indexer, loanAppId: number, poolManagerAppId: number, oracle: Oracle, escrowAddr: string): Promise<UserLoanInfo>;
/**
 *
 * Returns all loans that are liquidatable.
 *
 * @param indexerClient - Algorand indexer client to query
 * @param loanAppId - loan application to query about
 * @param poolManagerInfo - pool manager info which is returned by retrievePoolManagerInfo function
 * @param loanInfo - loan info which is returned by retrieveLoanInfo function
 * @param oraclePrices - oracle prices which is returned by getOraclePrices function
 * @param nextToken - token for retrieving next escrows
 * @returns Promise<{ loans: UserLoanInfo[], nextToken?: string}> object containing liquidatable loans and next token
 */
declare function retrieveLiquidatableLoans(indexerClient: Indexer, loanAppId: number, poolManagerInfo: PoolManagerInfo, loanInfo: LoanInfo, oraclePrices: OraclePrices, nextToken?: string): Promise<{
    loans: UserLoanInfo[];
    nextToken?: string;
}>;
/**
 *
 * Returns the maximum reduce collateral of a given f asset considering a target borrow utilisation ratio.
 * Returns 0 if no collateral in loan.
 * Returns at most the collateral balance.
 *
 * @param loan - user loan
 * @param colPoolAppId - price of asset borrowing 14 d.p.
 * @param targetBorrowUtilisationRatio - the utilisation ratio that you are targeting 4 d.p.
 * @returns bigint max f asset amount
 */
declare function getMaxReduceCollateralForBorrowUtilisationRatio(loan: UserLoanInfo, colPoolAppId: number, targetBorrowUtilisationRatio: bigint): bigint;
/**
 *
 * Returns the maximum borrow of a given asset considering a target borrow utilisation ratio.
 * Returns 0 if cannot borrow anything more.
 *
 * @param loan - user loan
 * @param assetPrice - price of asset borrowing 14 d.p.
 * @param borrowFactor - borrow factor of asset borrowing 4 d.p.
 * @param targetBorrowUtilisationRatio - the utilisation ratio that you are targeting 4 d.p.
 * @returns bigint max asset amount
 */
declare function getMaxBorrowForBorrowUtilisationRatio(loan: UserLoanInfo, assetPrice: bigint, borrowFactor: bigint, targetBorrowUtilisationRatio: bigint): bigint;
/**
 *
 * Returns a group transaction to create loan escrow.
 *
 * @param loanAppId - loan application to add escrow for
 * @param userAddr - account address for the user
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns { txns: Transaction[], escrow: Account } object containing group transaction and generated escrow account
 */
declare function prepareCreateUserLoan(loanAppId: number, userAddr: string, params: SuggestedParams): {
    txns: Transaction[];
    escrow: Account;
};
/**
 *
 * Returns a transaction to add a collateral to a loan escrow.
 *
 * @param loanAppId - loan application to add collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to add f asset of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction add collateral to loan escrow transaction
 */
declare function prepareAddCollateralToLoan(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, pool: Pool, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to sync collateral of a loan escrow.
 *
 * @param loanAppId - loan application to sync collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to sync collateral of
 * @param oracle - oracle application to retrieve asset price from
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] sync collateral group transaction
 */
declare function prepareSyncCollateralInLoan(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, pool: Pool, oracle: Oracle, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a transaction to sync collateral of a loan escrow.
 *
 * @param loanAppId - loan application to reduce collateral in
 * @param poolManagerAppId - pool manager application*
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the collateral (typically the user's deposit escrow)
 * @param pool - pool to reduce collateral of
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param amount - the amount of asset / f asset to reduce the collateral by
 * @param isfAssetAmount - whether the amount of collateral to reduce by is expressed in terms of f asset or asset
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] reduce collateral group transaction
 */
declare function prepareReduceCollateralFromLoan(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, pool: Pool, oracle: Oracle, lpAssets: LPToken[], baseAssetIds: number[], amount: number | bigint, isfAssetAmount: boolean, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a transaction to begin swap collateral in a loan escrow.
 * Must be groped together with swap_collateral_end call.
 *
 * @param loanAppId - loan application to swap collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the collateral (typically the user's deposit escrow)
 * @param pool - pool to swap collateral of
 * @param amount - the amount of asset / f asset to reduce the collateral by
 * @param isfAssetAmount - whether the amount of collateral to reduce by is expressed in terms of f asset or asset
 * @param txnIndexForSwapCollateralEnd - transaction index in the group transaction for the swap_collateral_end call.
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction swap collateral begin transaction
 */
declare function prepareSwapCollateralInLoanBegin(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, pool: Pool, amount: number | bigint, isfAssetAmount: boolean, txnIndexForSwapCollateralEnd: number, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to end swap collateral in a loan escrow.
 * Must be groped together with swap_collateral_begin call.
 *
 * @param loanAppId - loan application to swap collateral in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] swap collateral end group transaction
 */
declare function prepareSwapCollateralInLoanEnd(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, oracle: Oracle, lpAssets: LPToken[], baseAssetIds: number[], params: SuggestedParams): Transaction[];
/**
 *
 * Returns a transaction to remove collateral from loan escrow.
 *
 * @param loanAppId - loan application to remove collateral in
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to remove collateral of
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction remove collateral transaction
 */
declare function prepareRemoveCollateralFromLoan(loanAppId: number, userAddr: string, escrowAddr: string, pool: Pool, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to borrow using loan escrow.
 *
 * @param loanAppId - loan application to borrow in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the borrow (typically the user)
 * @param pool - pool to borrow from
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param borrowAmount - amount to borrow of asset
 * @param maxStableRate - maximum stable rate of the borrow, if zero then borrow is interpreted as a variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] borrow group transaction
 */
declare function prepareBorrowFromLoan(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, pool: Pool, oracle: Oracle, lpAssets: LPToken[], baseAssetIds: number[], borrowAmount: number | bigint, maxStableRate: number | bigint, // if zero then variable borrow
params: SuggestedParams): Transaction[];
/**
 *
 * Returns a transaction to switch borrow type of a loan borrow.
 *
 * @param loanAppId - loan application to switch borrow type in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to switch borrow type of
 * @param maxStableRate - maximum stable rate of the borrow, if zero then interpreted as switching a stable rate borrow to a variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] switch borrow type transaction
 */
declare function prepareSwitchBorrowTypeInLoan(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, pool: Pool, maxStableRate: number | bigint, // ignored if not switching to stable borrow
params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to repay borrow in a loan escrow using assets sent from user.
 *
 * @param loanAppId - loan application to repay borrow in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the rewards if any (typically the user's deposit escrow)
 * @param reserveAddr - account address to receive the protocol revenue from the percentage of the accrued interest
 * @param pool - pool to repay borrow of
 * @param repayAmount - amount of borrow to repay
 * @param isStable - whether the borrow that is being repaid is a stable or variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] reduce borrow with transaction group transaction
 */
declare function prepareRepayLoanWithTxn(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, reserveAddr: ReserveAddress, pool: Pool, repayAmount: number | bigint, isStable: boolean, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a group transaction to repay borrow in a loan escrow using collateral.
 *
 * @param loanAppId - loan application to repay borrow in
 * @param poolManagerAppId - pool manager application
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param receiverAddr - account address to receive the rewards if any (typically the user's deposit escrow)
 * @param reserveAddr - account address to receive the protocol revenue from the percentage of the accrued interest
 * @param pool - pool to repay borrow and use collateral of
 * @param repayAmount - amount of borrow to repay expressed in terms of the asset
 * @param isStable - whether the borrow that is being repaid is a stable or variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] repay borrow with collateral group transaction
 */
declare function prepareRepayLoanWithCollateral(loanAppId: number, poolManagerAppId: number, userAddr: string, escrowAddr: string, receiverAddr: string, reserveAddr: ReserveAddress, pool: Pool, repayAmount: number | bigint, isStable: boolean, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to repay borrow in a loan escrow using assets sent from user.
 *
 * @param loanAppId - loan application to repay borrow in
 * @param poolManagerAppId - pool manager application
 * @param liquidatorAddr - account address for the liquidator
 * @param escrowAddr - account address for the loan escrow
 * @param reserveAddr - account address to receive the protocol revenue from the percentage of the accrued interest
 * @param collateralPool - pool to seize collateral of
 * @param borrowPool - pool to repay borrow of
 * @param oracle - oracle application to retrieve asset prices from
 * @param lpAssets - list of lp assets in loan
 * @param baseAssetIds - list of base asset ids in loan (non-lp assets)
 * @param repayAmount - amount of borrow to repay expressed in terms of borrow pool asset
 * @param minCollateralAmount - minimum collateral amount for the liquidator to receive expressed in terms of collateral pool f asset
 * @param isStable - whether the borrow that is being repaid is a stable or variable rate borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] liquidate group transaction
 */
declare function prepareLiquidateLoan(loanAppId: number, poolManagerAppId: number, liquidatorAddr: string, escrowAddr: string, reserveAddr: ReserveAddress, collateralPool: Pool, borrowPool: Pool, oracle: Oracle, lpAssets: LPToken[], baseAssetIds: number[], repayAmount: number | bigint, minCollateralAmount: number | bigint, isStable: boolean, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a transaction to rebalance up borrow in a loan escrow.
 *
 * @param loanAppId - loan application to rebalance up borrow in
 * @param poolManagerAppId - pool manager application*
 * @param rebalancerAddr - account address for the rebalancer
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to rebalance
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction rebalance up transaction
 */
declare function prepareRebalanceUpLoan(loanAppId: number, poolManagerAppId: number, rebalancerAddr: string, escrowAddr: string, pool: Pool, params: SuggestedParams): Transaction;
/**
 *
 * Returns a transaction to rebalance up borrow in a loan escrow.
 *
 * @param loanAppId - loan application to rebalance down borrow in
 * @param poolManagerAppId - pool manager application
 * @param rebalancerAddr - account address for the rebalancer
 * @param escrowAddr - account address for the loan escrow
 * @param pool - pool to rebalance
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction rebalance down transaction
 */
declare function prepareRebalanceDownLoan(loanAppId: number, poolManagerAppId: number, rebalancerAddr: string, escrowAddr: string, pool: Pool, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to remove a user's loan escrow and return its minimum balance to the user.
 *
 * @param loanAppId - loan application to remove loan in
 * @param userAddr - account address for the user
 * @param escrowAddr - account address for the loan escrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] remove and close out loan escrow group transaction
 */
declare function prepareRemoveUserLoan(loanAppId: number, userAddr: string, escrowAddr: string, params: SuggestedParams): Transaction[];
/**
 *
 * Returns a transaction to begin flash loan
 * Must be groped together with flash_loan_end call.
 *
 * @param pool - pool to borrow from
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the loan
 * @param borrowAmount - the amount of the asset to borrow
 * @param txnIndexForFlashLoanEnd - transaction index in the group transaction for the flash_loan_end call.
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction flash loan begin transaction
 */
declare function prepareFlashLoanBegin(pool: Pool, userAddr: string, receiverAddr: string, borrowAmount: number | bigint, txnIndexForFlashLoanEnd: number, params: SuggestedParams): Transaction;
/**
 *
 * Returns a group transaction to end flash loan.
 * Must be groped together with flash_loan_begin call.
 *
 * @param pool - pool borrowed from
 * @param userAddr - account address for the user
 * @param reserveAddr - account address to receive the protocol revenue from the flash loan fee
 * @param repaymentAmount - the amount of the asset to repay (borrow amount plus flash loan fee)
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] flash loan end group transaction
 */
declare function prepareFlashLoanEnd(pool: Pool, userAddr: string, reserveAddr: ReserveAddress, repaymentAmount: number | bigint, params: SuggestedParams): Transaction[];
/**
 *
 * Wraps given transactions with flash loan.
 *
 * @param txns - txns to wrap flash loan around
 * @param pool - pool to borrow from
 * @param userAddr - account address for the user
 * @param receiverAddr - account address to receive the loan
 * @param reserveAddr - account address to receive the protocol revenue from the flash loan fee
 * @param borrowAmount - the amount of the asset to borrow
 * @param params - suggested params for the transactions with the fees overwritten
 * @param flashLoanFee - fee for flash loan as 16 d.p integer (default 0.1%)
 * @returns Transaction[] group transaction wrapped with flash loan
 */
declare function wrapWithFlashLoan(txns: Transaction[], pool: Pool, userAddr: string, receiverAddr: string, reserveAddr: ReserveAddress, borrowAmount: number | bigint, params: SuggestedParams, flashLoanFee?: bigint): Transaction[];
export { retrieveLoanInfo, retrieveLoansLocalState, retrieveLoanLocalState, userLoanInfo, retrieveUserLoansInfo, retrieveUserLoanInfo, retrieveLiquidatableLoans, getMaxReduceCollateralForBorrowUtilisationRatio, getMaxBorrowForBorrowUtilisationRatio, prepareCreateUserLoan, prepareAddCollateralToLoan, prepareSyncCollateralInLoan, prepareReduceCollateralFromLoan, prepareSwapCollateralInLoanBegin, prepareSwapCollateralInLoanEnd, prepareRemoveCollateralFromLoan, prepareBorrowFromLoan, prepareSwitchBorrowTypeInLoan, prepareRepayLoanWithTxn, prepareRepayLoanWithCollateral, prepareLiquidateLoan, prepareRebalanceUpLoan, prepareRebalanceDownLoan, prepareRemoveUserLoan, prepareFlashLoanBegin, prepareFlashLoanEnd, wrapWithFlashLoan, };
//# sourceMappingURL=loan.d.ts.map