"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.greedyUnstakeAllocationStrategy = exports.greedyStakeAllocationStrategy = void 0;
const mathLib_1 = require("../../mathLib");
const formulae_1 = require("../formulae");
const constants_1 = require("./constants");
const greedyStakeAllocationStrategy = (consensusState, amount) => {
    const { proposersBalances, maxProposerBalance } = consensusState;
    const allocation = new Array(proposersBalances.length);
    // sort in ascending order
    const indexed = proposersBalances.map((proposer, index) => (Object.assign(Object.assign({}, proposer), { index })));
    indexed.sort((p0, p1) => Number(p0.algoBalance - p1.algoBalance));
    // allocate to proposers in greedy approach
    let remaining = BigInt(amount);
    for (let i = 0; i < allocation.length && i < constants_1.MAX_APPL_CALLS; i++) {
        const { algoBalance: proposerAlgoBalance, index: proposerIndex } = indexed[i];
        // under-approximate capacity to leave wiggle room
        const algoCapacity = (0, mathLib_1.maximum)(maxProposerBalance - proposerAlgoBalance - constants_1.FIXED_CAPACITY_BUFFER, BigInt(0));
        const allocate = (0, mathLib_1.minimum)(remaining, algoCapacity);
        allocation[proposerIndex] = allocate;
        // exit if fully allocated
        remaining -= allocate;
        if (remaining <= 0)
            break;
    }
    // handle case where still remaining
    if (remaining > 0)
        throw Error("Insufficient capacity to stake");
    return allocation;
};
exports.greedyStakeAllocationStrategy = greedyStakeAllocationStrategy;
const greedyUnstakeAllocationStrategy = (consensusState, amount) => {
    const { proposersBalances, minProposerBalance } = consensusState;
    const allocation = new Array(proposersBalances.length);
    // sort in descending order
    const indexed = proposersBalances.map((proposer, index) => (Object.assign(Object.assign({}, proposer), { index })));
    indexed.sort((p0, p1) => Number(p1.algoBalance - p0.algoBalance));
    // allocate to proposers in greedy approach
    let remaining = BigInt(amount);
    for (let i = 0; i < allocation.length && i < constants_1.MAX_APPL_CALLS; i++) {
        const { algoBalance: proposerAlgoBalance, index: proposerIndex } = indexed[i];
        // under-approximate capacity to leave wiggle room
        const algoCapacity = (0, mathLib_1.maximum)(proposerAlgoBalance - minProposerBalance - constants_1.FIXED_CAPACITY_BUFFER, BigInt(0));
        const xAlgoCapacity = (0, formulae_1.convertAlgoToXAlgoWhenDelay)(algoCapacity, consensusState);
        const allocate = (0, mathLib_1.minimum)(remaining, xAlgoCapacity);
        allocation[proposerIndex] = allocate;
        // exit if fully allocated
        remaining -= allocate;
        if (remaining <= 0)
            break;
    }
    // handle case where still remaining
    if (remaining > 0)
        throw Error("Insufficient capacity to unstake - override with your own allocation");
    return allocation;
};
exports.greedyUnstakeAllocationStrategy = greedyUnstakeAllocationStrategy;
//# sourceMappingURL=greedy.js.map