"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareUnstakeTransactions = exports.prepareClaimDelayedStakeTransactions = exports.prepareDelayedStakeTransactions = exports.prepareImmediateStakeTransactions = exports.prepareDummyTransaction = exports.getConsensusState = void 0;
const algosdk_1 = require("algosdk");
const crypto_1 = require("crypto");
const index_1 = require("../index");
const utils_1 = require("../utils");
const abiContracts_1 = require("./abiContracts");
const allocationStrategies_1 = require("./allocationStrategies");
/**
 *
 * Returns information regarding the given consensus application.
 *
 * @param algodClient - Algorand client to query
 * @param consensusConfig - consensus application and xALGO config
 * @returns ConsensusState current state of the consensus application
 */
function getConsensusState(algodClient, consensusConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        const [{ globalState: state }, { round, value: boxValue }, params] = yield Promise.all([
            (0, utils_1.getApplicationGlobalState)(algodClient, consensusConfig.appId),
            yield (0, utils_1.getApplicationBox)(algodClient, consensusConfig.appId, utils_1.enc.encode("pr")),
            yield algodClient.getTransactionParams().do(),
        ]);
        if (state === undefined)
            throw Error("Could not find xAlgo application");
        // xALGO rate
        const atc = new algosdk_1.AtomicTransactionComposer();
        atc.addMethodCall({
            sender: "Q5Q5FC5PTYQIUX5PGNTEW22UJHJHVVUEMMWV2LSG6MGT33YQ54ST7FEIGA",
            signer: (0, algosdk_1.makeEmptyTransactionSigner)(),
            appID: consensusConfig.appId,
            method: (0, algosdk_1.getMethodByName)(abiContracts_1.xAlgoABIContract.methods, "get_xalgo_rate"),
            methodArgs: [],
            suggestedParams: params,
        });
        const simReq = new algosdk_1.modelsv2.SimulateRequest({
            txnGroups: [],
            allowEmptySignatures: true,
            allowUnnamedResources: true,
            extraOpcodeBudget: 70000,
        });
        const { methodResults } = yield atc.simulate(algodClient, simReq);
        const { returnValue } = methodResults[0];
        const [algoBalance, xAlgoCirculatingSupply, balances] = returnValue;
        // proposers
        const proposersBalances = (0, utils_1.parseUint64s)(Buffer.from(balances).toString("base64")).map((balance, index) => ({
            address: (0, algosdk_1.encodeAddress)(boxValue.subarray(index * 32, (index + 1) * 32)),
            algoBalance: balance,
        }));
        // global state
        const timeDelay = BigInt((0, utils_1.getParsedValueFromState)(state, "time_delay") || 0);
        const numProposers = BigInt((0, utils_1.getParsedValueFromState)(state, "num_proposers") || 0);
        const minProposerBalance = BigInt((0, utils_1.getParsedValueFromState)(state, "min_proposer_balance") || 0);
        const maxProposerBalance = BigInt((0, utils_1.getParsedValueFromState)(state, "max_proposer_balance") || 0);
        const fee = BigInt((0, utils_1.getParsedValueFromState)(state, "fee") || 0);
        const premium = BigInt((0, utils_1.getParsedValueFromState)(state, "premium") || 0);
        const totalPendingStake = BigInt((0, utils_1.getParsedValueFromState)(state, "total_pending_stake") || 0);
        const totalActiveStake = BigInt((0, utils_1.getParsedValueFromState)(state, "total_active_stake") || 0);
        const totalRewards = BigInt((0, utils_1.getParsedValueFromState)(state, "total_rewards") || 0);
        const totalUnclaimedFees = BigInt((0, utils_1.getParsedValueFromState)(state, "total_unclaimed_fees") || 0);
        const canImmediateStake = Boolean((0, utils_1.getParsedValueFromState)(state, "can_immediate_mint"));
        const canDelayStake = Boolean((0, utils_1.getParsedValueFromState)(state, "can_delay_mint"));
        return {
            currentRound: Number(round),
            algoBalance,
            xAlgoCirculatingSupply,
            proposersBalances,
            timeDelay,
            numProposers,
            minProposerBalance,
            maxProposerBalance,
            fee,
            premium,
            totalPendingStake,
            totalActiveStake,
            totalRewards,
            totalUnclaimedFees,
            canImmediateStake,
            canDelayStake,
        };
    });
}
exports.getConsensusState = getConsensusState;
function prepareDummyTransaction(consensusConfig, senderAddr, params) {
    const atc = new algosdk_1.AtomicTransactionComposer();
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: consensusConfig.appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.xAlgoABIContract.methods, "dummy"),
        methodArgs: [],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 1000 }),
    });
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return txns[0];
}
exports.prepareDummyTransaction = prepareDummyTransaction;
// assumes txns has either structure:
// period 1 [appl call, appl call, ...]
// period 2 [transfer, appl call, transfer, appl call, ...]
function getTxnsAfterResourceAllocation(consensusConfig, consensusState, txnsToAllocateTo, additionalAddresses, period, senderAddr, params) {
    const { appId, xAlgoId } = consensusConfig;
    // make copy of txns
    const txns = txnsToAllocateTo.slice();
    const availableCalls = txns.length / period;
    // add xALGO asset and proposers box
    txns[period - 1].appForeignAssets = [xAlgoId];
    const box = { appIndex: appId, name: utils_1.enc.encode("pr") };
    const { boxes } = txns[period - 1];
    if (boxes) {
        boxes.push(box);
    }
    else {
        txns[period - 1].boxes = [box];
    }
    // get all accounts we need to add
    const accounts = additionalAddresses;
    consensusState.proposersBalances.forEach(({ address }) => accounts.push((0, algosdk_1.decodeAddress)(address)));
    // add accounts in groups of 4
    const MAX_FOREIGN_ACCOUNT_PER_TXN = 4;
    for (let i = 0; i < accounts.length; i += MAX_FOREIGN_ACCOUNT_PER_TXN) {
        // which txn to use
        const callNum = Math.floor(i / MAX_FOREIGN_ACCOUNT_PER_TXN) + 1;
        let txnIndex;
        // check if we need to add dummy call
        if (callNum <= availableCalls) {
            txnIndex = callNum * period - 1;
        }
        else {
            txns.unshift(prepareDummyTransaction(consensusConfig, senderAddr, params));
            txnIndex = 0;
        }
        // add proposer accounts
        txns[txnIndex].appAccounts = accounts.slice(i, i + 4);
    }
    return txns;
}
/**
 *
 * Returns a group transaction to stake ALGO and get xALGO immediately.
 *
 * @param consensusConfig - consensus application and xALGO config
 * @param consensusState - current state of the consensus application
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to send
 * @param minReceivedAmount - min amount of xALGO expected to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param proposerAllocations - determines which proposers the ALGO sent goes to
 * @param note - optional note to distinguish who is the minter (must pass to be eligible for revenue share)
 * @returns Transaction[] stake transactions
 */
function prepareImmediateStakeTransactions(consensusConfig, consensusState, senderAddr, amount, minReceivedAmount, params, proposerAllocations = (0, allocationStrategies_1.greedyStakeAllocationStrategy)(consensusState, amount), note) {
    const { appId } = consensusConfig;
    const atc = new algosdk_1.AtomicTransactionComposer();
    proposerAllocations.forEach((splitMintAmount, proposerIndex) => {
        if (splitMintAmount === BigInt(0))
            return;
        // calculate min received amount by proportional of total mint amount
        const splitMinReceivedAmount = (0, index_1.mulScale)(BigInt(minReceivedAmount), splitMintAmount, BigInt(amount));
        // generate txns for single proposer
        const { address: proposerAddress } = consensusState.proposersBalances[proposerIndex];
        const sendAlgo = {
            txn: (0, utils_1.transferAlgoOrAsset)(0, senderAddr, proposerAddress, splitMintAmount, params),
            signer: utils_1.signer,
        };
        atc.addMethodCall({
            sender: senderAddr,
            signer: utils_1.signer,
            appID: appId,
            method: (0, algosdk_1.getMethodByName)(abiContracts_1.xAlgoABIContract.methods, "immediate_mint"),
            methodArgs: [sendAlgo, proposerIndex, splitMinReceivedAmount],
            suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
            note,
        });
    });
    // allocate resources
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return getTxnsAfterResourceAllocation(consensusConfig, consensusState, txns, [], 2, senderAddr, params);
}
exports.prepareImmediateStakeTransactions = prepareImmediateStakeTransactions;
/**
 *
 * Returns a group transaction to stake ALGO and get xALGO after 320 rounds.
 *
 * @param consensusConfig - consensus application and xALGO config
 * @param consensusState - current state of the consensus application
 * @param senderAddr - account address for the sender
 * @param amount - amount of ALGO to send
 * @param params - suggested params for the transactions with the fees overwritten
 * @param includeBoxMinBalancePayment - whether to include ALGO payment to app for box min balance
 * @param proposerAllocations - determines which proposers the ALGO sent goes to
 * @param note - optional note to distinguish who is the minter (must pass to be eligible for revenue share)
 * @returns Transaction[] stake transactions
 */
function prepareDelayedStakeTransactions(consensusConfig, consensusState, senderAddr, amount, params, includeBoxMinBalancePayment = true, proposerAllocations = (0, allocationStrategies_1.greedyStakeAllocationStrategy)(consensusState, amount), note) {
    const { appId } = consensusConfig;
    const atc = new algosdk_1.AtomicTransactionComposer();
    proposerAllocations.forEach((splitMintAmount, proposerIndex) => {
        if (splitMintAmount === BigInt(0))
            return;
        // generate txns for single proposer
        const { address: proposerAddress } = consensusState.proposersBalances[proposerIndex];
        const sendAlgo = {
            txn: (0, utils_1.transferAlgoOrAsset)(0, senderAddr, proposerAddress, splitMintAmount, params),
            signer: utils_1.signer,
        };
        const nonce = (0, crypto_1.randomBytes)(2); // TODO: safeguard against possible clash?
        const boxName = Uint8Array.from([...utils_1.enc.encode("dm"), ...(0, algosdk_1.decodeAddress)(senderAddr).publicKey, ...nonce]);
        atc.addMethodCall({
            sender: senderAddr,
            signer: utils_1.signer,
            appID: appId,
            method: (0, algosdk_1.getMethodByName)(abiContracts_1.xAlgoABIContract.methods, "delayed_mint"),
            methodArgs: [sendAlgo, proposerIndex, nonce],
            boxes: [{ appIndex: appId, name: boxName }],
            suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
            note,
        });
    });
    // allocate resources
    let txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    txns = getTxnsAfterResourceAllocation(consensusConfig, consensusState, txns, [], 2, senderAddr, params);
    // add box min balance payment if specified
    if (includeBoxMinBalancePayment) {
        const minBalance = BigInt(36100);
        txns.unshift((0, utils_1.transferAlgoOrAsset)(0, senderAddr, (0, algosdk_1.getApplicationAddress)(appId), minBalance, params));
    }
    return txns;
}
exports.prepareDelayedStakeTransactions = prepareDelayedStakeTransactions;
/**
 *
 * Returns a group transaction to claim xALGO from delayed stake after 320 rounds.
 *
 * @param consensusConfig - consensus application and xALGO config
 * @param consensusState - current state of the consensus application
 * @param senderAddr - account address for the sender
 * @param receiverAddr - account address for the receiver
 * @param nonce - what was used to generate the delayed mint box
 * @param params - suggested params for the transactions with the fees overwritten
 * @returns Transaction[] stake transactions
 */
function prepareClaimDelayedStakeTransactions(consensusConfig, consensusState, senderAddr, receiverAddr, nonce, params) {
    const { appId } = consensusConfig;
    const atc = new algosdk_1.AtomicTransactionComposer();
    const boxName = Uint8Array.from([...utils_1.enc.encode("dm"), ...(0, algosdk_1.decodeAddress)(receiverAddr).publicKey, ...nonce]);
    atc.addMethodCall({
        sender: senderAddr,
        signer: utils_1.signer,
        appID: appId,
        method: (0, algosdk_1.getMethodByName)(abiContracts_1.xAlgoABIContract.methods, "claim_delayed_mint"),
        methodArgs: [receiverAddr, nonce],
        boxes: [{ appIndex: appId, name: boxName }],
        suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 3000 }),
    });
    // allocate resources
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return getTxnsAfterResourceAllocation(consensusConfig, consensusState, txns, [(0, algosdk_1.decodeAddress)(receiverAddr)], 1, senderAddr, params);
}
exports.prepareClaimDelayedStakeTransactions = prepareClaimDelayedStakeTransactions;
/**
 *
 * Returns a group transaction to unstake xALGO and get ALGO.
 *
 * @param consensusConfig - consensus application and xALGO config
 * @param consensusState - current state of the consensus application
 * @param senderAddr - account address for the sender
 * @param amount - amount of xALGO to send
 * @param minReceivedAmount - min amount of ALGO expected to receive
 * @param params - suggested params for the transactions with the fees overwritten
 * @param proposerAllocations - determines which proposers the ALGO received comes from
 * @param note - optional note to distinguish who is the burner (must pass to be eligible for revenue share)
 * @returns Transaction[] unstake transactions
 */
function prepareUnstakeTransactions(consensusConfig, consensusState, senderAddr, amount, minReceivedAmount, params, proposerAllocations = (0, allocationStrategies_1.greedyUnstakeAllocationStrategy)(consensusState, amount), note) {
    const { appId, xAlgoId } = consensusConfig;
    const atc = new algosdk_1.AtomicTransactionComposer();
    proposerAllocations.forEach((splitBurnAmount, proposerIndex) => {
        if (splitBurnAmount === BigInt(0))
            return;
        // calculate min received amount by proportional of total burn amount
        const splitMinReceivedAmount = (0, index_1.mulScale)(BigInt(minReceivedAmount), splitBurnAmount, BigInt(amount));
        // generate txns for single proposer
        const sendXAlgo = {
            txn: (0, utils_1.transferAlgoOrAsset)(xAlgoId, senderAddr, (0, algosdk_1.getApplicationAddress)(appId), splitBurnAmount, params),
            signer: utils_1.signer,
        };
        atc.addMethodCall({
            sender: senderAddr,
            signer: utils_1.signer,
            appID: appId,
            method: (0, algosdk_1.getMethodByName)(abiContracts_1.xAlgoABIContract.methods, "burn"),
            methodArgs: [sendXAlgo, proposerIndex, splitMinReceivedAmount],
            suggestedParams: Object.assign(Object.assign({}, params), { flatFee: true, fee: 2000 }),
            note,
        });
    });
    // allocate resources
    const txns = atc.buildGroup().map(({ txn }) => {
        txn.group = undefined;
        return txn;
    });
    return getTxnsAfterResourceAllocation(consensusConfig, consensusState, txns, [], 2, senderAddr, params);
}
exports.prepareUnstakeTransactions = prepareUnstakeTransactions;
//# sourceMappingURL=consensus.js.map